/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.common;

import com.kerio.im.common.error.Code;
import com.kerio.im.common.error.ServiceException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.ServiceLoader;

/**
 * Make loading services smooth.
 * @author tskoda
 */
public final class Lookup {
    
    private static final Map<Class, ServiceLoader> CACHE = new HashMap<Class, ServiceLoader>();
    
    /**
     * Returns instance of Service interface. Services are ment to be unique in
     * current system runtime. Always the same instance is returned on multiple 
     * calls with single type parameter.
     * 
     * @param type interface
     * @return implementation of interface if exists and is unique, never null
     * @throws ServiceException when multiple or no implementation found
     */
    public static  <T extends Service> T service(Class<T> type) throws ServiceException {
        T result = null;
        
        ServiceLoader<T> providers = CACHE.get(type);
        if (providers == null) {
            providers = ServiceLoader.load(type);
            CACHE.put(type, providers);
        }
        
        Iterator<T> iterator = providers.iterator();
        if (iterator.hasNext()) {
            T uniqueImplementation = iterator.next();
            
            if (!iterator.hasNext()) {
                result = uniqueImplementation;
            } else {
                throw new ServiceException("Multiple implementations of " + type.getName() + " found", Code.CLASS_PATH_ERROR);
            }
        } 
        
        if (result == null) {
            throw new ServiceException("Implementation of " + type.getName() + " is missing", Code.CLASS_PATH_ERROR);
        }
        
        return result;
    }
    
    /**
     * Return all found instances of interface type. More implelentation could 
     * exists. New instance of an interface is created on every method call.
     * @param type interface
     * @return List of implementation of interface or abstract class type, never return null
     */
    public static <T extends ServiceUnit> List<T> all(Class<T> type) {
        ArrayList<T> result = new ArrayList<T>();
        ServiceLoader<T> providers = ServiceLoader.load(type);
        Iterator<T> iterator = providers.iterator();
        
        if (iterator.hasNext()) {
            result.add(iterator.next());
        }
        
        return result;
    }
    
}
