/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.common;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author tskoda
 */
public class FileUtilTest {

    public FileUtilTest() {
    }

    @Test
    public void testDeleteFile() throws Exception {
        File toDelete = File.createTempFile("todelete", "tmp");
        try {
            assertTrue(toDelete.isFile());
            assertTrue(FileUtil.deleteFile(toDelete.getParentFile(), toDelete.getName()));
            assertFalse(toDelete.exists());
        } finally {
            toDelete.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFileNotFound() throws Exception {
        File folder = createTempFolder();
        File toDelete = new File(folder, "notfound.tmp");
        try {
            assertFalse(toDelete.exists());
            assertTrue(FileUtil.deleteFile(toDelete.getParentFile(), toDelete.getName()));
            assertFalse(toDelete.exists());
        } finally {
            folder.deleteOnExit();
        }
    }

    @Test(expected = IOException.class)
    public void testDeleteFileIsDirectory() throws Exception {
        File folder = createTempFolder();
        try {
            assertTrue(folder.isDirectory());
            FileUtil.deleteFile(folder.getParentFile(), folder.getName());
        } finally {
            folder.deleteOnExit();
        }
    }

    public void testDeleteFileLocked() throws Exception {
        File folder = createTempFolder();
        File toDelete = File.createTempFile("locked", "tmp", folder);
        FileInputStream fio = new FileInputStream(toDelete);
        try {
            FileLock lock = fio.getChannel().lock();
            assertTrue(lock.isValid());
            try {
                assertTrue(toDelete.isFile());
                assertFalse(FileUtil.deleteFile(folder, toDelete.getName()));
                assertTrue(toDelete.isFile());
            } finally {
                lock.release();
            }
        } finally {
            try {
                fio.close();
            } catch (IOException e) {
                System.err.println("Failed to close input stream");
            }

            folder.deleteOnExit();
            toDelete.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFolder() throws Exception {
        File folder = createTempFolder();
        try {
            assertTrue(folder.isDirectory());
            assertTrue(FileUtil.deleteFolder(folder));
            assertFalse(folder.exists());
        } finally {
            folder.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFolderNotFound() throws Exception {
        File folder = createTempFolder();
        File notfound = new File(folder, "notfound");
        try {
            assertFalse(notfound.isDirectory());
            assertTrue(FileUtil.deleteFolder(notfound));
            assertFalse(notfound.isDirectory());
        } finally {
            folder.deleteOnExit();
        }
    }

    @Test(expected = IOException.class)
    public void testDeleteFolderIsFile() throws Exception {
        File folder = createTempFolder();
        File file = createTempFile(folder);
        try {
            assertTrue(file.isFile());
            assertTrue(FileUtil.deleteFolder(file));
        } finally {
            folder.deleteOnExit();
            file.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFolderWithFile() throws Exception {
        File folder = createTempFolder();
        File file = createTempFile(folder);
        try {
            assertTrue(folder.isDirectory());
            assertTrue(file.isFile());
            assertFalse(FileUtil.deleteFolder(folder));
            assertTrue(folder.isDirectory());
            assertTrue(file.isFile());
        } finally {
            folder.deleteOnExit();
            file.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFolderWithDirectory() throws Exception {
        File folder = createTempFolder();
        File subfolder = createTempFolder(folder);
        try {
            assertTrue(folder.isDirectory());
            assertTrue(subfolder.isDirectory());
            assertFalse(FileUtil.deleteFolder(folder));
            assertTrue(folder.isDirectory());
            assertTrue(subfolder.isDirectory());
        } finally {
            folder.deleteOnExit();
            subfolder.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFolderWithFilesEmpty() throws IOException {
        File folder = createTempFolder();
        try {
            assertTrue(folder.isDirectory());
            assertTrue(folder.listFiles().length == 0);
            assertTrue(FileUtil.deleteFolderWithFiles(folder));
            assertFalse(folder.exists());
        } finally {
            folder.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFolderWithFilesFileOnly() throws IOException {
        File folder = createTempFolder();
        File one = createTempFile(folder);
        File two = createTempFile(folder);
        try {
            assertTrue(folder.isDirectory());
            assertTrue(one.isFile() && one.getParentFile().equals(folder));
            assertTrue(two.isFile() && two.getParentFile().equals(folder));

            assertTrue(FileUtil.deleteFolderWithFiles(folder));

            assertFalse(folder.exists());
            assertFalse(one.exists());
            assertFalse(two.exists());

        } finally {
            folder.deleteOnExit();
            one.deleteOnExit();
            two.deleteOnExit();
        }
    }

    @Test
    public void testDeleteFolderWithFilesAndSubdirectory() throws IOException {
        File folder = createTempFolder();
        File one = createTempFile(folder);
        File two = createTempFolder(folder);
        try {
            assertTrue(folder.isDirectory());
            assertTrue(one.isFile());
            assertTrue(two.isDirectory());

            assertFalse(FileUtil.deleteFolderWithFiles(folder));

            assertTrue(folder.exists());
            assertFalse(one.exists());
            assertTrue(two.exists());

        } finally {
            folder.deleteOnExit();
            one.deleteOnExit();
            two.deleteOnExit();
        }
    }

    @Test(expected = IOException.class)
    public void testDeleteFolderWithFilesNotADirectory() throws IOException {
        File file = File.createTempFile("todelete", "tmp");
        try {
            assertTrue(file.isFile());
            assertFalse(FileUtil.deleteFolderWithFiles(file));
        } finally {
            file.deleteOnExit();

        }
    }

    public static File createTempFile(File folder) throws IOException {
        return File.createTempFile("test", "tmp", folder);
    }

    public static File createTempFolder(File folder) throws IOException {
        File temp;

        temp = File.createTempFile("db_", Long.toString(System.nanoTime()), folder);

        if (!(temp.delete())) {
            throw new IOException("Could not delete temp file: " + temp.getAbsolutePath());
        }

        if (!(temp.mkdir())) {
            throw new IOException("Could not create temp directory: " + temp.getAbsolutePath());
        }
        return temp;
    }

    public static File createTempFolder() throws IOException {
        return createTempFolder(null);
    }
}
