/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core;

import com.kerio.im.core.api.CoreService;
import com.kerio.im.core.api.service.AccessPolicyService;
import com.kerio.im.core.api.service.AuthService;
import com.kerio.im.core.api.service.ConfigurationService;
import com.kerio.im.core.api.service.ConversationService;
import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.core.api.service.MessageService;
import com.kerio.im.core.api.service.SecurityService;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.api.service.ThreadService;
import com.kerio.im.core.api.service.VCardService;
import com.kerio.im.core.auth.Authentication;
import com.kerio.im.core.auth.OnlineAuthentication;
import com.kerio.im.core.backup.BackupListener;
import com.kerio.im.core.cert.Certificates;
import com.kerio.im.core.configuration.StoreConfiguration;
import com.kerio.im.core.conversation.archive.Archive;
import com.kerio.im.core.conversation.ConversationServiceImpl;
import com.kerio.im.core.domain.Domains;
import com.kerio.im.core.logging.LoggingManager;
import com.kerio.im.core.remote.RemoteControl;
import com.kerio.im.core.service.SecurityManager;
import com.kerio.im.core.service.ServiceManager;
import com.kerio.im.core.user.UserAccessPolicy;
import com.kerio.im.core.user.Users;
import com.kerio.im.core.vcard.VCards;
import com.kerio.im.external.ExtService;
import com.kerio.im.xmpp.api.XmppService;

/**
 *
 * @author tskoda
 */
public class CoreRuntime implements CoreService {

    private final LoggingManager logging;
    private final StoreConfiguration configuration;
    private final Authentication offlineAuthManager;
    private final RemoteControl remote;
    private final ServiceBus bus;
    private final ThreadPool threadPool;
    private final ErrorHandler errorHandler;
    private final UserAccessPolicy accessPolicy;
    private final ConversationServiceImpl conversations;
    private final GroupService groups;
    private final SecurityManager ciphersManager;
    //these services become active since xmppservice accessible
    private XmppService xmppService;
    private ServiceManager ports;
    //these services become active since external service accessible
    private ExtService extService;
    private Certificates certificates;
    private Domains domains;
    private Users users;
    private OnlineAuthentication authManager;
    private BackupListener backup;
    private VCards vcards;
    private Archive archive;
    

    public CoreRuntime(LoggingManager logging, StoreConfiguration configuration, Authentication offlineAuthManager, RemoteControl remote, ServiceBus bus, ThreadPool threadPool, ErrorHandler errorHandler, UserAccessPolicy accessPolicy, ConversationServiceImpl conversations, GroupService groups, SecurityManager ciphersManager) {
        this.logging = logging;
        this.configuration = configuration;
        this.offlineAuthManager = offlineAuthManager;
        this.remote = remote;
        this.bus = bus;
        this.threadPool = threadPool;
        this.errorHandler = errorHandler;
        this.accessPolicy = accessPolicy;
        this.conversations = conversations;
        this.groups = groups;
        this.ciphersManager = ciphersManager;
    }

    public void setXmppService(final XmppService xmppService) {
        this.xmppService = xmppService;
        /* now we are listening on service port, waiting for configuration message */
        this.ports = new ServiceManager(configuration, bus, this.xmppService.ports());
    }

    public void setExtService(final ExtService extService) {
        this.extService = extService;
        this.certificates = new Certificates(extService, this.xmppService.domains(), this.bus, this.configuration);
        this.vcards = new VCards(this.xmppService, extService, this.bus);
        this.users = new Users(this.configuration, this.xmppService, this.extService, this.bus, this.threadPool);
        this.domains = new Domains(this.configuration, this.bus, this.threadPool, this.extService, this.xmppService.domains());
        this.authManager = new OnlineAuthentication(this.users, this.ports, this.extService);
        this.offlineAuthManager.set(this.authManager);
        this.ports.set(extService, this.bus, this.threadPool, this.remote, this.domains);
        this.backup = new BackupListener(this.xmppService.domains());
        this.backup.subscribe(bus);
        this.accessPolicy.setExtService(this.users, extService, this.bus);
        this.archive = Archive.open(this.conversations, this.extService, this.configuration, this.bus);
    }

    @Override
    public ThreadService threads() {
        return this.threadPool;
    }

    @Override
    public ConfigurationService configuration() {
        return this.configuration;
    }

    @Override
    public AuthService auth() {
        return this.offlineAuthManager;
    }

    @Override
    public MessageService message() {
        return this.remote;
    }

    @Override
    public ServiceBus bus() {
        return this.bus;
    }

    @Override
    public ErrorHandler errors() {
        return this.errorHandler;
    }

    @Override
    public VCardService vcards() {
        return this.vcards;
    }

    @Override
    public AccessPolicyService accessPolicy() {
        return this.accessPolicy;
    }

    @Override
    public ConversationService conversations() {
        return this.conversations;
    }

    @Override
    public GroupService groups() {
        return groups;
    }

    @Override
    public SecurityService securityService() {
        return this.ciphersManager;
    }
    
}
