/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.configuration;

import com.kerio.im.common.Key;
import com.kerio.im.common.TypedProperties;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.core.api.service.ConfigurationService;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.remote.CommandEvent;
import java.io.File;
import java.util.Map;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class StoreConfiguration implements ConfigurationService {

    private static final Logger LOG = Logger.getLogger(StoreConfiguration.class.getName());
    /**
     * Bridge configuration. Serves as a store for all configuration values
     */
    private final TypedProperties properties;

    /**
     * Creates ConfigurationManager with empty global configuration. Call open
     * method before use.
     *
     * @param managers
     */
    public StoreConfiguration(Map<String, String> cmdProperties, ServiceBus bus) {
        this.properties = new TypedProperties();
        cmd(cmdProperties);
        bus.subscribe(CommandEvent.class, new CommandEventListener(this.properties));
    }

    protected final void cmd(Map<String, String> cmdProperties) {
        LOG.fine("Processing startup configuration");

        //required
        File workingDir = TypedProperties.file(GlobalKey.SERVICE_DIR, cmdProperties);
        if (!workingDir.isDirectory()) {
            throw new IllegalArgumentException("Working directory " + workingDir.getAbsolutePath() + " does not exist");
        } else {
            this.properties.put(GlobalKey.SERVICE_DIR, workingDir);
        }
        //required
        File dataDir = TypedProperties.file(GlobalKey.DATA_DIR, cmdProperties);
        if (!dataDir.isDirectory()) {
            throw new IllegalArgumentException("Store directory " + dataDir.getAbsolutePath() + " does not exist");
        } else {
            this.properties.put(GlobalKey.DATA_DIR, dataDir);
        }

        //required
        this.properties.put(GlobalKey.SERVICE_REMOTE_PORT, TypedProperties.integer(GlobalKey.SERVICE_REMOTE_PORT, cmdProperties));

        //required
        this.properties.put(GlobalKey.SERVICE_REMOTE_LOGIN, TypedProperties.string(GlobalKey.SERVICE_REMOTE_LOGIN, cmdProperties));

        //required
        this.properties.put(GlobalKey.SERVICE_REMOTE_DOMAIN, TypedProperties.string(GlobalKey.SERVICE_REMOTE_DOMAIN, cmdProperties));

        //TODO make it required
        this.properties.put(GlobalKey.HOSTNAME, TypedProperties.string(GlobalKey.HOSTNAME, cmdProperties, "localhost"));

        //optional
        this.properties.put(GlobalKey.ARCHIVE_SUBDIR, TypedProperties.string(GlobalKey.ARCHIVE_SUBDIR, cmdProperties, null));

        //required
        this.properties.put(GlobalKey.ROSTER_LIMIT, TypedProperties.integer(GlobalKey.ROSTER_LIMIT, cmdProperties));

        LOG.info("Processing startup configuration successful");
    }

    public <T> void put(Key<T> key, T value) {
        this.properties.put(key, value);
    }

    @Override
    public <T> T get(Key<T> key) {
        return this.properties.get(key);
    }

    @Override
    public <T> T get(Key<T> key, T defValue) {
        return properties.get(key, defValue);
    }

    @Override
    public TypedProperties all() {
        return TypedProperties.asCopy(this.properties);
    }

    protected static class CommandEventListener implements ServiceBus.EventListener<CommandEvent> {

        private static final String COMMAND = "config";
        public static final int TEN_MINUTES_IN_SECONDS = 60 * 10;
        private static final long FILE_LIMIT_DEFAULT = 50l;

        private final TypedProperties properties;

        public CommandEventListener(TypedProperties properties) {
            this.properties = properties;
        }

        @Override
        public void event(CommandEvent event) {
            if (event.interested(COMMAND)) {
                //optional
                this.properties.put(GlobalKey.REFRESH_PERIOD_KEY, TypedProperties.integer(GlobalKey.REFRESH_PERIOD_KEY, event.getContent(), TEN_MINUTES_IN_SECONDS));
                //optional
                this.properties.put(GlobalKey.ARCHIVE_FILE_LIMIT, TypedProperties.loong(GlobalKey.ARCHIVE_FILE_LIMIT, event.getContent(), FILE_LIMIT_DEFAULT));
            }
        }
    }
}
