/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class ArchiveBufferSchema {

    private static final Logger LOG = Logger.getLogger(ArchiveBufferSchema.class.getName());

    private static final String VERSION = "1.0";
    

    public enum SCHEMA_NAME {

        archive
    }
    
    public enum LIMITS {
        JID(511),
        TEXT(32672),
        FILENAME(255);
        
        public final int limit;

        private LIMITS(int limit) {
            this.limit = limit;
        }

    }

    public enum SCHEMA {

        //table properties
        tbl_properties(
                "create table %1$s.properties ("
                + "      name varchar(50) NOT NULL PRIMARY KEY,"
                + "     value varchar(100))", SCHEMA_NAME.archive),
        //table users
        tbl_users(
                "create table %1$s.users ("
                + "        id bigint GENERATED BY DEFAULT AS IDENTITY NOT NULL PRIMARY KEY,"
                + "       jid varchar(511))", SCHEMA_NAME.archive), //255 username, 255 domain, 1 @ - we do not store resource part
        idx_users_jid(
                "create unique index idx_users_jid "
                + "         on %1$s.users ( jid )", SCHEMA_NAME.archive),
        //table records
        tbl_records(
                "create table %1$s.records ("
                + "        id bigint GENERATED BY DEFAULT AS IDENTITY NOT NULL PRIMARY KEY,"
                + "      when timestamp DEFAULT CURRENT_TIMESTAMP,"
                + "    sender bigint NOT NULL references %1$s.users(id),"
                + " recipient bigint NOT NULL references %1$s.users(id),"
                + "      type int NOT NULL,"
                + "      text varchar(32672) default NULL,"
                + " file_name varchar(255) default NULL,"
                + " file_size bigint default NULL)", SCHEMA_NAME.archive),
        idx_records_sender(
                "create index idx_records_sender on %1$s.records ( sender )", SCHEMA_NAME.archive),
        idx_records_recipient(
                "create index idx_records_recipient on %1$s.records ( recipient )", SCHEMA_NAME.archive),
        idx_records_when(
                "create index idx_records_when on %1$s.records ( when )", SCHEMA_NAME.archive);

        final String sql;

        private SCHEMA(String create, Object... values) {
            this.sql = String.format(create, values);
        }
    }

    public enum DATA {

        schema_version("insert into %1$s.properties (name, value) values ('schema-version', '%2$s')", SCHEMA_NAME.archive, VERSION);

        final String sql;

        private DATA(String insert, Object... params) {
            this.sql = String.format(insert, params);
        }
    }

    public enum SQL {

        getParticipants(
                " select u.id as id, u.jid as jid"
                + " from %1$s.users u "
                + "where u.jid = ? or u.jid = ?", SCHEMA_NAME.archive),
        addParticipant(
                "insert into %1$s.users (jid) "
                + "              values (?)", SCHEMA_NAME.archive),
        storeRecord(
                "insert into %1$s.records (sender, recipient, type, text, file_name, file_size)"
                + "                values (     ?,         ?,    ?,    ?,         ?,         ?)", SCHEMA_NAME.archive);

        public final String sql;

        private SQL(String sql, Object... params) {
            this.sql = String.format(sql, params);
        }

    }

    public enum RECORD_TYPE {

        message,
        file;

        public int toInt() {
            return ordinal();
        }

    }

    public static void apply(DerbyConnectionFactory factory) throws SQLException {
        Connection connection = factory.newConnection();
        try {
            Statement statement = connection.createStatement();
            try {
                LOG.log(Level.INFO, "Applying schema version {0}", VERSION);
                for (SCHEMA command : SCHEMA.values()) {
                    LOG.log(Level.INFO, "    + create {0}", command);
                    statement.execute(command.sql);
                }
                for (DATA data : DATA.values()) {
                    LOG.log(Level.INFO, "    + insert {0}", data);
                    statement.execute(data.sql);
                }
            } finally {
                DerbyConnectionFactory.close(statement);
            }
            connection.commit();
        } catch (SQLException e) {
            DerbyConnectionFactory.rollback(connection);
        } finally {
            DerbyConnectionFactory.close(connection);
        }
    }

    private ArchiveBufferSchema() {
    }
}
