/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import com.kerio.im.common.error.Code;
import com.kerio.im.common.error.ServiceException;
import com.kerio.im.external.ExtService;
import java.io.File;
import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class ArchiveBufferStorage implements ConversationStorage {

    //private static final SimpleDateFormat DB_NAME_FORMAT = new SimpleDateFormat("yyyy-MMM-'active'");
    private static final String DB_NAME = "active";
    private static final Logger LOG = Logger.getLogger(ArchiveBufferStorage.class.getName());

    public static ArchiveBufferStorage open(ExtService.ArchiveSettings archiveSettings, String subfolder) {
        try {
            DerbyConnectionFactory database = DerbyConnectionFactory.create(DerbyDestination.open(new File(archiveSettings.getPath(), subfolder), DB_NAME));
            ArchiveBufferStorage result = new ArchiveBufferStorage(database);
            try {
                result.open();
                return result;
            } catch (Exception e) {
                result.close(); //silent close
                throw new ServiceException("Failed to open conversation storage", e, Code.INTERNAL_SERVER_ERROR);
            }
        } catch (IOException e) {
            throw new ServiceException("Failed to open database storage", e, Code.INTERNAL_SERVER_ERROR);
        }
    }

    private final DerbyConnectionFactory connectionFactory;
    private Connection connection;
    private PreparedStatement getParticipants;
    private PreparedStatement addParticipant;
    private PreparedStatement storeRecord;

    public ArchiveBufferStorage(DerbyConnectionFactory connectionFactory) {
        this.connectionFactory = connectionFactory;
    }

    void open() throws IOException, SQLException {
        if (this.connectionFactory.open()) {
            ArchiveBufferSchema.apply(this.connectionFactory);
        } else {
            //check schema version and perform upgrade
        }
        this.connection = this.connectionFactory.newConnection();
        this.getParticipants = this.connection.prepareStatement(ArchiveBufferSchema.SQL.getParticipants.sql);
        this.addParticipant = this.connection.prepareStatement(ArchiveBufferSchema.SQL.addParticipant.sql, Statement.RETURN_GENERATED_KEYS);
        this.storeRecord = this.connection.prepareStatement(ArchiveBufferSchema.SQL.storeRecord.sql);
    }

    @Override
    public void close() {
        synchronized (this) {
            try {
                DerbyConnectionFactory.close(getParticipants);
                DerbyConnectionFactory.close(addParticipant);
                DerbyConnectionFactory.close(storeRecord);
                DerbyConnectionFactory.close(this.connection); //silently closes connection
            } finally {
                this.connectionFactory.shutdown();
            }
        }
    }

    private Integer addUser(String jid) throws SQLException {
        Integer result = null;
        synchronized (this) {
            try {
                this.addParticipant.setString(1, jid);
                this.addParticipant.execute();

                ResultSet idResult = this.addParticipant.getGeneratedKeys();
                try {
                    if (idResult.next()) {
                        result = idResult.getInt(1);
                    } else {
                        throw new SQLException("Failed to insert participant " + jid);
                    }
                } finally {
                    DerbyConnectionFactory.close(idResult);
                }
                this.connection.commit();
            } catch (SQLException e) {
                DerbyConnectionFactory.rollback(connection);
                throw e;
            }
        }
        return result;
    }

    private Map<String, Integer> getUserIds(String sender, String recipient) throws SQLException {
        Map<String, Integer> result = new HashMap<String, Integer>(2);
        synchronized (this) {
            if (sender != null) {
                this.getParticipants.setString(1, sender);
            } else {
                this.getParticipants.setNull(1, Types.VARCHAR);
            }

            if (recipient != null) {
                this.getParticipants.setString(2, recipient);
            } else {
                this.getParticipants.setNull(2, Types.VARCHAR);
            }
            //read jids
            ResultSet idResult = this.getParticipants.executeQuery();
            try {
                while (idResult.next()) {
                    result.put(idResult.getString("jid"), idResult.getInt("id"));
                }
            } finally {
                DerbyConnectionFactory.close(idResult);
            }

        }
        return result;
    }

    private void storeRecord(String sender, String recipient, String message) {
        synchronized (this) {
            try {
                Map<String, Integer> userIds = getUserIds(sender, recipient);
                this.storeRecord.setInt(1, userIds.containsKey(sender) ? userIds.get(sender) : addUser(sender));
                this.storeRecord.setInt(2, userIds.containsKey(recipient) ? userIds.get(recipient) : addUser(recipient));
                this.storeRecord.setInt(3, ArchiveBufferSchema.RECORD_TYPE.message.toInt());
                this.storeRecord.setString(4, message);
                this.storeRecord.setNull(5, Types.VARCHAR);
                this.storeRecord.setNull(6, Types.BIGINT);
                this.storeRecord.execute();
                this.connection.commit();
            } catch (SQLException e) {
                DerbyConnectionFactory.rollback(this.connection);
                LOG.log(Level.INFO, "Failed to store message to buffer storage", e);
            }
        }
    }

    private void storeFileTransfer(String sender, String recipient, String fileName, long size) {
        synchronized (this) {
            try {
                Map<String, Integer> userIds = getUserIds(sender, recipient);
                this.storeRecord.setInt(1, userIds.containsKey(sender) ? userIds.get(sender) : addUser(sender));
                this.storeRecord.setInt(2, userIds.containsKey(recipient) ? userIds.get(recipient) : addUser(recipient));
                this.storeRecord.setInt(3, ArchiveBufferSchema.RECORD_TYPE.file.toInt());
                this.storeRecord.setNull(4, Types.VARCHAR);
                this.storeRecord.setString(5, fileName);
                this.storeRecord.setLong(6, size);
                this.storeRecord.execute();
                this.connection.commit();
            } catch (SQLException e) {
                DerbyConnectionFactory.rollback(this.connection);
                LOG.log(Level.INFO, "Failed to store message to buffer storage", e);
            }
        }
    }

    @Override
    public SessionStorage latest(String room) {
        return new ChatSessionStorage(room);
    }

    @Override
    public SessionStorage latest(String sender, String recipient) {
        return new ChatSessionStorage(recipient);
    }

    private class ChatSessionStorage implements SessionStorage {

        private final String recipient;

        public ChatSessionStorage(String recipient) {
            this.recipient = recipient;
        }

        @Override
        public void say(String who, String what, String raw) {
            ArchiveBufferStorage.this.storeRecord(who, this.recipient, what);
        }

        @Override
        public void transfer(String who, String fileName, long size, String raw) {
            ArchiveBufferStorage.this.storeFileTransfer(who, this.recipient, fileName, size);
        }

    }

}
