/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import com.kerio.im.core.api.event.ShutdownEvent;
import com.kerio.im.core.api.service.ConversationService;
import com.kerio.im.core.api.service.ServiceBus;
import java.time.LocalDateTime;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;

/**
 * Allows registering listeners for gathering content of conversations.
 *
 * @author tskoda
 */
public class ConversationServiceImpl implements ConversationService {

    public interface AsyncConversationService extends ConversationService {

        ExecutorService getExecutor();
        void close();
    }

    private final ConcurrentHashMap<String, AsyncConversationService> listeners;
    private final ServiceBus bus;

    public ConversationServiceImpl(ServiceBus bus) {
        this.listeners = new ConcurrentHashMap<String, AsyncConversationService>();
        this.bus = bus;
        this.bus.subscribe(ShutdownEvent.class, new ShutdownEventListener());
    }

    public void register(String name, AsyncConversationService asyncService) {
        this.listeners.put(name, asyncService);
    }

    public AsyncConversationService unregister(String name) {
        return this.listeners.remove(name);
    }

    @Override
    public void message(final String from, final String to, final MessageInfo info, final LocalDateTime time, final String raw) {
        for (final AsyncConversationService listener : this.listeners.values()) {
            listener.getExecutor().submit(new Runnable() {
                @Override
                public void run() {
                    listener.message(from, to, info, time, raw);
                }
            });
        }
    }

    @Override
    public void file(final String from, final String to, final FileInfo info,  final LocalDateTime time, final String raw) {
        for (final AsyncConversationService listener : this.listeners.values()) {
            listener.getExecutor().submit(new Runnable() {
                @Override
                public void run() {
                    listener.file(from, to, info, time, raw);
                }
            });
        }
    }

    private class ShutdownEventListener implements ServiceBus.EventListener<ShutdownEvent> {

        @Override
        public void event(ShutdownEvent event) {
            ConversationServiceImpl.this.listeners.clear();
        }
    }

}
