/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import com.google.common.io.ByteStreams;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

/**
 *
 * @author tskoda
 */
public class FileFactory {

    private static final Logger LOG = Logger.getLogger(FileFactory.class.getName());

    public enum TYPE {

        TXT(".txt"),
        ZIP(".zip"),
        PART(".part");

        public final String suffix;

        private TYPE(String suffix) {
            this.suffix = suffix;
        }
    }

    static final String FILE_NAME_PREFIX = "xmpp-archive-";
    static final String FILE_NAME = "%1$tY-%1$tm-%1$td-%1$tH-%1$tM";
    static final String FILE_NAME_FORMAT = FILE_NAME_PREFIX + FILE_NAME + TYPE.TXT.suffix;
    static final String FILE_NAME_TMP_FORMAT = FILE_NAME_PREFIX + FILE_NAME + TYPE.PART.suffix;
    static final String FILE_NAME_ZIP_FORMAT = FILE_NAME_PREFIX + FILE_NAME + TYPE.ZIP.suffix;

    private static class SuffixFileFilter implements FilenameFilter {

        private final TYPE type;

        public SuffixFileFilter(TYPE type) {
            this.type = type;
        }

        @Override
        public boolean accept(File dir, String name) {
            return name.endsWith(type.suffix) && name.startsWith(FILE_NAME_PREFIX);
        }

    }

    private final File parent;

    public FileFactory(File parent) throws IOException {
        this.parent = parent;
    }

    public void zip(File source) throws IOException {

        File finalDestination = withType(source, TYPE.ZIP);
        if (finalDestination.exists()) {
            throw new IOException("Failed to create zip file, destination " + finalDestination.getAbsolutePath() + " already exists");
        }

        File tempDestination = withType(source, TYPE.PART);
        if (tempDestination.exists() && !tempDestination.delete()) {
            throw new IOException("Failed to delete temporary zip file " + tempDestination.getAbsolutePath());
        }

        zip(source, tempDestination);

        if (!tempDestination.renameTo(finalDestination)) {
            LOG.log(Level.INFO, "Failed to rename {0} to {1}", new Object[]{source.getAbsolutePath(), tempDestination.getAbsolutePath()});
        }

        if ((finalDestination.exists()) && (finalDestination.length() > 0)) { //delete original only if zip exists
            if (!source.delete()) {
                LOG.log(Level.INFO, "Failed to delete source file {0}", source.getAbsolutePath());
            }
        } else {
            throw new IOException("Failed to create zip file " + source.getAbsolutePath());
        }
    }

    public void zip(File source, File destination) throws IOException {
        if (!source.exists()) {
            throw new IOException("Failed to create zip file, source " + source.getAbsolutePath() + " doesn't exist");
        }

        if (destination.exists()) {
            throw new IOException("Failed to create zip file, destination " + destination.getAbsolutePath() + " already exists");
        }

        ZipOutputStream zout = new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(destination)));
        try {
            ZipEntry entry = new ZipEntry(source.getName());
            zout.putNextEntry(entry);

            BufferedInputStream bin = new BufferedInputStream(new FileInputStream(source));
            try {
                copy(bin, zout);
            } finally {
                bin.close();
            }

            zout.closeEntry();
        } finally {
            zout.close();
        }
    }

    public void copy(File source, File destination) throws IOException {
        BufferedOutputStream bof = new BufferedOutputStream(new FileOutputStream(destination));
        try {
            BufferedInputStream bin = new BufferedInputStream(new FileInputStream(source));
            try {
                copy(bin, bof);
            } finally {
                bin.close();
            }
        } finally {
            bof.close();
        }
    }

    private void copy(InputStream is, OutputStream os) throws IOException {
        ByteStreams.copy(is, os);

        //Files.copy(source, os); // this reads entire source into memory!
    }

    public File consolidate(Date date) throws IOException {
        File result = null;

        //check if xmpp subfolder exists or create it
        if ((!parent.exists() && (!parent.mkdir()))) {
            throw new IOException("Destination " + parent.getAbsolutePath() + " is not a directory and can't be created");
        } else if (!parent.isDirectory()) {
            throw new IOException("Destination " + parent.getAbsolutePath() + " exists but is not a directory");
        }

        //delete all tmp files (they can be incomplete), zip the txt again
        File[] tmps = withType(FileFactory.TYPE.PART);
        LOG.log(Level.INFO, "Found {0} temporary archive files. Deleting all (if any) ... ", tmps.length);
        for (File tmp : tmps) {
            LOG.log(Level.INFO, "Deleting temporary archive file {0}", tmp.getAbsolutePath());
            if (!tmp.delete()) {
                LOG.log(Level.INFO, "Deleting temporary archive file {0} failed", tmp.getAbsolutePath());
            }
        }

        //check the pairs of .txt and .zip with the same filename
        Set<File> zips = new HashSet(Arrays.asList(withType(FileFactory.TYPE.ZIP)));
        List<File> txts = new ArrayList<File>(Arrays.asList(withType(FileFactory.TYPE.TXT))); //Arrays.asList does not support iterator.remove method

        if ((zips.size() > 0)) {
            LOG.log(Level.INFO, "Found {0} unpacked archive files. Deleting all already packed (if any) ... ", txts.size());
            Iterator<File> txtIt = txts.iterator();
            while (txtIt.hasNext()) {
                File txt = txtIt.next();
                File zip = withType(txt, FileFactory.TYPE.ZIP);
                if (zips.contains(zip)) {
                    LOG.log(Level.INFO, "Deleting already packed archive file {0}", txt.getAbsolutePath());
                    if (!txt.delete()) {
                        LOG.log(Level.INFO, "Deleting already packed archive file {0} failed", txt.getAbsolutePath());
                    }
                    //anyway dont consider it as an "active" file
                    txtIt.remove();
                }
            }
        }

        if (txts.size() > 1) {
            //in txts there are only txt files without zip pair, we sort them lexographically
            Collections.sort(txts);

            for (int i = (txts.size() - 1); i >= 0; i--) {
                if (i == (txts.size() - 1)) { //return only the youngest (with lexicographically highest name)
                    result = txts.get(i);
                } else {
                    try {
                        zip(txts.get(i));
                        LOG.log(Level.INFO, "Zipping old txt archive {0} successful", txts.get(i).getAbsolutePath());
                    } catch (IOException e) {
                        LOG.log(Level.INFO, "Zipping old txt archive failed", e);
                    }
                }
            }
        } else if (txts.size() == 1) {
            result = txts.get(0);
        }

        return result != null ? result : next(date);
    }

    public File next(Date date) throws IOException {

        File result = file(date);

        if (result.exists()) {
            throw new IOException("Failed to find unused filename,  " + file(new Date()));
        }

        return result;
    }

    public File withType(File file, TYPE type) {
        String name = removeSuffix(file.getName());
        return new File(this.parent, name + type.suffix);
    }

    public File withType(String source, TYPE type) {
        return withType(new File(this.parent, source), type);
    }

    private File file(Date messageDate) {
        return new File(this.parent, String.format(FILE_NAME_FORMAT, messageDate));
    }

    private String removeSuffix(String filename) {
        String result;
        int dotSeparator = filename.lastIndexOf(".");
        if ((dotSeparator > 0) && (dotSeparator < filename.length())) {
            result = filename.substring(0, dotSeparator);
        } else {
            result = filename;
        }

        return result;
    }

    private File[] withType(TYPE type) {

        File[] result = this.parent.listFiles(new SuffixFileFilter(type));
        if (result == null) {
            result = new File[0];
        }

        return result;
    }

    public String getDestination() {
        return this.parent.getAbsolutePath();
    }

}
