/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import com.kerio.im.core.conversation.archive.FileStoreWriter.SizeCheckException;
import java.io.File;
import java.io.IOException;
import java.time.LocalDateTime;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class FileStore {

    private static final Logger LOG = Logger.getLogger(FileStore.class.getName());

    public static FileStore open(File destination, long limit) throws IOException {
        FileFactory filestoreManager = new FileFactory(destination);
        FileStoreWriterFactory archiveStoreFactory = new FileStoreWriterFactory(filestoreManager, limit);
        return new FileStore(filestoreManager, archiveStoreFactory);
    }

    private final FileStoreWriterFactory fileWriterFactory;
    private final FileFactory fileManager;
    private FileStoreWriter fileWriter;

    FileStore(FileFactory fileFactory, FileStoreWriterFactory archiveStoreFactory) {
        this.fileWriterFactory = archiveStoreFactory;
        this.fileManager = fileFactory;
    }

    public void open() throws IOException {

        File existing = this.fileManager.consolidate(new Date());
        LOG.log(Level.INFO, "Consolidating archive folder structure successful");

        this.fileWriter = fileWriterFactory.create(existing);
        LOG.log(Level.INFO, "Creating file writer successful, destination {0}", this.fileWriter.file().getAbsolutePath());
    }

    public void store(LocalDateTime time, String raw) throws IOException {
        synchronized (this) {
            if (this.fileWriter != null) {
                try {
                    this.fileWriter.write(time, raw);
                    LOG.log(Level.FINEST, "Archiving raw message/event sucessful");
                    
                } catch (SizeCheckException e) {
                    try {
                        LOG.log(Level.INFO, "File limit reached, packing current archive storage {0}", this.fileWriter.file().getAbsolutePath());
                        this.fileWriter.close();
                        this.fileManager.zip(this.fileWriter.file());
                        LOG.log(Level.INFO, "Packing current archive storage successful");
                    } catch (IOException io) {
                        LOG.log(Level.INFO, "Packing current archive storage failed, leaving unpacked", io);
                    } finally {
                        this.fileWriter = fileWriterFactory.create();
                        LOG.log(Level.INFO, "Openning archive storage successful, new destination {0}", this.fileWriter.file().getAbsolutePath());

                        this.fileWriter.write(time, raw);
                        LOG.log(Level.FINEST, "Archiving raw message/event sucessful");
                    }
                }
            } else {
                throw new IOException("Archiving raw message/event failed, archive is closed");
            }
        }
    }

    public void export(String source, String path) throws IOException {
        synchronized (this) {
            File exportFile = new File(path);
            File exportFolder = exportFile.getParentFile();

            if (!exportFolder.isDirectory()) {
                throw new IOException("Exporting archive failed, destination folder doesn't exist");
            } else if (exportFile.exists()) {
                throw new IOException("Exporting archive failed, destination file already exists");
            }

            File txt = this.fileManager.withType(source, FileFactory.TYPE.TXT);
            File zip = this.fileManager.withType(source, FileFactory.TYPE.ZIP);
            if (txt.exists()) {
                this.fileManager.zip(txt, exportFile);
                LOG.log(Level.INFO, "Packing {0} to {1} successful", new Object[]{source, path});
            } else if (zip.exists()) {
                this.fileManager.copy(zip, exportFile);
                LOG.log(Level.INFO, "Copying {0} to {1} successful", new Object[]{source, path});
            } else {
                throw new IOException("Exporting archive failed, source doesn't exist");
            }
        }
    }

    public void close() {
        synchronized (this) {
            if (this.fileWriter != null) {
                try {
                    this.fileWriter.close();
                    LOG.log(Level.INFO, "Closing file writer successful");
                } finally {
                    this.fileWriter = null;
                }
            }
        }
    }

    public String getDestination() {
        return this.fileManager.getDestination();
    }

}
