/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.remote;

import com.kerio.im.core.ErrorHandler;
import com.kerio.im.core.api.service.MessageService;
import com.kerio.im.core.api.service.ServiceBus;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.json.JSONException;
import org.json.JSONObject;

/**
 *
 * @author tskoda
 */
public class RemoteControl implements MessageService {

    /**
     * Package private protocol messages.
     */
    enum PROTOCOL {

        result,
        reason;

        boolean is(String command) {
            return this.toString().equalsIgnoreCase(command);
        }
    }
    private static final Logger LOG = Logger.getLogger(RemoteControl.class.getName());
    private final ServiceBus bus;
    private final ErrorHandler errorHandler;

    public RemoteControl(final ServiceBus bus, final ErrorHandler errorHandler) {
        this.bus = bus;
        this.errorHandler = errorHandler;
    }

    @Override
    public void messageReceived(String message, ResponseHandler responseHandler) {

        try {
            //Map<String, String> result = ;
            CommandEvent event = createCommandEvent(message, new HashMap<String, String>(), responseHandler);
            LOG.log(Level.INFO, "Received message '{0}' with id '{1}'", new Object[]{event.getType(), event.getId()});
            this.bus.publish(event); //sequential, returns after all listeners have been notified

            if (! event.isAsync()) {
                answer(event);
            }

        } catch (Exception e) {
            //TODO error code
            LOG.log(Level.WARNING, "Received message processing failed", e);
            //this.errorHandler.error(Code.COMMAND_ERROR);
        }

    }

    public void answer(CommandEvent event) {
        
        try {
            String response = createResponse(event, event.getResponse());
            event.getResponseHandler().response(response);
            LOG.log(Level.FINE, "Received message with id '{0}' processed successfuly", event.getId());
        } catch (Exception e) {
            LOG.log(Level.WARNING, "Received message processing failed", e);
        }
    }

    protected CommandEvent createCommandEvent(String message, Map<String, String> result, ResponseHandler responseHandler) throws JSONException {
        JSONObject jsonMessage = new JSONObject(message);
        String command = jsonMessage.getString("op");
        String commandId = jsonMessage.getString("opId");
        Map<String, String> data = new HashMap<String, String>();

        JSONObject jsonData = jsonMessage.getJSONObject("param");
        String key;
        for (Iterator<String> it = jsonData.keys(); it.hasNext();) {
            key = it.next();
            data.put(key, jsonData.getString(key));
        }

        return new CommandEvent(commandId, command, data, result, responseHandler);
    }

    protected String createResponse(CommandEvent event, Map<String, String> content) throws JSONException {
        JSONObject jsonMessage = new JSONObject();
        jsonMessage.put("op", event.getType());
        jsonMessage.put("opId", event.getId());

        JSONObject jsonContent = new JSONObject();
        if (event.isFail()) {
            jsonContent.put(PROTOCOL.result.toString(), "failed");
            jsonContent.put(PROTOCOL.reason.toString(), event.getFailMessage());
        } else {
            jsonContent.put(PROTOCOL.result.toString(), "success");
        }

        for (Map.Entry<String, String> pair : content.entrySet()) {
            jsonContent.put(pair.getKey(), pair.getValue());
        }

        jsonMessage.put("param", jsonContent);
        return jsonMessage.toString();
    }
}
