/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.user;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.core.api.service.AccessPolicyService;
import com.kerio.im.core.api.service.PacketType;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.configuration.StoreConfiguration;
import com.kerio.im.core.remote.CommandEvent;
import com.kerio.im.external.ExtService;
import com.kerio.im.xmpp.api.XmppConstants;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class UserAccessPolicy implements AccessPolicyService {

    private static final Logger LOG = Logger.getLogger(UserAccessPolicy.class.getName());
    private static final Boolean DENY = Boolean.FALSE;
    private static final Boolean ALLOW = Boolean.TRUE;
    private static final String DENY_INCOMING = "Message delivery failed: recipient can receive emails from his/her own domain only";
    private static final String DENY_OUTGOING = "Message delivery failed: sending messages outside your domain is disabled for your account";

    enum State {

        UNKNOWN,
        ENABLED,
        DISABLED
    }
    private final String controlUser;
    private final String controlDomain;
    private Users users;
    private State state = State.UNKNOWN;

    public UserAccessPolicy(StoreConfiguration configuration) {
        this.state = State.UNKNOWN;
        this.controlDomain = configuration.get(GlobalKey.SERVICE_REMOTE_DOMAIN);
        this.controlUser = configuration.get(GlobalKey.SERVICE_REMOTE_LOGIN);
    }

    void setState(State state) {
        this.state = state;
    }
    
    State getState() {
        return this.state;
    }

    public void setExtService(Users users, ExtService extService, ServiceBus bus) {
        this.users = users;
        bus.subscribe(CommandEvent.class, new EventFederation(this));
        setState(extService.getSettings().isSendOutsideEnabled() ? State.ENABLED : State.DISABLED);
    }

    @Override
    public boolean incoming(String from, String fromDomain, String to, String toDomain, PacketType type) {
        boolean result = filter(from, fromDomain, to, toDomain);

        if (result == DENY) {
            switch (type) {
                case MESSAGE:
                    LOG.log(Level.WARNING, "Delivery not authorized (instant message from <{0}@{1}> to <{2}@{3}>: recipient can receive instant messages from own domain only)", new Object[]{from, fromDomain, to, toDomain});
                default:
                    LOG.log(Level.FINER, "Blocking incoming packet {4} from {0}@{1} to {2}@{3}", new Object[]{from, fromDomain, to, toDomain, type});
            }
        }
        return result;
    }

    @Override
    public boolean outgoing(String to, String toDomain, String from, String fromDomain, PacketType type) {
        boolean result = filter(to, toDomain, from, fromDomain);

        if (result == DENY) {
            switch (type) {
                case MESSAGE:
                    LOG.log(Level.WARNING, "Delivery not authorized (instant message from <{0}@{1}> to <{2}@{3}>: sender can send instant messages to own domain only)", new Object[]{from, fromDomain, to, toDomain});
                default:
                    LOG.log(Level.FINER, "Blocking outgoing packet {4} to {0}@{1} from {2}@{3}", new Object[]{to, toDomain, from, fromDomain, type});
            }
        }

        return result;
    }
    
    private boolean filter(String user1, String domain1, String user2, String domain2) {
        boolean result = ALLOW; //e.g. presence packet change could have no to - we allow all incomplete connection

        if ((domain1 != null) && (user2 != null) && (domain2 != null)) { //we must allow packet without fromDomain set
            if (isSameDomain(XmppConstants.GROUP_CHAT_NAME, domain1, domain2)) {
                //always allow communication inside the same domain
                result = allowEqualDomains();
            } else {
                //it depends on "federation" setting and user rights
                result = allowDifferentDomains(user2.toLowerCase(), domain2.toLowerCase());
            }
        }
        
        return result;
    }
    
    /**
     * Returns true only if domain1 and domain2 are case insensitive equal, 
     * or domain1 is equal to domain2 with prefix
     * or domain2 is equal to domain1 with prefix
     * @param domain1
     * @param domain2
     * @return 
     */
    boolean isSameDomain(String prefix, String domain1, String domain2) {
        
        return domain1.equalsIgnoreCase(domain2) 
                || ((prefix.length() < domain1.length()) && (domain1.substring(prefix.length() + 1).equalsIgnoreCase(domain2)))
                || ((prefix.length() < domain2.length()) && (domain2.substring(prefix.length() + 1).equalsIgnoreCase(domain1)));
        
    }

    private boolean allowEqualDomains() {
        return ALLOW;
    }

    private boolean allowDifferentDomains(String user, String userDomain) {
        boolean result;
        switch (this.state) {
            case ENABLED:
                if (getUserAccess(user, userDomain)) {
                    result = ALLOW;
                } else {
                    result = DENY;
                }
                break;
            case DISABLED:
                //TODO log?
                result = DENY;
                break;
            case UNKNOWN: //allow only control user
                if (isControlUser(user, userDomain)) {
                    result = ALLOW;
                } else {
                    result = DENY;
                    //TODO log?
                }
                break;
            default:
                //TODO log?
                result = DENY;
        }

        return result;
    }

    private boolean isControlUser(String login, String domain) {
        return this.controlUser.equalsIgnoreCase(login) && this.controlDomain.equalsIgnoreCase(domain);
    }

    private boolean getUserAccess(String login, String domain) {
        boolean result;
        UserControl control = this.users.getControl(domain);
        if (control == null) { //this is not locally managed domain, allowing packet
            result = ALLOW;
        } else { //locally managed domain, deny for unknown user
            User user = control.find(login);
            if (user != null) {
                result = user.isAllowedSendOutside();
            } else {
                result = DENY;
            }
        }

        return result;
    }

    @Override
    public String incomingResponse() {
        return DENY_INCOMING;
    }

    @Override
    public String outgoingResponse() {
        return DENY_OUTGOING;
    }
}
