/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.user;

import com.kerio.im.bean.Domain;
import com.kerio.im.bean.User;
import com.kerio.im.common.CollectionUtil;
import com.kerio.im.common.Function;
import com.kerio.im.common.error.Code;
import com.kerio.im.xmpp.api.XmppException;
import com.kerio.im.xmpp.api.XmppUsers;
import java.util.Collection;
import java.util.EnumSet;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class UserControl {

    private static final Logger LOG = Logger.getLogger(UserControl.class.getName());
    private final GroupProvider groups;
    private final UserCache userCache;
    private final XmppUsers xmppUsers;
    private final Domain domain;

    public UserControl(Domain domain, XmppUsers xmppUsers, UserCache userCache) {
        this.xmppUsers = xmppUsers;
        this.userCache = userCache;
        this.domain = domain;
        this.groups = new GroupProvider();
    }

    User find(String login) {
        return this.userCache.find(login);
    }

    public void bye() {
        synchronized (this) {
            this.xmppUsers.bye();
        }
    }

    public Domain getDomain() {
        return domain;
    }

    public void add(User user) {
        synchronized (this) {
            if (find(user.getLogin()) == null) {
                try {
                    new UserAdded(this.xmppUsers, this.userCache).ex(user);
                } catch (Exception e) {
                    throw new XmppException("Failed to add user to domain {0}", e, Code.RESOURCE_ERROR, this.domain.getName());
                }
            }
        }
    }

    public void remove(String login, String domain) {
        synchronized (this) {
            if (find(login) != null) {
                try {
                    new UserRemoved(xmppUsers, this.userCache).ex(new User(login, domain, null, false, false, false));
                } catch (Exception e) {
                    throw new XmppException("Failed to add user to domain {0}", e, Code.RESOURCE_ERROR, this.domain.getName());
                }
            }
        }
    }

    public void update(String login, String domain, String fullname, Boolean enabled, Boolean visible, Boolean allowedSendOutside) {
        synchronized (this) {
            User existing = find(login);
            if (existing != null) {
                String newFullName = fullname != null ? fullname : existing.getFullName();
                Boolean newEnabled = enabled != null ? enabled : existing.isEnabled();
                Boolean newVisible = visible != null ? visible : existing.isVisible();
                Boolean newAllowed = allowedSendOutside != null ? allowedSendOutside : existing.isAllowedSendOutside();
                User updated = new User(login, domain, newFullName, newEnabled, newVisible, newAllowed);
                this.userCache.addOrUpdate(updated);
                try {
                    UserMerge userMerge = new UserMerge(xmppUsers);
                    userMerge.merge(updated, existing);
                } catch (Exception e) {
                    throw new XmppException("Failed to update user {0}", e, Code.RESOURCE_ERROR, existing.getPrimaryEmail());
                }

            }
        }
    }

    public void sync() {
        synchronized (this) {
            this.userCache.refresh();
            try {
                //compare currently known users (fresh list) with those existing in xmpp
                final Collection<User> extUsers = this.userCache.all();
                final Map<String, User> xmppUsersMap = this.xmppUsers.all();
                CollectionUtil.sync(
                        extUsers,
                        xmppUsersMap.values(),
                        new UserAdded(xmppUsers, userCache),
                        new UserRemoved(xmppUsers, userCache),
                        new UserPaired(new UserMerge(xmppUsers), xmppUsersMap));
            } catch (Exception e) { //TODO all methods throw only XMPPException?
                throw new XmppException("Failed to merge users in domain {0}", e, Code.RESOURCE_ERROR, this.domain.getName());
            }
        }
    }

    /**
     * This method check if user exist in database. If not it perform its
     * initial full synchronization. It puts all existing known users into
     * his/her roster.
     *
     * @param userName
     */
    public boolean check(final String userName) {
        synchronized (this) {
            return this.xmppUsers.check(this.userCache.find(userName));
        }
    }

    public void repair() {
        synchronized (this) {
            try {
                for (User user : this.userCache.all()) {
                    LOG.log(Level.FINER, "Repairing user {0}", user.getPrimaryEmail());
                    xmppUsers.repair(user);
                }
            } catch (Exception e) {
                throw new XmppException("Failed to repair users in domain {0}", e, Code.RESOURCE_ERROR, this.domain.getName());
            }
        }
    }

    public void repair(String userName, String domain) {
        synchronized (this) {
            User user = find(userName);
            if (user != null) {
                try {
                    xmppUsers.repair(user);
                } catch (Exception e) {
                    throw new XmppException("Failed to repair users in domain {0}", e, Code.RESOURCE_ERROR, this.domain.getName());
                }
            } else {
                LOG.log(Level.INFO, "Failed to repair user {0}@{1}, user not found", new Object[]{userName, domain});
            }
        }
    }

    static class UserAdded implements Function<User> {

        private final XmppUsers xmppUsers;
        private final UserCache userStore;

        public UserAdded(XmppUsers xmppUsers, UserCache userStore) {
            this.xmppUsers = xmppUsers;
            this.userStore = userStore;
        }

        @Override
        public void ex(User item) {
            LOG.log(Level.FINER, "Adding user {0}@{1}", new Object[]{item.getLogin(), item.getDomain()});
            this.userStore.addOrUpdate(item);
            this.xmppUsers.add(item);
        }
    }

    static class UserRemoved implements Function<User> {

        private final XmppUsers xmppUsers;
        private final UserCache userStore;

        public UserRemoved(XmppUsers xmppUsers, UserCache userStore) {
            this.xmppUsers = xmppUsers;
            this.userStore = userStore;
        }

        @Override
        public void ex(User item) {
            LOG.log(Level.FINER, "Removing user {0}@{1}", new Object[]{item.getLogin(), item.getDomain()});
            this.userStore.remove(item);
            this.xmppUsers.remove(item);
        }
    }

    static class UserPaired implements Function<User> {

        private final UserMerge merge;
        private final Map<String, User> xmppUsersMap;

        public UserPaired(UserMerge merge, Map<String, User> xmppUsersMap) {
            this.merge = merge;
            this.xmppUsersMap = xmppUsersMap;
        }

        @Override
        public void ex(User item) {
            this.merge.merge(item, xmppUsersMap.get(item.getLogin()));
        }
    }

    static class UserMerge {

        private final XmppUsers xmppUsers;

        public UserMerge(XmppUsers xmppUsers) {
            this.xmppUsers = xmppUsers;
        }

        public void merge(User storeUser, User xmppUser) {
            EnumSet<XmppUsers.Changes> changes = EnumSet.noneOf(XmppUsers.Changes.class);
            renamed(storeUser, xmppUser, changes);
            statusChanged(storeUser, xmppUser, changes);
            galChanged(storeUser, xmppUser, changes);

            if (!changes.isEmpty()) {
                LOG.log(Level.FINER, "Updating user {0}@{1}", new Object[]{storeUser.getLogin(), storeUser.getDomain()});
                this.xmppUsers.update(storeUser, changes);
            }
        }

        private void renamed(User storeUser, User xmppUser, EnumSet<XmppUsers.Changes> changes) {

            if ((storeUser != null) && (xmppUser != null)) {
                if ((storeUser.getFullName() == null) && (xmppUser.getFullName() != null)) {
                    changes.add(XmppUsers.Changes.NAME);
                } else if ((storeUser.getFullName() != null) && (!storeUser.getFullName().equals(xmppUser.getFullName()))) {
                    changes.add(XmppUsers.Changes.NAME);
                }
            }
        }

        private void statusChanged(User storeUser, User xmppUser, EnumSet<XmppUsers.Changes> changes) {
            if ((storeUser != null) && (xmppUser != null)) {
                if (storeUser.isEnabled() != xmppUser.isEnabled()) {
                    changes.add(XmppUsers.Changes.STATUS);
                }
            }
        }

        private void galChanged(User storeUser, User xmppUser, EnumSet<XmppUsers.Changes> changes) {
            if ((storeUser != null) && (xmppUser != null)) {
                if (storeUser.isVisible() != xmppUser.isVisible()) {
                    changes.add(XmppUsers.Changes.VISIBILITY);
                }
            }
        }
    }
}
