/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.user;

import com.kerio.im.bean.Domain;
import com.kerio.im.external.ExtService;
import com.kerio.im.xmpp.api.XmppService;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class UserControlList {

    private static final Logger LOG = Logger.getLogger(UserControlList.class.getName());
    private final XmppService xmppService;
    private final ExtService extService;
    private final Map<String, UserControl> userControlMap;

    public UserControlList(XmppService xmppService, ExtService extService) {
        this.xmppService = xmppService;
        this.extService = extService;
        this.userControlMap = new HashMap<String, UserControl>();
    }

    public void remove(Domain domain) {
        synchronized (this) {
            LOG.log(Level.FINE, "Removing domain {0} from synchronization", domain.getName());
            if (this.userControlMap.containsKey(domain.getName())) {
                UserControl removed = this.userControlMap.remove(domain.getName());
                if (removed != null) {
                    removed.bye();
                    LOG.log(Level.INFO, "Removing domain {0} from synchronization successful", domain.getName());
                }
            } else {
                LOG.log(Level.INFO, "Removing domain {0} from synchronization failed, domain not found", domain.getName());
            }
        }
    }

    public void create(Domain domain) {

        UserControl userControl = null;
        synchronized (this) {
            LOG.log(Level.FINE, "Adding domain {0} for synchronization", domain.getName());
            if (!this.userControlMap.containsKey(domain.getName())) {
                userControl = new UserControl(domain, this.xmppService.users(domain.getName()), new UserCache(domain, this.extService));
                this.userControlMap.put(domain.getName(), userControl);
            }
        }
        try {
            if (userControl != null) {
                userControl.sync();
                LOG.log(Level.INFO, "Adding domain {0} for synchronization successful", domain.getName());
            } else {
                LOG.log(Level.INFO, "Adding domain {0} for synchronization failed, domain already exist", domain.getName());
            }
        } catch (Exception e) {
            LOG.log(Level.INFO, "Adding domain " + domain.getName() + " for synchronization failed", e);
            LOG.log(Level.WARNING, "Adding domain {0} for synchronization failed, {1}", new Object[]{domain.getName(), e.getMessage()});
        }
    }

    UserControl get(String domain) {
        synchronized (this) {
            return this.userControlMap.get(domain);
        }
    }

    public void sync(Domain domain) {
        synchronized (this) {
            LOG.log(Level.FINE, "Syncing all users from domain {0}", domain.getName());

            try {
                UserControl control = get(domain.getName());
                if (control != null) {
                    control.sync();
                    LOG.log(Level.INFO, "Syncing all users from domain {0} successful", domain.getName());
                } else {
                    LOG.log(Level.INFO, "Syncing all users from domain {0} failed, domain not in sync", domain.getName());
                }
            } catch (Exception e) {
                LOG.log(Level.INFO, "Syncing all users from domain " + domain.getName() + " failed",  e);
                LOG.log(Level.WARNING, "Syncing all users from domain {0} failed, {1}", new Object[]{domain.getName(), e.getMessage()});
            }
            
        }
    }

    public void repair(Domain domain) {
        synchronized (this) {
            LOG.log(Level.FINE, "Repairing all users from domain {0}", domain.getName());

            try {
                UserControl control = get(domain.getName());
                if (control != null) {
                    control.repair();
                    LOG.log(Level.INFO, "Repairing all users from domain {0} successful", domain.getName());
                } else {
                    LOG.log(Level.INFO, "Repairing all users from domain {0} failed, domain not in sync", domain.getName());
                }
            } catch (Exception e) {
                LOG.log(Level.WARNING, "Repairing all users from domain failed", e);
                LOG.log(Level.WARNING, "Repairing all users from domain {0} failed, {1}", new Object[]{domain.getName(), e.getMessage()});
            }
            
        }
    }
}
