/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.vcard;

import com.kerio.im.bean.VCard;
import com.kerio.im.bean.VCard.VCardFactory;
import com.kerio.im.external.ExtService;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author tskoda
 */
public class VCardCache implements ExtService.VCardCache {

    private static final Logger LOG = Logger.getLogger(VCardCache.class.getName());
    /**
     * User in form of primary user email address.
     */
    private final Map<String, VCard> userToVCard;
    /**
     * Public contact identifier to user in form of primary user email address.
     */
    private final Map<String, String> idToUser;
    /**
     * Domain for users.
     */
    private final String domain;
    /**
     * Last sync key. Marks changes in public contact folder.
     */
    private long syncKey;
    /**
     * Source for vcard cache. Is managed by external service. Could be null.
     */
    private String source;

    public VCardCache(String domain) {
        this.domain = domain;
        this.userToVCard = new HashMap<String, VCard>();
        this.idToUser = new HashMap<String, String>();
    }

    public void refresh(ExtService extService, VCardFactory factory) {
        synchronized (this) {
            this.syncKey = extService.getAllVCards(this.syncKey, this.domain, this, factory);
        }
    }

    public VCard get(String login) {
        synchronized (this) {
            VCard userVCard = this.userToVCard.get(getPrimaryEmail(login));
            if (userVCard != null) {
                LOG.log(Level.FINEST, "Getting vcard for user {0}@{1} successful", new Object[]{login, this.domain});
            } else {
                LOG.log(Level.INFO, "Getting vcard for user {0}@{1} failed, user not in cache", new Object[]{login, this.domain});
            }

            return userVCard;
        }
    }

    private String getPrimaryEmail(String login) {
        return login + "@" + this.domain;
    }

    @Override
    public boolean containsId(String itemId) {
        return (this.idToUser.containsKey(itemId) && this.userToVCard.containsKey(this.idToUser.get(itemId)));
    }

    @Override
    public VCard getById(String itemId) {
        return this.userToVCard.get(this.idToUser.get(itemId));

    }

    @Override
    public void clear() {
        this.userToVCard.clear();
        this.idToUser.clear();
    }

    @Override
    public void removeById(String itemId) {
        String user = this.idToUser.remove(itemId);
        if (user != null) {
            this.userToVCard.remove(user);
        }
    }

    @Override
    public int count() {
        return this.userToVCard.size();
    }

    @Override
    public void put(VCard vcard, String itemId, String username) {
        if ((vcard != null) && (itemId != null) && (username != null)) {
            String lowerCaseUserName = username.toLowerCase();
            this.idToUser.put(itemId, lowerCaseUserName);
            this.userToVCard.put(lowerCaseUserName, vcard);
        }
    }

    @Override
    public void setSource(String source) {
        this.source = source;
    }

    @Override
    public String getSource() {
        return source;
    }
}
