/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation;

import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import static org.hamcrest.CoreMatchers.*;

/**
 *
 * @author tskoda
 */
public class DerbyConnectionFactoryTest {

    private DerbyDestination mockDestination;
    private DerbyConnectionFactory.DerbyConnectionFactoryImpl mockFactoryImpl;
    private Connection mockConnection;

    @BeforeClass
    public static void setUpClass() {
    }

    @AfterClass
    public static void tearDownClass() {
    }

    @Before
    public void setUp() throws SQLException, IOException {
        this.mockDestination = mock(DerbyDestination.class);
        when(this.mockDestination.toPath()).thenReturn("mock-test-db");
        this.mockFactoryImpl = mock(DerbyConnectionFactory.DerbyConnectionFactoryImpl.class);
        this.mockConnection = mock(Connection.class);
    }

    @After
    public void tearDown() {
    }

    @Test(expected = IOException.class)
    public void testOpenExistFailToRepair() throws Exception {
        when(this.mockDestination.exists()).thenReturn(Boolean.TRUE); //destination exists
        when(this.mockFactoryImpl.newConnection(anyString())).thenThrow(new SQLException()); //destination is not consistent
        doThrow(new IOException()).when(this.mockDestination).repair(); //destination can not be repaired

        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);
        try {
            cf.open();
        } finally {
            verify(this.mockFactoryImpl, times(1)).newConnection(anyString());
            verify(this.mockDestination, times(1)).repair();
        }
    }

    @Test(expected = SQLException.class)
    public void testOpenExistFailAfterSuccessRepair() throws Exception {
        when(this.mockDestination.exists()).thenReturn(Boolean.TRUE); //destination exists
        when(this.mockFactoryImpl.newConnection(anyString())).thenThrow(new SQLException()); //destination is not consistent, even after repair

        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);
        try {
            cf.open();
        } finally {
            verify(this.mockFactoryImpl, times(2)).newConnection(anyString());
            verify(this.mockDestination, times(1)).repair();
        }
    }

    @Test
    public void testOpenExistPassAfterSuccessRepair() throws Exception {
        when(mockDestination.exists()).thenReturn(Boolean.TRUE);
        when(this.mockFactoryImpl.newConnection(anyString())).thenThrow(new SQLException()).thenReturn(mockConnection); //destination is not consistent, repair is successful
        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);

        assertThat(cf.open(), is(true)); //db was created
        verify(this.mockFactoryImpl, times(2)).newConnection(anyString());
        verify(this.mockDestination, times(1)).repair();

    }

    @Test
    public void testOpenExist() throws Exception {
        when(mockDestination.exists()).thenReturn(Boolean.TRUE);
        when(this.mockFactoryImpl.newConnection(anyString())).thenReturn(this.mockConnection);
        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);

        assertThat(cf.open(), is(false)); //db was NOT created
        verify(this.mockFactoryImpl, times(1)).newConnection(anyString());
        verify(this.mockDestination, times(0)).repair();
    }

    @Test(expected = IOException.class)
    public void testOpenNotExistFailToRepair() throws Exception {
        when(this.mockDestination.exists()).thenReturn(Boolean.FALSE); //destination exists
        when(this.mockFactoryImpl.newConnection(anyString())).thenThrow(new SQLException()); //destination is not consistent
        doThrow(new IOException()).when(this.mockDestination).repair(); //destination can not be repaired

        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);
        try {
            cf.open();
        } finally {
            verify(this.mockFactoryImpl, times(1)).newConnection(anyString());
            verify(this.mockDestination, times(1)).repair();
        }
    }

    @Test(expected = SQLException.class)
    public void testOpenNotExistFailAfterSuccessRepair() throws Exception {
        when(this.mockDestination.exists()).thenReturn(Boolean.FALSE); //destination exists
        when(this.mockFactoryImpl.newConnection(anyString())).thenThrow(new SQLException()); //destination is not consistent, even after repair

        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);
        try {
            cf.open();
        } finally {
            verify(this.mockFactoryImpl, times(2)).newConnection(anyString());
            verify(this.mockDestination, times(1)).repair();
        }
    }

    @Test
    public void testOpenNotExistPassAfterSuccessRepair() throws Exception {
        when(mockDestination.exists()).thenReturn(Boolean.FALSE);
        when(this.mockFactoryImpl.newConnection(anyString())).thenThrow(new SQLException()).thenReturn(mockConnection); //destination is not consistent, repair is successful
        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);

        assertThat(cf.open(), is(true)); //db was created
        verify(this.mockFactoryImpl, times(2)).newConnection(anyString());
        verify(this.mockDestination, times(1)).repair();

    }

    @Test
    public void testOpenNotExist() throws Exception {
        when(mockDestination.exists()).thenReturn(Boolean.FALSE);
        when(this.mockFactoryImpl.newConnection(anyString())).thenReturn(this.mockConnection);
        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);

        assertThat(cf.open(), is(true)); //db was created
        verify(this.mockFactoryImpl, times(1)).newConnection(anyString());
        verify(this.mockDestination, times(0)).repair();
    }
    
    public void testNewConnection() throws SQLException {
        when(this.mockFactoryImpl.newConnection(anyString())).thenReturn(this.mockConnection);
        DerbyConnectionFactory cf = new DerbyConnectionFactory(mockDestination, mockFactoryImpl);
         assertThat(cf.newConnection(), is(sameInstance(this.mockConnection))); //db was created
    }

}
