/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import com.kerio.im.common.error.Code;
import com.kerio.im.core.api.event.ShutdownEvent;
import com.kerio.im.core.conversation.ConversationServiceImpl;
import com.kerio.im.core.remote.CommandEvent;
import com.kerio.im.external.ExtService;
import java.io.IOException;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.mockito.Mockito.*;

/**
 *
 * @author tskoda
 */
public class ArchiveTest {

    private ConversationServiceImpl mockConversationRegister;
    //private ExtService mockExtService;
    private FileStore mockArchiveStore;

    public ArchiveTest() {
    }

    @Before
    public void setUp() {
        this.mockConversationRegister = mock(ConversationServiceImpl.class);
        // this.mockExtService = mock(ExtService.class);
        this.mockArchiveStore = mock(FileStore.class);
    }

    @After
    public void tearDown() {

    }

    private ExtService.ArchiveSettings settings(final boolean state) {
        return new ExtService.ArchiveSettings() {

            @Override
            public String getPath() {
                return "path";
            }

            @Override
            public boolean isEnabled() {
                return state;
            }
        };
    }


    @Test
    public void testRefreshOff() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive spyArchive = spy(archive);

        spyArchive.refresh(false);

        verify(spyArchive, never()).close();
        verify(mockArchiveStore, never()).open();
    }

    @Test
    public void testRefreshOn() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(true);

        verify(mockConversationRegister, never()).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, never()).close();
        verify(mockArchiveStore).open();
        verify(mockConversationRegister).register(eq(Archive.CONVERSATION_LISTENER_NAME), any(AsyncFileStore.class));
    }

    @Test
    public void testRefreshOffNoChange() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(false);
        archive.refresh(false);

        verify(mockConversationRegister, never()).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, never()).close();
        verify(mockArchiveStore, never()).open();
        verify(mockConversationRegister, never()).register(eq(Archive.CONVERSATION_LISTENER_NAME), any(AsyncFileStore.class));

    }

    @Test
    public void testRefreshOnNoChange() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(true);
        archive.refresh(true);

        verify(mockConversationRegister, never()).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, never()).close();
        verify(mockArchiveStore, times(1)).open();
        verify(mockConversationRegister, times(1)).register(eq(Archive.CONVERSATION_LISTENER_NAME), any(AsyncFileStore.class));

    }

    @Test
    public void testRefreshChange() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(true);
        verify(mockConversationRegister, never()).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, never()).close();
        verify(mockArchiveStore).open();
        verify(mockConversationRegister).register(eq(Archive.CONVERSATION_LISTENER_NAME), any(AsyncFileStore.class));

        archive.refresh(false);
        verify(mockConversationRegister, times(1)).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, times(1)).close();

    }

    @Test
    public void testExportOff() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(true);

        archive.export("source", "destination");
        verify(mockArchiveStore).export(eq("source"), eq("destination"));
    }

    @Test
    public void testExportOn() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(true);

        archive.export("source", "destination");
        verify(mockArchiveStore).export(eq("source"), eq("destination"));

    }

    @Test
    public void testCloseOnFirst() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        archive.close();

        verify(mockConversationRegister, times(1)).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, times(1)).close();
    }

    @Test
    public void testCloseOff() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(false);
        archive.close();

        verify(mockConversationRegister, times(1)).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, times(1)).close();
    }

    @Test
    public void testCloseOn() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);

        archive.refresh(true);
        archive.close();

        verify(mockConversationRegister, times(1)).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, times(1)).close();

    }

    @Test
    public void testCloseRegistered() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        
        archive.refresh(true);

        ConversationServiceImpl.AsyncConversationService mockPreviously = mock(ConversationServiceImpl.AsyncConversationService.class);
        when(mockConversationRegister.unregister(Archive.CONVERSATION_LISTENER_NAME)).thenReturn(mockPreviously);

        archive.close();

        verify(mockPreviously).close();
        verify(mockConversationRegister, times(1)).unregister(Archive.CONVERSATION_LISTENER_NAME);
        verify(mockArchiveStore, times(1)).close();

    }

    @Test
    public void testCommandUnknown() {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive.ArchiveEventListener listener = archive.new ArchiveEventListener(mock(ExtService.class));

        CommandEvent mockCommand = mock(CommandEvent.class);
        when(mockCommand.interested(any(String.class))).thenReturn(false);

        listener.event(mockCommand);
        verify(mockCommand).interested("archive");
        verifyNoMoreInteractions(mockCommand);
    }

    @Test
    public void testCommandEvent() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive spyArchive = spy(archive);
        ExtService mockExtService = mock(ExtService.class);
        Archive.ArchiveEventListener listener = spyArchive.new ArchiveEventListener(mockExtService);

        when(mockExtService.getArchiveSettings()).thenReturn(settings(true));

        CommandEvent mockCommand = mock(CommandEvent.class);
        when(mockCommand.interested(any(String.class))).thenReturn(true);
        when(mockCommand.get("event")).thenReturn("changed");

        listener.event(mockCommand);

        verify(spyArchive).refresh(eq(true));
    }

    @Test
    public void testCommandEventUnknown() {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive spyArchive = spy(archive);

        ExtService mockExtService = mock(ExtService.class);
        Archive.ArchiveEventListener listener = spyArchive.new ArchiveEventListener(mockExtService);

        when(mockExtService.getArchiveSettings()).thenReturn(settings(true));

        CommandEvent mockCommand = mock(CommandEvent.class);
        when(mockCommand.interested(any(String.class))).thenReturn(true);
        when(mockCommand.get("event")).thenReturn("unknwon");

        listener.event(mockCommand);

        verifyNoMoreInteractions(spyArchive);
        verify(mockCommand).fail(any(String.class));
    }

    @Test
    public void testCommandExport() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive spyArchive = spy(archive);

        ExtService mockExtService = mock(ExtService.class);
        Archive.ArchiveEventListener listener = spyArchive.new ArchiveEventListener(mockExtService);

        CommandEvent mockCommand = mock(CommandEvent.class);
        when(mockCommand.interested(any(String.class))).thenReturn(true);
        when(mockCommand.get("event")).thenReturn("export");
        when(mockCommand.get("export.source")).thenReturn("source");
        when(mockCommand.get("export.destination")).thenReturn("destination");

        listener.event(mockCommand);

        verify(spyArchive).export("source", "destination");

    }

    @Test
    public void testCommandNoSource() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive spyArchive = spy(archive);

        ExtService mockExtService = mock(ExtService.class);
        Archive.ArchiveEventListener listener = spyArchive.new ArchiveEventListener(mockExtService);

        CommandEvent mockCommand = mock(CommandEvent.class);
        when(mockCommand.interested(any(String.class))).thenReturn(true);
        when(mockCommand.get("event")).thenReturn("export");
        // let it be null when(mockCommand.get("export.source")).thenReturn("source");
        when(mockCommand.get("export.destination")).thenReturn("destination");

        listener.event(mockCommand);

        verify(spyArchive, never()).close();
        verify(spyArchive, never()).refresh(anyBoolean());
        verify(mockCommand).fail(any(String.class));

    }

    @Test
    public void testCommandNoDest() throws IOException {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive spyArchive = spy(archive);
        Archive.ArchiveEventListener listener = spyArchive.new ArchiveEventListener(mock(ExtService.class));

        CommandEvent mockCommand = mock(CommandEvent.class);
        when(mockCommand.interested(any(String.class))).thenReturn(true);
        when(mockCommand.get("event")).thenReturn("export");
        when(mockCommand.get("export.source")).thenReturn("source");
        // let it be null when(mockCommand.get("export.destination")).thenReturn("destination");

        listener.event(mockCommand);

        verify(spyArchive, never()).close();
        verify(spyArchive, never()).refresh(anyBoolean());
        verify(mockCommand).fail(any(String.class));

    }

    @Test
    public void testCommandShutdown() {
        Archive archive = new Archive(this.mockConversationRegister, this.mockArchiveStore);
        Archive spyArchive = spy(archive);
        Archive.ArchiveShutdownListener listener = spyArchive.new ArchiveShutdownListener();

        listener.event(new ShutdownEvent(Code.OK));

        verify(spyArchive).close();
    }

}
