/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import com.google.common.io.ByteStreams;
import com.kerio.im.core.conversation.TempFolder;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Date;
import java.util.Enumeration;
import java.util.concurrent.TimeUnit;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;

/**
 *
 * @author tskoda
 */
public class FileStoreManagerTest {

    private static final byte[] TEST_MESSAGE = "<message type=\"chat\" to=\"adium@tskoda-macbook.kerio.local\" id=\"aabea\"><body>testovaci zprava</body><active xmlns=\"http://jabber.org/protocol/chatstates\"/></message>".getBytes();
    private static final byte[] TEST_FILE = "<iq type=\"set\" to=\"ichat@tskoda-macbook.kerio.local/tskoda-macbook\" id=\"aacaa\"><si xmlns=\"http://jabber.org/protocol/si\" profile=\"http://jabber.org/protocol/si/profile/file-transfer\" id=\"ft_151a\"><file xmlns=\"http://jabber.org/protocol/si/profile/file-transfer\" size=\"9932529\" name=\"debug.log\"><range/></file><feature xmlns=\"http://jabber.org/protocol/feature-neg\"><x xmlns=\"jabber:x:data\" type=\"form\"><field type=\"list-single\" var=\"stream-method\"><option><value>http://jabber.org/protocol/bytestreams</value></option><option><value>http://jabber.org/protocol/ibb</value></option></field></x></feature></si></iq>".getBytes();

    private TempFolder temp;

    public FileStoreManagerTest() {
    }

    @Before
    public void setUp() throws IOException {
        this.temp = new TempFolder();
        System.err.println("test is using temp folder: " + this.temp.raw().getAbsolutePath());
    }

    @After
    public void tearDown() {
        if (this.temp != null) {
            this.temp.destroy();
        }
    }

    private File withContent(File file, byte[]   ... content) throws IOException {
        BufferedOutputStream buf = new BufferedOutputStream(new FileOutputStream(file, true));
        try {
            for (byte[] b : content) {
                buf.write(b);
            }
        } finally {
            buf.close();
        }
        return file;
    }

    private File withContent(File file) throws IOException {
        return withContent(file, TEST_MESSAGE, TEST_FILE);
    }

    private File unzip(File source) throws IOException {
        File result = null;
        ZipFile zipFile = new ZipFile(source);
        Enumeration<? extends ZipEntry> entries = zipFile.entries();
        if (entries.hasMoreElements()) {
            ZipEntry entry = entries.nextElement();
            result = this.temp.createFile();

            InputStream in = zipFile.getInputStream(entry);
            try {
                OutputStream out = new FileOutputStream(result);
                try {
                    ByteStreams.copy(in, out);
                } finally {
                    out.close();
                }
            } finally {
                in.close();
            }
        }
        return result;
    }
    
    @Test
    public void testConstructor_FolderDoesNotExist() throws Exception {
        File fakeFile = temp.get("idontexist");
        assertThat(fakeFile.exists(), is(false));
        FileFactory fsm = new FileFactory(fakeFile); //no exception
        assertThat(fakeFile.isDirectory(), is(false)); //folder is not created
    }
    
    @Test
    public void testConstructor_FolderIsFile() throws Exception {
        File file = temp.createFile("imfile");
        
        assertThat(file.isFile(), is(true)); 
        
        FileFactory fsm = new FileFactory(file); //no exception       
        assertThat(file.isFile(), is(true)); //file is untouched
    }

    @Test
    public void testConsolidate_FolderDoesNotExist() throws Exception {
        File fakeFile = temp.get("idontexist");
        assertThat(fakeFile.exists(), is(false));
        FileFactory fsm = new FileFactory(fakeFile);
        fsm.consolidate(new Date());
        
        assertThat(fakeFile.isDirectory(), is(true));
    }

    @Test
    public void testConsolidate_FolderIsFile() throws Exception {
        File file = temp.createFile("imfile");
        
        assertThat(file.isFile(), is(true));
        try {
            FileFactory fsm = new FileFactory(file);
            fsm.consolidate(new Date());
            fail();
        } catch (IOException e) {
            //expected
        }
             
        assertThat(file.isFile(), is(true));
    }

    @Test
    public void testZip_File() throws Exception {

        FileFactory manager = new FileFactory(temp.raw());
        File originalFile = withContent(temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, new Date())));
        long originalSize = originalFile.length();
        File zippedFile = temp.get(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, new Date()));
        File zippedTempFile = temp.get(String.format(FileFactory.FILE_NAME_TMP_FORMAT, new Date()));

        assertThat(originalFile.exists(), is(true));
        assertTrue(originalSize > 0);
        assertThat(zippedFile.exists(), is(false));
        assertThat(zippedTempFile.exists(), is(false));

        manager.zip(originalFile);

        assertThat(originalFile.exists(), is(false));
        assertThat(zippedFile.exists(), is(true));
        assertThat(zippedTempFile.exists(), is(false));

        File unzipedFile = unzip(zippedFile);
        assertThat(unzipedFile.exists(), is(true));
        assertThat(originalSize, is(unzipedFile.length()));
    }

    @Test
    public void testZip_file_sourceNotExist() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        File originalFile = temp.get(String.format(FileFactory.FILE_NAME_FORMAT, new Date()));
        File zippedFile = temp.get(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, new Date()));
        File zippedTempFile = temp.get(String.format(FileFactory.FILE_NAME_TMP_FORMAT, new Date()));

        assertThat(originalFile.exists(), is(false));
        assertThat(zippedFile.exists(), is(false));
        assertThat(zippedTempFile.exists(), is(false));

        try {
            manager.zip(originalFile);
            fail();
        } catch (IOException e) {
            //expected
        }

        assertThat(originalFile.exists(), is(false));
        assertThat(zippedFile.exists(), is(false));
        assertThat(zippedTempFile.exists(), is(false));
    }

    @Test
    public void testZip_file_destinationAlreadyExist() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        File originalFile = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, new Date()));
        long originalFileModification = originalFile.lastModified();
        File zippedFile = temp.createFile(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, new Date()));
        long zippedFileModification = zippedFile.lastModified();
        File zippedTempFile = temp.get(String.format(FileFactory.FILE_NAME_TMP_FORMAT, new Date()));

        assertThat(originalFile.exists(), is(true));
        assertThat(zippedFile.exists(), is(true));
        assertThat(zippedTempFile.exists(), is(false));

        try {
            manager.zip(originalFile);
            fail();
        } catch (IOException e) {
            //expected
        }

        assertThat(originalFile.exists(), is(true));
        assertThat(originalFile.lastModified(), is(originalFileModification));
        assertThat(zippedFile.exists(), is(true));
        assertThat(zippedFile.lastModified(), is(zippedFileModification));
        assertThat(zippedTempFile.exists(), is(false));
    }

    @Test
    public void testZip_file_tmpAlreadyExist() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        File originalFile = withContent(temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, new Date())));
        long originalSize = originalFile.length();
        File zippedFile = temp.get(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, new Date()));
        File zippedTempFile = temp.createFile(String.format(FileFactory.FILE_NAME_TMP_FORMAT, new Date()));

        assertThat(originalFile.exists(), is(true));
        assertThat(zippedFile.exists(), is(false));
        assertThat(zippedTempFile.exists(), is(true));

        manager.zip(originalFile);

        assertThat(originalFile.exists(), is(false));
        assertThat(zippedFile.exists(), is(true));
        assertThat(zippedTempFile.exists(), is(false));

        File unzipedFile = unzip(zippedFile);
        assertThat(unzipedFile.exists(), is(true));
        assertThat(originalSize, is(unzipedFile.length()));
    }

    @Test
    public void testZip_File_File() throws Exception {
        FileFactory manager = new FileFactory(temp.raw());
        File originalFile = withContent(temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, new Date())));
        File customDestination = temp.get("custom-destination.txt");

        assertThat(originalFile.exists(), is(true));
        assertTrue(originalFile.length() > 0);
        assertThat(customDestination.exists(), is(false));

        manager.zip(originalFile, customDestination);

        assertThat(originalFile.exists(), is(true));
        assertThat(customDestination.exists(), is(true));

        File unzipedFile = unzip(customDestination);
        assertThat(unzipedFile.exists(), is(true));
        assertThat(originalFile.length(), is(unzipedFile.length()));
    }

    @Test
    public void testZip_File_File_sourceNotExist() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        File originalFile = temp.get(String.format(FileFactory.FILE_NAME_FORMAT, new Date()));
        File customDestination = temp.get("custom-destination.txt");

        assertThat(originalFile.exists(), is(false));
        assertThat(customDestination.exists(), is(false));

        try {
            manager.zip(originalFile, customDestination);
            fail();
        } catch (IOException e) {
            //expected
        }

        assertThat(originalFile.exists(), is(false));
        assertThat(customDestination.exists(), is(false));
    }

    @Test
    public void testZip_File_File_destinationAlreadyExist() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        File originalFile = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, new Date()));
        File customDestination = temp.createFile("custom-destination.txt");
        long customDestinationModification = customDestination.lastModified();

        assertThat(originalFile.exists(), is(true));
        assertThat(customDestination.exists(), is(true));

        try {
            manager.zip(originalFile, customDestination);
            fail();
        } catch (IOException e) {
            //expected
        }

        assertThat(originalFile.exists(), is(true));
        assertThat(customDestination.exists(), is(true));
        assertThat(customDestination.lastModified(), is(customDestinationModification));
    }

    @Test
    public void testCopy() throws Exception {
        FileFactory manager = new FileFactory(temp.raw());
        File source = withContent(temp.createFile("source"));
        File destination = temp.get("destination");

        assertThat(source.exists(), is(true));
        assertThat(destination.exists(), is(false));

        manager.copy(source, destination);

        assertThat(source.exists(), is(true));
        assertThat(destination.exists(), is(true));
        assertThat(destination.length(), is(source.length()));
    }

    @Test
    public void testCopy_sourceNotExists() throws Exception {

        FileFactory manager = new FileFactory(temp.raw());
        File source = temp.get("source");
        File destination = temp.get("destination");

        assertThat(source.exists(), is(false));
        assertThat(destination.exists(), is(false));

        try {
            manager.copy(source, destination);
            fail();
        } catch (IOException e) {
            //expected
        }
    }

    @Test
    public void testCopy_destinationExists() throws Exception {
        FileFactory manager = new FileFactory(temp.raw());
        File source = temp.get("source");
        File destination = temp.createFile("destination");

        assertThat(source.exists(), is(false));
        assertThat(destination.exists(), is(true));

        try {
            manager.copy(source, destination);
            fail();
        } catch (IOException e) {
            //expected
        }
    }

    @Test
    public void testNext() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        Date now = new Date();
        assertThat(manager.next(now), is(temp.get(String.format(FileFactory.FILE_NAME_FORMAT, now))));
    }

    @Test
    public void testNext_Exist() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());

        Date now = new Date();
        assertThat(temp.createFile(manager.next(now).getName()).exists(), is(true));

        try {
            manager.next(now);
            fail();
        } catch (IOException e) {
            //expected
        }
    }

    @Test
    public void testNext_minuteDiff() throws IOException {

        FileFactory manager = new FileFactory(temp.raw());

        Date now = new Date();
        Date oneSecondBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(1, TimeUnit.SECONDS));
        Date oneMinuteBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(1, TimeUnit.MINUTES));

        File txt = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, now));
        assertThat(txt.exists(), is(true));

        try {
            manager.next(oneSecondBefore); //fails
            fail();
        } catch (IOException e) {
            //expected
        }

        manager.next(oneMinuteBefore); //pass

    }

    @Test
    public void testWithType_File_FileStoreManagerTYPE() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());

        for (FileFactory.TYPE sourceType : FileFactory.TYPE.values()) {
            assertThat(manager.withType(temp.get("aaa.zip.txt.tmp" + sourceType.suffix), FileFactory.TYPE.TXT), is(new File(temp.raw(), "aaa.zip.txt.tmp.txt")));
        }

        for (FileFactory.TYPE sourceType : FileFactory.TYPE.values()) {
            assertThat(manager.withType(temp.get("aaa.zip.txt.tmp" + sourceType.suffix), FileFactory.TYPE.ZIP), is(new File(temp.raw(), "aaa.zip.txt.tmp.zip")));
        }

        for (FileFactory.TYPE sourceType : FileFactory.TYPE.values()) {
            assertThat(manager.withType(temp.get("aaa.zip.txt.tmp" + sourceType.suffix), FileFactory.TYPE.PART), is(new File(temp.raw(), "aaa.zip.txt.tmp.part")));
        }
    }

    @Test
    public void testWithType_String_FileStoreManagerTYPE() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());

        for (FileFactory.TYPE sourceType : FileFactory.TYPE.values()) {
            assertThat(manager.withType("aaa.zip.txt.tmp" + sourceType.suffix, FileFactory.TYPE.TXT), is(new File(temp.raw(), "aaa.zip.txt.tmp.txt")));
        }

        for (FileFactory.TYPE sourceType : FileFactory.TYPE.values()) {
            assertThat(manager.withType("aaa.zip.txt.tmp" + sourceType.suffix, FileFactory.TYPE.ZIP), is(new File(temp.raw(), "aaa.zip.txt.tmp.zip")));
        }

        for (FileFactory.TYPE sourceType : FileFactory.TYPE.values()) {
            assertThat(manager.withType("aaa.zip.txt.tmp" + sourceType.suffix, FileFactory.TYPE.PART), is(new File(temp.raw(), "aaa.zip.txt.tmp.part")));
        }

    }

    @Test
    public void testConsolidate_EmptyFolder() throws Exception {
        FileFactory manager = new FileFactory(temp.raw());
        assertThat(temp.raw().list().length, is(0));

        Date now = new Date();
        File expected = temp.get(String.format(FileFactory.FILE_NAME_FORMAT, now));
        assertThat(manager.consolidate(now), is(expected));

        assertThat(temp.raw().list().length, is(0));
    }

    /*
     * All tmp are deleted nevertheless other files present or not
     * file.tmp, file.txt, file.zip, file1.tmp, file2.txt, file2.tmp
     */
    @Test
    public void testConsolidate_tempDeleted() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        Date fourMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(4, TimeUnit.MINUTES));
        File txt = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, fourMinutesBefore));
        File tmp = temp.createFile(String.format(FileFactory.FILE_NAME_TMP_FORMAT, fourMinutesBefore));

        Date threeMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(3, TimeUnit.MINUTES));
        File tmp2 = temp.createFile(String.format(FileFactory.FILE_NAME_TMP_FORMAT, threeMinutesBefore));
        File zip2 = temp.createFile(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, threeMinutesBefore));

        Date twoMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(2, TimeUnit.MINUTES));
        File tmp3 = temp.createFile(String.format(FileFactory.FILE_NAME_TMP_FORMAT, twoMinutesBefore));

        assertThat(tmp.exists(), is(true));
        assertThat(txt.exists(), is(true));
        assertThat(zip2.exists(), is(true));
        assertThat(tmp2.exists(), is(true));
        assertThat(tmp3.exists(), is(true));

        manager.consolidate(new Date());

        assertThat(tmp.exists(), is(false));
        assertThat(txt.exists(), is(true));
        assertThat(zip2.exists(), is(true));
        assertThat(tmp2.exists(), is(false));
        assertThat(tmp3.exists(), is(false));

    }

    /*
     * Empty folder or only zips, passed date file name is returned
     * empty --> *newfile.txt
     */
    @Test
    public void testConsolidate_empty() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        assertThat(temp.raw().list().length, is(0));

        Date fourDaysBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(4, TimeUnit.DAYS));
        File expected = temp.get(String.format(FileFactory.FILE_NAME_FORMAT, fourDaysBefore));

        File result = manager.consolidate(fourDaysBefore);

        assertThat(temp.raw().list().length, is(0));
        assertThat(result.getAbsolutePath(), is(expected.getAbsolutePath()));
    }

    /*
     * Empty folder or only zips, passed date file name is returned
     * file.zip, file2.zip --> *newfile.txt
     */
    @Test
    public void testConsolidate_onlyZips() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());

        Date fourDaysBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(4, TimeUnit.DAYS));
        File zip = temp.createFile(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, fourDaysBefore));

        Date threeDaysBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(3, TimeUnit.DAYS));
        File zip1 = temp.createFile(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, threeDaysBefore));

        assertThat(temp.raw().list().length, is(2));
        assertThat(zip.exists(), is(true));
        assertThat(zip1.exists(), is(true));

        Date oneDayBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(1, TimeUnit.DAYS));
        File expected = temp.get(String.format(FileFactory.FILE_NAME_FORMAT, oneDayBefore));

        File result = manager.consolidate(oneDayBefore);

        assertThat(result.getAbsolutePath(), is(expected.getAbsolutePath()));
        assertThat(temp.raw().list().length, is(2));
        assertThat(zip.exists(), is(true));
        assertThat(zip1.exists(), is(true));
    }

    /* 
     * All txt are zipped except the last one (no zip at all)
     * file.txt --> *file.txt
     */
    @Test
    public void testConsolidate_noZip() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());

        Date fourMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(4, TimeUnit.MINUTES));
        File txt = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, fourMinutesBefore));

        assertThat(temp.raw().list().length, is(1));
        assertThat(txt.exists(), is(true));

        Date now = new Date();
        File result = manager.consolidate(now);

        assertThat(temp.raw().list().length, is(1));
        assertThat(result.getAbsolutePath(), is(txt.getAbsolutePath()));
    }

    /* 
     * All txt are zipped except the last one
     * file.txt, file1.txt, file2.txt --> file.zip, file1.zip, *file2.txt
     */
    @Test
    public void testConsolidate_oldTxtAreZipped() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());

        Date fourMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(4, TimeUnit.MINUTES));
        File txt = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, fourMinutesBefore));
        File zip = temp.get(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, fourMinutesBefore));

        Date threeMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(3, TimeUnit.MINUTES));
        File txt2 = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, threeMinutesBefore));
        File zip2 = temp.get(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, threeMinutesBefore));

        Date twoMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(2, TimeUnit.MINUTES));
        File txt3 = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, twoMinutesBefore));

        assertThat(temp.raw().list().length, is(3));
        assertThat(txt.exists(), is(true));
        assertThat(txt2.exists(), is(true));
        assertThat(txt3.exists(), is(true));

        Date now = new Date();
        File result = manager.consolidate(now);

        assertThat(temp.raw().list().length, is(3));
        assertThat(txt.exists(), is(false));
        assertThat(zip.exists(), is(true));
        assertThat(txt2.exists(), is(false));
        assertThat(zip2.exists(), is(true));
        assertThat(txt3.exists(), is(true));
        assertThat(result.getAbsolutePath(), is(result.getAbsolutePath()));

    }

    /* 
     * Zip and TXT duplicates
     * file.txt, file.zip --> file.zip, *newfile.txt
     */
    @Test
    public void testConsolidate_allDuplicates() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());

        Date fourMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(4, TimeUnit.MINUTES));
        File txt = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, fourMinutesBefore));
        File zip = temp.createFile(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, fourMinutesBefore));

        assertThat(temp.raw().list().length, is(2));
        assertThat(txt.exists(), is(true));
        assertThat(zip.exists(), is(true));

        Date now = new Date();
        File expected = temp.get(String.format(FileFactory.FILE_NAME_FORMAT, now));

        File result = manager.consolidate(now);

        assertThat(temp.raw().list().length, is(1));
        assertThat(txt.exists(), is(false));
        assertThat(zip.exists(), is(true));
        assertThat(result.getAbsolutePath(), is(expected.getAbsolutePath()));
    }

    /* 
     * Zip and TXT duplicates
     * file.txt, file.zip, file1.txt --> file.zip, *file1.txt
     */
    @Test
    public void testConsolidate_duplicates() throws IOException {

        FileFactory manager = new FileFactory(temp.raw());

        Date fourMinutesBefore = new Date(new Date().getTime() - TimeUnit.MILLISECONDS.convert(4, TimeUnit.MINUTES));
        File txt = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, fourMinutesBefore));
        File zip = temp.createFile(String.format(FileFactory.FILE_NAME_ZIP_FORMAT, fourMinutesBefore));

        Date threeMinutesBefore = new Date();
        File txt2 = temp.createFile(String.format(FileFactory.FILE_NAME_FORMAT, threeMinutesBefore));

        assertThat(temp.raw().list().length, is(3));
        assertThat(txt.exists(), is(true));
        assertThat(zip.exists(), is(true));
        assertThat(txt2.exists(), is(true));

        File result = manager.consolidate(new Date()); //!passing current time

        assertThat(temp.raw().list().length, is(2));
        assertThat(txt.exists(), is(false));
        assertThat(zip.exists(), is(true));
        assertThat(txt2.exists(), is(true));
        assertThat(result.getAbsolutePath(), is(txt2.getAbsolutePath()));

    }

    /*
     * Foreign files are not touched.
     * foreign.tmp, foreign.txt, foreign.zip --> foreign.tmp, foreign.txt, foreign.zip, *newfile.txt
     */
    @Test
    public void testConsolidate_foreign() throws IOException {
        FileFactory manager = new FileFactory(temp.raw());
        File foreignTxt = temp.createFile("foreign" + FileFactory.TYPE.TXT);
        File foreignZip = temp.createFile("foreign" + FileFactory.TYPE.ZIP);
        File foreignTemp = temp.createFile("foreign" + FileFactory.TYPE.PART);

        assertThat(temp.raw().list().length, is(3));
        assertThat(foreignTxt.exists(), is(true));
        assertThat(foreignZip.exists(), is(true));
        assertThat(foreignTemp.exists(), is(true));

        manager.consolidate(new Date());

        assertThat(temp.raw().list().length, is(3));
        assertThat(foreignTxt.exists(), is(true));
        assertThat(foreignZip.exists(), is(true));
        assertThat(foreignTemp.exists(), is(true));
    }

}
