/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import com.kerio.im.core.conversation.TempFolder;
import com.kerio.im.core.conversation.archive.FileStoreWriter.SizeCheckException;
import java.io.File;
import java.io.IOException;
import java.time.LocalDateTime;
import java.time.Month;
import java.util.Date;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import org.mockito.Mockito;

/**
 *
 * @author tskoda
 */
public class FileStoreTest {

    private File fileReturnedFromFileStoreManager_consolidate;
    private File fileAfterReopen;
    private FileStoreWriter archiveStoreAfterReopen;
    private FileFactory mockFileStoreManager;
    private FileStoreWriterFactory mockArchiveStoreFactory;
    private File fileAfterOpen;
    private FileStoreWriter archiveStoreAfterOpen;

    private TempFolder tmp;

    public FileStoreTest() {
    }

    @Before
    public void setUp() throws IOException {
        this.fileReturnedFromFileStoreManager_consolidate = new File("store/manager/file");
        this.mockFileStoreManager = mock(FileFactory.class);
        when(this.mockFileStoreManager.consolidate(Mockito.any(Date.class))).thenReturn(this.fileReturnedFromFileStoreManager_consolidate);

        this.fileAfterReopen = new File("archive/store/file");
        this.archiveStoreAfterReopen = mock(FileStoreWriter.class);
        when(this.archiveStoreAfterReopen.file()).thenReturn(this.fileAfterReopen);

        this.fileAfterOpen = new File("archive/store/date/file");
        this.archiveStoreAfterOpen = mock(FileStoreWriter.class);
        when(this.archiveStoreAfterOpen.file()).thenReturn(this.fileAfterOpen);

        this.mockArchiveStoreFactory = mock(FileStoreWriterFactory.class);
        when(this.mockArchiveStoreFactory.create(Mockito.any(File.class))).thenReturn(this.archiveStoreAfterOpen);
        when(this.mockArchiveStoreFactory.create()).thenReturn(this.archiveStoreAfterReopen);

        this.tmp = new TempFolder();
    }

    @After
    public void tearDown() {
        if (this.tmp != null) {
            this.tmp.destroy();
        }
    }

    @Test
    public void testOpen() throws Exception {

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.open();

        verify(this.mockFileStoreManager, times(1)).consolidate(Mockito.any(Date.class));
        verify(this.mockArchiveStoreFactory, times(1)).create(this.fileReturnedFromFileStoreManager_consolidate);
    }

    @Test
    public void testOpen_storeManagerException() throws Exception {
        when(this.mockFileStoreManager.consolidate(Mockito.any(Date.class))).thenThrow(new IOException());
        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        try {
            store.open();
            fail();
        } catch (IOException e) {
            //expected
        }
    }

    @Test
    public void testOpen_archiveStoreFactoryException() throws Exception {
        when(this.mockArchiveStoreFactory.create(Mockito.any(File.class))).thenThrow(new IOException());
        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        try {
            store.open();
            fail();
        } catch (IOException e) {
            //expected
        }
    }

    @Test
    public void testStore() throws Exception {

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.open();
        
        LocalDateTime time = LocalDateTime.of(2012, Month.MARCH, 12, 13, 14, 15);
        String message = "some message";

        store.store(time, message);
        verify(this.archiveStoreAfterOpen, times(1)).write(time, message);
    }

    @Test
    public void testStore_notOpenned() throws Exception {
        FileStore store = new FileStore(mockFileStoreManager, mockArchiveStoreFactory);
        try {
            store.store(LocalDateTime.MAX, "some message");
            fail();
        } catch (IOException e) {
            //expected
        }
    }

    @Test
    public void testStore_IOException() throws Exception {
        doThrow(new IOException()).when(this.archiveStoreAfterOpen).write(any(LocalDateTime.class), anyString());

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.open();

        try {
            store.store(LocalDateTime.MIN, "some message");
            fail();
        } catch (IOException e) {
            //expected
        }
    }

    @Test
    public void testStore_FileFull() throws Exception {
        doThrow(new SizeCheckException()).when(this.archiveStoreAfterOpen).write(any(LocalDateTime.class), anyString()); //means file is full

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.open();
        
        LocalDateTime time = LocalDateTime.of(2012, Month.MARCH, 12, 13, 14, 15);
        String message = "some message";

        store.store(time, message);

        //called on open, only once
        verify(this.mockArchiveStoreFactory, times(1)).create(Mockito.any(File.class));
        //the first wrote
        verify(this.archiveStoreAfterOpen, times(1)).write(time, message);
        //the old archive store should be closed and zipped when full
        verify(this.archiveStoreAfterOpen, times(1)).close();
        verify(this.mockFileStoreManager, times(1)).zip(this.fileAfterOpen);
        //open new one
        verify(this.mockArchiveStoreFactory, times(1)).create();
        //write the message again
        verify(this.archiveStoreAfterReopen, times(1)).write(time, message); //the second try to write it
    }
    
        @Test
    public void testStore_FileFull_PackingException() throws Exception {
        doThrow(new SizeCheckException()).when(this.archiveStoreAfterOpen).write(any(LocalDateTime.class), anyString()); //means file is full
        doThrow(new IOException()).when(this.mockFileStoreManager).zip(Mockito.any(File.class));

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.open();
        
        LocalDateTime time = LocalDateTime.of(2012, Month.MARCH, 12, 13, 14, 15);
        String messageContent = "some message";

        store.store(time, messageContent);

        //called on open, only once
        verify(this.mockArchiveStoreFactory, times(1)).create(Mockito.any(File.class));
        //the first wrote
        verify(this.archiveStoreAfterOpen, times(1)).write(time, messageContent);
        //the old archive store should be closed and zipped when full
        verify(this.archiveStoreAfterOpen, times(1)).close();
        verify(this.mockFileStoreManager, times(1)).zip(this.fileAfterOpen);
        //open new one
        verify(this.mockArchiveStoreFactory, times(1)).create();
        //write the message again
        verify(this.archiveStoreAfterReopen, times(1)).write(time, messageContent); //the second try to write it
    }

    @Test
    public void testExport_destinationParentDoesNotExist() throws Exception {
        File sourceZip = tmp.createFile("source.zip");
        File sourceTxt = tmp.createFile("source.txt");
        File destinationParent = tmp.get("destination-parent");
        File destination = new File(destinationParent, "destination-file");

        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.TXT))).thenReturn(sourceTxt);
        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.ZIP))).thenReturn(sourceZip);

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        try {
            store.export(sourceTxt.getAbsolutePath(), destination.getAbsolutePath());
            fail();
        } catch (IOException e) {
            //expected
        }

        verify(this.mockFileStoreManager, never()).withType(Mockito.any(String.class), Mockito.any(FileFactory.TYPE.class));
        verify(this.mockFileStoreManager, never()).copy(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class));
    }

    @Test
    public void testExport_destinationExist() throws Exception {
        File sourceZip = tmp.createFile("source.zip");
        File sourceTxt = tmp.createFile("source.txt");
        File destinationParent = tmp.createFolder("destination-parent");
        File destination = new File(destinationParent, "destination-file");
        tmp.createFileCustom(destination);

        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.TXT))).thenReturn(sourceTxt);
        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.ZIP))).thenReturn(sourceZip);

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        try {
            store.export(sourceTxt.getAbsolutePath(), destination.getAbsolutePath());
            fail();
        } catch (IOException e) {
            //expected
        }

        verify(this.mockFileStoreManager, never()).withType(Mockito.any(String.class), Mockito.any(FileFactory.TYPE.class));
        verify(this.mockFileStoreManager, never()).copy(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class));
    }

    @Test
    public void testExport_destinationParentFile() throws Exception {
        File sourceZip = tmp.createFile("source.zip");
        File sourceTxt = tmp.createFile("source.txt");
        File destinationParent = tmp.createFile("destination-parent");
        File destination = new File(destinationParent, "destination-file");

        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.TXT))).thenReturn(sourceTxt);
        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.ZIP))).thenReturn(sourceZip);

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        try {
            store.export(sourceTxt.getAbsolutePath(), destination.getAbsolutePath());
            fail();
        } catch (IOException e) {
            //expected
        }

        verify(this.mockFileStoreManager, never()).withType(Mockito.any(String.class), Mockito.any(FileFactory.TYPE.class));
        verify(this.mockFileStoreManager, never()).copy(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class));
    }

    @Test
    public void testExport_sourceDoesNotExist() throws IOException {
        File sourceZip = tmp.get("source.zip");
        File sourceTxt = tmp.get("source.txt");
        File destinationParent = tmp.createFolder("destination-parent");
        File destination = new File(destinationParent, "destination-file");

        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.TXT))).thenReturn(sourceTxt);
        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.ZIP))).thenReturn(sourceZip);

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        try {
            store.export(sourceTxt.getAbsolutePath(), destination.getAbsolutePath());
            fail();
        } catch (IOException e) {
            //expected
        }

        verify(this.mockFileStoreManager, times(2)).withType(Mockito.any(String.class), Mockito.any(FileFactory.TYPE.class));
        verify(this.mockFileStoreManager, never()).copy(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class));
    }

    @Test
    public void testExport_sourceZipExistsTxt() throws IOException {
        File sourceZip = tmp.get("source.zip");
        File sourceTxt = tmp.createFile("source.txt");
        File destinationParent = tmp.createFolder("destination-parent");
        File destination = new File(destinationParent, "destination-file");

        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.TXT))).thenReturn(sourceTxt);
        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.ZIP))).thenReturn(sourceZip);

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.export(sourceZip.getAbsolutePath(), destination.getAbsolutePath());

        verify(this.mockFileStoreManager, times(2)).withType(Mockito.any(String.class), Mockito.any(FileFactory.TYPE.class));
        verify(this.mockFileStoreManager, never()).copy(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, times(1)).zip(eq(sourceTxt), eq(destination));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class));
    }

    @Test
    public void testExport_sourceTxtExistsZip() throws IOException {
        File sourceZip = tmp.createFile("source.zip");
        File sourceTxt = tmp.get("source.txt");
        File destinationParent = tmp.createFolder("destination-parent");
        File destination = new File(destinationParent, "destination-file");

        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.TXT))).thenReturn(sourceTxt);
        when(this.mockFileStoreManager.withType(Mockito.any(String.class), eq(FileFactory.TYPE.ZIP))).thenReturn(sourceZip);

        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.export(sourceZip.getAbsolutePath(), destination.getAbsolutePath());

        verify(this.mockFileStoreManager, times(2)).withType(Mockito.any(String.class), Mockito.any(FileFactory.TYPE.class));
        verify(this.mockFileStoreManager, times(1)).copy(eq(sourceZip), eq(destination));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class), Mockito.any(File.class));
        verify(this.mockFileStoreManager, never()).zip(Mockito.any(File.class));
    }

    @Test
    public void testCloseNotOpenned() {
        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.close();
        verify(this.archiveStoreAfterOpen, times(0)).close();
    }
    
    @Test
    public void testClose() throws IOException {
        FileStore store = new FileStore(this.mockFileStoreManager, this.mockArchiveStoreFactory);
        store.open();
        
        store.close();
        verify(this.archiveStoreAfterOpen, times(1)).close();
    }

}
