/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.conversation.archive;

import com.kerio.im.core.conversation.TempFolder;
import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.time.LocalDateTime;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.*;

/**
 *
 * @author tskoda
 */
public class FileStoreWriterTest {

    private TempFolder temp;

    public FileStoreWriterTest() {
    }

    @Before
    public void setUp() throws IOException {
        this.temp = new TempFolder();
    }

    @After
    public void tearDown() {
        if (temp != null) {
            temp.destroy();
        }
    }

    private byte[] read(File file) throws IOException {
        BufferedInputStream reader = new BufferedInputStream(new FileInputStream(file));
        try {
            ByteArrayOutputStream sb = new ByteArrayOutputStream();
            byte[] buffer = new byte[512];
            int count;
            while ((count = reader.read(buffer)) != -1) {
                sb.write(buffer, 0, count);
            }
            return sb.toByteArray();
            
        } finally {
            reader.close();
        }
    }

    private void write(File file, String text) throws IOException {
        BufferedWriter writer = new BufferedWriter(new FileWriter(file));
        try {
            writer.write(text);
        } finally {
            writer.close();
        }
    }
    
    private byte[] getUTF8bytes(Object... components) throws UnsupportedEncodingException {
        StringBuilder buffer = new StringBuilder();
        for (Object component: components) {
            if (component instanceof LocalDateTime) {
                buffer.append("[");
                buffer.append(((LocalDateTime)component).format(FileStoreWriter.DATE_TIME_FORMAT));
                buffer.append("] ");    
            } else {
                buffer.append(component.toString());
            }   
        }
        return buffer.toString().getBytes("UTF-8");
    }

    @Test
    public void testWriteDestinationExist() throws Exception {
        File destination = temp.createFile("destination");
        FileStoreWriter.SizeCheck check = new FileStoreWriter.SizeCheck(destination, 10);
        FileStoreWriter writer = new FileStoreWriter(destination, check);

        LocalDateTime time = LocalDateTime.now();
        writer.write(time, "hi");
        writer.close();

        assertThat(read(destination), is(getUTF8bytes(time, "hi", "\n")));
    }

    @Test
    public void testWriteDestinationAppend() throws Exception {
        File destination = temp.createFile("destination");
        write(destination, "hello\n");
        assertThat(read(destination), is("hello\n".getBytes("UTF-8")));

        FileStoreWriter.SizeCheck check = new FileStoreWriter.SizeCheck(destination, 10);
        FileStoreWriter writer = new FileStoreWriter(destination, check);

        LocalDateTime time = LocalDateTime.now();
        writer.write(time, "hi");
        writer.close();

        assertThat(read(destination), is(getUTF8bytes("hello", "\n", time, "hi", "\n")));
    }

    @Test
    public void testWriteNewLine() throws Exception {
        File destination = temp.createFile("destination");

        FileStoreWriter.SizeCheck check = new FileStoreWriter.SizeCheck(destination, 10);
        FileStoreWriter writer = new FileStoreWriter(destination, check);
        
        LocalDateTime time = LocalDateTime.now();
        writer.write(time, "hi\nhello");
        writer.close();

        assertThat(read(destination), is(getUTF8bytes(time, "hihello", "\n")));
    }

    @Test
    public void testWriteNewLineOnly() throws Exception {
        File destination = temp.createFile("destination");

        FileStoreWriter.SizeCheck check = new FileStoreWriter.SizeCheck(destination, 10);
        FileStoreWriter writer = new FileStoreWriter(destination, check);
        
        LocalDateTime time = LocalDateTime.now();
        writer.write(time, "\n");
        writer.close();

        assertThat(read(destination), is(getUTF8bytes("")));
    }

    @Test
    public void testWriteEmpty() throws Exception {
        File destination = temp.createFile("destination");
        FileStoreWriter.SizeCheck check = new FileStoreWriter.SizeCheck(destination, 10);
        FileStoreWriter writer = new FileStoreWriter(destination, check);

        LocalDateTime time = LocalDateTime.now();
        writer.write(time, "");
        writer.close();

        assertThat(read(destination), is("".getBytes("UTF-8")));

    }

    @Test
    public void testWriteDestinationFolder() throws Exception {
        File destination = temp.createFolder("destination_folder");
        assertThat(destination.isDirectory(), is(true));

        FileStoreWriter.SizeCheck check = new FileStoreWriter.SizeCheck(destination, 10);
        
        try {
            FileStoreWriter writer = new FileStoreWriter(destination, check);
        } catch (IOException e) {
            //expected
        }

        assertThat(destination.isDirectory(), is(true));
    }

    @Test
    public void testWriteTooBig() throws Exception {

        LocalDateTime time = LocalDateTime.now();
        String content = "content";

        int maximumContentLength = getUTF8bytes(content, "\n", time, content, "\n").length + 1;
        
        File destination = temp.createFile("destination_folder");
        write(destination, content + "\n");
        assertThat(read(destination), is(getUTF8bytes(content, "\n")));
        
        FileStoreWriter.SizeCheck check = new FileStoreWriter.SizeCheck(destination, maximumContentLength, 1); //first can be written, second not, check every single write
        FileStoreWriter writer = new FileStoreWriter(destination, check);
        
        writer.write(time, content); //first write is successful
        writer.write(time, content); //second write is successful
        try {
            writer.write(time, content); //third write fails
            fail();
        } catch (FileStoreWriter.SizeCheckException e) {
            //expected
        }
        
        //first content was written at the beginning, two contents through writer
        assertThat(read(destination), is(getUTF8bytes(content, "\n", time, content, "\n", time, content, "\n")));
    }

    @Test
    public void testClose() {
    }

    @Test
    public void testFile() {
    }

}
