/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.core.service;

import java.net.UnknownHostException;
import org.junit.Test;
import static org.junit.Assert.*;
import static com.kerio.im.core.service.FilterTestUtil.*;
import java.math.BigInteger;
import java.net.InetAddress;

/**
 *
 * @author tskoda
 */
public class ServiceListenerFiltersTest {

    public ServiceListenerFiltersTest() {
    }

    @Test
    public void testLocalhostFilter() throws UnknownHostException {
        FilterLocalhost filter = new FilterLocalhost();

        assertFalse(filter.pass(ip(Long.MIN_VALUE)));
        assertFalse(filter.pass(ip(-1)));
        assertFalse(filter.pass(ip(0)));
        assertFalse(filter.pass(ip(Long.MAX_VALUE)));

        assertTrue(filter.pass(ip(127, 0, 0, 1)));

    }

    @Test
    public void testRangeFilter() throws UnknownHostException {
        FilterRange filter = new FilterRange("0.0.0.0", "0.0.0.0");

        assertFalse(filter.pass(ip(Long.MIN_VALUE)));
        assertFalse(filter.pass(ip(-1)));
        assertFalse(filter.pass(ip(Long.MAX_VALUE)));

        assertTrue(filter.pass(ip(0)));
        assertTrue(filter.pass(ip(0, 0, 0, 0)));
        assertFalse(filter.pass(ip(0, 0, 0, 1)));

        filter = new FilterRange("192.168.1.1", "192.168.1.2");

        assertFalse(filter.pass(ip(Long.MIN_VALUE)));
        assertFalse(filter.pass(ip(-1)));
        assertFalse(filter.pass(ip(0)));
        assertFalse(filter.pass(ip(Long.MAX_VALUE)));

        assertFalse(filter.pass(ip(192, 168, 1, 0)));
        assertTrue(filter.pass(ip(192, 168, 1, 1)));
        assertTrue(filter.pass(ip(192, 168, 1, 2)));
        assertFalse(filter.pass(ip(192, 168, 1, 3)));

        filter = new FilterRange("192.168.1.1", "192.168.2.1");

        assertFalse(filter.pass(ip(Long.MIN_VALUE)));
        assertFalse(filter.pass(ip(-1)));
        assertFalse(filter.pass(ip(0)));
        assertFalse(filter.pass(ip(Long.MAX_VALUE)));

        assertFalse(filter.pass(ip(192, 168, 1, 0)));
        assertTrue(filter.pass(ip(192, 168, 1, 1)));
        assertTrue(filter.pass(ip(192, 168, 2, 1)));
        assertFalse(filter.pass(ip(192, 168, 2, 2)));

        filter = new FilterRange("192.168.1.1", "192.169.1.1");

        assertFalse(filter.pass(ip(Long.MIN_VALUE)));
        assertFalse(filter.pass(ip(-1)));
        assertFalse(filter.pass(ip(0)));
        assertFalse(filter.pass(ip(Long.MAX_VALUE)));

        assertFalse(filter.pass(ip(10, 168, 1, 1)));
        assertFalse(filter.pass(ip(192, 168, 1, 0)));
        assertTrue(filter.pass(ip(192, 168, 1, 1)));
        assertTrue(filter.pass(ip(192, 169, 1, 1)));
        assertFalse(filter.pass(ip(192, 169, 1, 2)));
        assertFalse(filter.pass(ip(200, 169, 1, 1)));

        filter = new FilterRange("192.168.1.1", "193.168.1.1");

        assertFalse(filter.pass(ip(Long.MIN_VALUE)));
        assertFalse(filter.pass(ip(-1)));
        assertFalse(filter.pass(ip(0)));
        assertFalse(filter.pass(ip(Long.MAX_VALUE)));

        assertFalse(filter.pass(ip(10, 168, 1, 1)));
        assertFalse(filter.pass(ip(192, 168, 1, 0)));
        assertTrue(filter.pass(ip(192, 168, 1, 1)));
        assertTrue(filter.pass(ip(192, 255, 255, 255)));
        assertTrue(filter.pass(ip(193, 168, 1, 1)));
        assertFalse(filter.pass(ip(193, 168, 1, 2)));
        assertFalse(filter.pass(ip(200, 168, 1, 1)));
    }

    @Test
    public void testHostFilter() throws UnknownHostException {
        FilterHost filter = new FilterHost("1.1.1.1");

        assertFalse(filter.pass(ip(Long.MIN_VALUE)));
        assertFalse(filter.pass(ip(-1)));
        assertFalse(filter.pass(ip(0)));
        assertFalse(filter.pass(ip(Long.MAX_VALUE)));

        assertTrue(filter.pass(ip(1, 1, 1, 1)));
    }
    
    @Test
    public void testHostFilterVIPV6() throws UnknownHostException {
        FilterHost filter = new FilterHost("2001:718:1803:44:aa20:66ff:fe59:eb0a");
        assertTrue(filter.pass(ip("2001:718:1803:44:aa20:66ff:fe59:eb0a")));
        assertFalse(filter.pass(decrement(ip("2001:718:1803:44:aa20:66ff:fe59:eb0a"))));
        assertFalse(filter.pass(increment(ip("2001:718:1803:44:aa20:66ff:fe59:eb0a"))));
    }

    @Test
    public void testMaskClassC() throws UnknownHostException {

        FilterMask filter = FilterMask.create("192.168.1.0", "255.255.255.0");
        assertTrue(filter.pass(ip(192, 168, 1, 0))); //net
        assertTrue(filter.pass(ip(192, 168, 1, 1))); //client
        assertTrue(filter.pass(ip(192, 168, 1, 255))); //broadcast
        assertTrue(filter.pass(ip(192, 168, 1, 100)));

        assertFalse(filter.pass(ip(192, 168, 2, 0)));
        assertFalse(filter.pass(ip(192, 168, 0, 254)));
        assertFalse(filter.pass(ip(192, 169, 1, 1)));
        assertFalse(filter.pass(ip(192, 168, 2, 0)));
        
        filter = FilterMask.create("192.168.1.0", "24");
        assertFalse(filter.pass(ip(192,168,2,0)));
        assertTrue(filter.pass(ip(192,168,1,0))); //net
        assertTrue(filter.pass(ip(192,168,1,1))); //client
        assertTrue(filter.pass(ip(192,168,1,255))); //broadcast
        
        filter = FilterMask.create("192.168.1.0", "255.255.255.128");
        assertFalse(filter.pass(ip(192,168,2,0)));
        assertTrue(filter.pass(ip(192,168,1,0)));
        assertTrue(filter.pass(ip(192,168,1,127)));
        assertFalse(filter.pass(ip(192,168,1,128)));
        
        filter = FilterMask.create("192.168.1.128", "255.255.255.128");
        assertFalse(filter.pass(ip(192,168,2,0)));
        assertFalse(filter.pass(ip(192,168,1,127)));
        assertTrue(filter.pass(ip(192,168,1,128)));
        assertTrue(filter.pass(ip(192,168,1,129)));
        assertTrue(filter.pass(ip(192,168,1,255)));
        
        filter = FilterMask.create("192.168.1.0", "255.255.255.192");
        assertFalse(filter.pass(ip(192,168,2,0)));
        assertTrue(filter.pass(ip(192,168,1,0)));
        assertTrue(filter.pass(ip(192,168,1,63)));
        assertFalse(filter.pass(ip(192,168,1,64)));
        
        filter = FilterMask.create("192.168.1.64", "255.255.255.192");
        assertFalse(filter.pass(ip(192,168,2,0)));
        assertFalse(filter.pass(ip(192,168,1,63)));
        assertTrue(filter.pass(ip(192,168,1,64)));
        assertTrue(filter.pass(ip(192,168,1,127)));
        assertFalse(filter.pass(ip(192,168,1,128)));
        
        filter = FilterMask.create("192.168.1.128", "255.255.255.192");
        assertFalse(filter.pass(ip(192,168,2,0)));
        assertFalse(filter.pass(ip(192,168,1,127)));
        assertTrue(filter.pass(ip(192,168,1,128)));
        assertTrue(filter.pass(ip(192,168,1,191)));
        assertFalse(filter.pass(ip(192,168,1,192)));
        
        filter = FilterMask.create("192.168.1.192", "255.255.255.192");
        assertFalse(filter.pass(ip(192,168,2,0)));
        assertFalse(filter.pass(ip(192,168,1,191)));
        assertTrue(filter.pass(ip(192,168,1,192)));
        assertTrue(filter.pass(ip(192,168,1,255)));
    }

    @Test
    public void testPrefixClasC() throws UnknownHostException {
        FilterMask filter = FilterMask.create("192.168.1.0", "24");
        assertTrue(filter.pass(ip(192, 168, 1, 0))); //net
        assertTrue(filter.pass(ip(192, 168, 1, 1))); //client
        assertTrue(filter.pass(ip(192, 168, 1, 255))); //broadcast
        assertTrue(filter.pass(ip(192, 168, 1, 100))); //broadcast

        assertFalse(filter.pass(ip(192, 168, 2, 0)));
        assertFalse(filter.pass(ip(192, 168, 0, 254)));
        assertFalse(filter.pass(ip(192, 169, 1, 1)));
        assertFalse(filter.pass(ip(192, 168, 2, 0)));
    }

    @Test
    public void testMaskClassB() throws UnknownHostException {
        FilterMask filter = FilterMask.create("172.16.0.1", "255.255.0.0");

        assertTrue(filter.pass(ip(172, 16, 0, 1))); //net
        assertTrue(filter.pass(ip(172, 16, 255, 255))); //broadcast
        assertTrue(filter.pass(ip(172, 16, 0, 254)));
        assertTrue(filter.pass(ip(172, 16, 16, 16)));

        assertFalse(filter.pass(ip(172, 15, 255, 254)));
        assertFalse(filter.pass(ip(172, 14, 0, 1)));
    }

    @Test
    public void testPrefixClassB() throws UnknownHostException {
        FilterMask filter = FilterMask.create("172.16.0.1", "16");

        assertTrue(filter.pass(ip(172, 16, 0, 1))); //net
        assertTrue(filter.pass(ip(172, 16, 255, 255))); //broadcast
        assertTrue(filter.pass(ip(172, 16, 0, 254)));
        assertTrue(filter.pass(ip(172, 16, 16, 16)));

        assertFalse(filter.pass(ip(172, 15, 255, 254)));
        assertFalse(filter.pass(ip(172, 14, 0, 1)));
    }

    @Test
    public void testMaskClassA() throws UnknownHostException {
        FilterMask filter = FilterMask.create("10.0.0.1", "255.0.0.0");

        assertTrue(filter.pass(ip(10, 0, 0, 1)));  //net
        assertTrue(filter.pass(ip(10, 255, 255, 255))); //broadcast
        assertTrue(filter.pass(ip(10, 0, 0, 2)));
        assertTrue(filter.pass(ip(10, 255, 255, 254)));

        assertFalse(filter.pass(ip(9, 255, 255, 254)));
        assertFalse(filter.pass(ip(11, 0, 0, 1)));
    }

    @Test
    public void testPrefixClassA() throws UnknownHostException {
        FilterMask filter = FilterMask.create("10.0.0.1", "8");

        assertTrue(filter.pass(ip(10, 0, 0, 1)));  //net
        assertTrue(filter.pass(ip(10, 255, 255, 255))); //broadcast
        assertTrue(filter.pass(ip(10, 0, 0, 2)));
        assertTrue(filter.pass(ip(10, 255, 255, 254)));

        assertFalse(filter.pass(ip(9, 255, 255, 254)));
        assertFalse(filter.pass(ip(11, 0, 0, 1)));
    }

    @Test
    public void testPrefixIPV6() throws UnknownHostException {
        {
            FilterMask filter = FilterMask.create("2001:718:1803:44:aa20:66ff:fe59:eb0a", "65");

            assertTrue(filter.pass(ip("2001:0718:1803:0044:8000:0000:0000:0000")));
            assertTrue(filter.pass(ip("2001:0718:1803:0044:800A:0C00:00D0:0000")));
            assertTrue(filter.pass(ip("2001:0718:1803:0044:ffff:ffff:ffff:ffff")));

            assertFalse(filter.pass(decrement(ip6("2001:0718:1803:0044:8000:0000:0000:0000"))));
            assertFalse(filter.pass(increment(ip6("2001:0718:1803:0045:0000:0000:0000:ffff"))));
        }
        {
            FilterMask filter = FilterMask.create("2001:718:1803:44:aa20:66ff:fe59:eb0a", "10");
            
            assertTrue(filter.pass(ip("2000:0000:0000:0000:0000:0000:0000:0000")));
            assertTrue(filter.pass(ip("2001:0718:1803:0044:800A:0C00:00D0:0000")));
            assertTrue(filter.pass(ip("203f:ffff:ffff:ffff:ffff:ffff:ffff:ffff")));

            assertFalse(filter.pass(decrement(ip6("2000:0000:0000:0000:0000:0000:0000:0000"))));
            assertFalse(filter.pass(increment(ip6("203f:ffff:ffff:ffff:ffff:ffff:ffff:ffff"))));
        }
        {
            FilterMask filter = FilterMask.create("2001:718:1803:44:aa20:66ff:fe59:eb0a", "126");
            
            assertTrue(filter.pass(ip("2001:0718:1803:0044:aa20:66ff:fe59:eb08")));
            assertTrue(filter.pass(ip("2001:0718:1803:0044:aa20:66ff:fe59:eb09")));
            assertTrue(filter.pass(ip("2001:0718:1803:0044:aa20:66ff:fe59:eb0b")));

            assertFalse(filter.pass(decrement(ip6("2001:0718:1803:0044:aa20:66ff:fe59:eb08"))));
            assertFalse(filter.pass(increment(ip6("2001:0718:1803:0044:aa20:66ff:fe59:eb0b"))));
        }
    }

    private InetAddress increment(InetAddress address) throws UnknownHostException {
        return InetAddress.getByAddress(new BigInteger(1, address.getAddress()).add(BigInteger.ONE).toByteArray());
    }

    private InetAddress decrement(InetAddress address) throws UnknownHostException {
        return InetAddress.getByAddress(new BigInteger(1, address.getAddress()).subtract(BigInteger.ONE).toByteArray());
    }


}
