/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.external;

import com.kerio.im.bean.VCard;
import com.kerio.im.bean.VCard.VCardFactory;
import java.util.Set;

/**
 *
 * @author tskoda
 */
public interface ExtService {

    /**
     * Get set containing all domains managed by external service.
     *
     * @return never null
     */
    void getAllDomains(DomainHandler handler);

    /**
     * Get all users from provided domain.
     *
     * @domain users domain
     * @handler handler called for every external user
     * @return never null
     */
    void getAllUsers(String domain, UserHandler handler);

    /**
     * Return all ports configured for service id.
     *
     * @param serviceId identifier of xmpp service
     * @param handler handle called for every configured port
     */
    void getAllPorts(String serviceId, ServiceHandler handler);

    /**
     * Return access limits for service
     *
     * @param serviceId service identifier
     * @param handler handler for found access limits
     */
    void getPortAccess(String serviceId, AccessHandler handler);

    /**
     * Authenticate user.
     *
     * @param string user name in form login@domain
     * @param password user password
     * @param serviceId service identifier
     * @param connectionId connection identifier
     * @param ip local service ip address
     * @param port remote service port
     * @param secure if user is connected by secured channel
     * @return true only if user is authenticated
     */
    boolean auth(String username, byte[] password, String serviceId, String connectionId, boolean secure);
    /**
     * Called wen user/server connected to xmpp.
     * @param serviceId
     * @param connectionId
     * @param ip
     * @param port
     * @param secure 
     * 
     * @return true if connection is allowed, oterwise false and connection must be refused
     */
    boolean connect(String serviceId, String connectionId, String ip, int port, boolean secure);
    
    /**
     * Called when user/server connection ends.
     * @param serviceId service identifier
     * @param connectionId connection identifier
     */
    void disconnect(String serviceId, String connectionId);

    /**
     * Get security settings.
     * @return 
     */
    void getSecurity(SecurityAccessHandler handler);

    /**
     * Get all pairs of public and private certificate keys.
     */
    void getServerCertificates(CertificateHandler handler);
    
    Settings getSettings();
    
    ArchiveSettings getArchiveSettings();

    /**
     * Fills/remove/update all vcards from passed domain.
     *
     * @param syncKey last sync key (marks point in vcard history), if null full
     * sync is executed
     * @param domain
     * @param factory
     * @param cache
     * @return new sync key
     */
    long getAllVCards(long syncKey, String domain, VCardCache cache, VCardFactory factory);
    
    class Settings {
        
        private final boolean sendOutsideEnabled;

        public Settings(boolean sendOutsideEnabled) {
            this.sendOutsideEnabled = sendOutsideEnabled;
        }

        public boolean isSendOutsideEnabled() {
            return sendOutsideEnabled;
        }        
    }
    
    interface ArchiveSettings {
        String getPath();
        boolean isEnabled();
    }

    interface VCardCache {

        boolean containsId(String itemId);

        VCard getById(String itemId);

        void clear();

        void removeById(String itemId);

        /**
         *
         * @param vcard filled vcard
         * @param itemId contact id
         * @param username login@fomain
         */
        void put(VCard vcard, String itemId, String username);

        int count();
        
        void setSource(String source);
        String getSource();
    }
    
    interface CertificateHandler {
        
        /**
         * 
         * @param id
         * @return True if further process this certificate, false if skip.
         */
        boolean process(String id);
        
        /**
         * 
         * @param id
         * @param aliases
         * @param certificate
         * @param key
         * @param defaultCertificate
         * @param valid
         * @return True if stop processing certificates.
         */
        boolean handle(
                String id, 
                Set<String> aliases, 
                String certificate, 
                String key,
                boolean defaultCertificate,
                boolean valid); 
                
    }

    interface DomainHandler {

        void handle(String id, String name);
    }

    interface UserHandler {

        void handle(String login, String fullname, boolean enabled, boolean inGal, boolean domainRestriction);
    }

    interface ServiceHandler {

        static enum TYPE {

            all,
            localhost,
            address
        }

        void port(int port, TYPE type, String address, String serviceName);
        AccessHandler getAccess();
    }
    
    interface SecurityAccessHandler extends AccessHandler {
        
        void exceptions(boolean hasExceptions);
        
        void required(boolean tlsRequired);
    }

    interface AccessHandler {
        
        /**
         * Used as a notification about unknown or corrupted filter.
         */
        void filterUnknown();

        void filterRange(String from, String to, boolean enabled);

        void filterMask(String net, String mask, boolean enabled);

        void filterLocalhost(boolean enabled);

        void filterHost(String ip, boolean enabled);
        
        void limit(int count);
    }
}
