/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.connect;

import com.kerio.im.common.InetAddressUtil;
import com.kerio.im.common.error.Code;
import com.kerio.im.common.error.ServiceException;
import com.kerio.im.external.ExtService;
import com.kerio.lib.json.api.connect.admin.iface.Archive;
import com.kerio.lib.json.api.connect.admin.iface.ConnectCertificate;
import com.kerio.lib.json.api.connect.admin.iface.Domains;
import com.kerio.lib.json.api.connect.admin.iface.InstantMessaging;
import com.kerio.lib.json.api.connect.admin.iface.InstantMessaging.GetResult;
import com.kerio.lib.json.api.connect.admin.iface.SecurityPolicy;
import com.kerio.lib.json.api.connect.admin.iface.Services;
import com.kerio.lib.json.api.connect.admin.iface.Session;
import com.kerio.lib.json.api.connect.admin.iface.Users;
import com.kerio.lib.json.api.connect.admin.iface.Users.AuthenticateConnectionFromExternalServiceResult;
import com.kerio.lib.json.api.connect.admin.iface.Users.ConnectFromExternalServiceResult;
import com.kerio.lib.json.api.connect.admin.iface.common.IpAddressGroups;
import com.kerio.lib.json.api.connect.admin.struct.AddressType;
import com.kerio.lib.json.api.connect.admin.struct.AuthResult;
import com.kerio.lib.json.api.connect.admin.struct.Domain;
import com.kerio.lib.json.api.connect.admin.struct.Listener;
import com.kerio.lib.json.api.connect.admin.struct.SecurityPolicyOptions;
import com.kerio.lib.json.api.connect.admin.struct.Service;
import com.kerio.lib.json.api.connect.admin.struct.User;
import com.kerio.lib.json.api.connect.admin.struct.common.ApiApplication;
import com.kerio.lib.json.api.connect.admin.struct.common.CertificateType;
import com.kerio.lib.json.api.connect.admin.struct.common.NamedMultiValue;
import com.kerio.lib.json.api.connect.admin.struct.common.NamedValue;
import com.kerio.lib.json.api.connect.admin.struct.common.SearchQuery;
import java.net.URL;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.codec.binary.Base64;

/**
 *
 * @author tskoda
 */
public class AdminApi extends ApiClient {

    private static final String API_PATH = "/admin/api/jsonrpc/";
    public static final Integer PORT_DEFAULT = 4040;
    private static final Logger LOG = Logger.getLogger(AdminApi.class.getName());

    private static class ArchiveSettingsIml implements ExtService.ArchiveSettings {

        private final String path;
        private final boolean enabled;

        public ArchiveSettingsIml(String path, boolean enabled) {
            this.path = path;
            this.enabled = enabled;
        }

        @Override
        public String getPath() {
            return this.path;
        }

        @Override
        public boolean isEnabled() {
            return this.enabled;
        }

    }

    private static class PortErrorHandler implements AddressFilter.ErrorHandler {

        private final String serviceName;
        private final ExtService.AccessHandler handler;

        public PortErrorHandler(ExtService.AccessHandler handler, String serviceName) {
            this.handler = handler;
            this.serviceName = serviceName;
        }

        @Override
        public void groupNotFound(String groupId) throws ServiceException {
            handler.filterUnknown();
            LOG.log(Level.INFO, "IP access limit of {0} service incomplete, unknown IP group '{1}' specified. Service may be accessible to less IP addresses than intended.", new Object[]{this.serviceName, groupId});
            LOG.log(Level.WARNING, "IP access limit of {0} service incomplete, unknown IP group '{1}' specified. Service may be accessible to less IP addresses than intended.", new Object[]{this.serviceName, groupId});
        }
    }

    private static class SecurityErrorHandler implements AddressFilter.ErrorHandler {

        private final ExtService.SecurityAccessHandler handler;

        public SecurityErrorHandler(ExtService.SecurityAccessHandler handler) {
            this.handler = handler;
        }

        @Override
        public void groupNotFound(String groupId) throws ServiceException {
            handler.filterUnknown();
            LOG.log(Level.INFO, "Security policy exception incomplete, unknown IP group '{0}' specified. Unsecure connection/authentication may be allowed to less IP addresses than intended.", new Object[]{groupId});
            LOG.log(Level.WARNING, "Security policy exception incomplete, unknown IP group '{0}' specified. Unsecure connection/authentication may be allowed to less IP addresses than intended.", new Object[]{groupId});
        }
    }

    private final ApiApplication app;
    private final String username;
    private final String password;

    public AdminApi(String username, String password, String server, Integer port, Boolean secure) {
        super(ApiClient.hostname(server, port, secure), API_PATH);
        this.app = new ApiApplication();
        this.app.setName("Instant Messaging Admin API client");
        this.app.setVendor("Kerio Technologies, inc");
        this.app.setVersion("1.0");

        this.username = username;
        this.password = password;
    }

    @Override
    void login() {
        LOG.log(Level.FINE, "Connecting to ADMIN API on {0}", this.getSession().getHostname());
        getApi(Session.class).login(this.username, this.password, this.app);
    }

    @Override
    void logout() {
        getApi(Session.class).logout();
    }

    @Override
    String getName() {
        return this.app.getName();
    }

    public Runnable getPing() {
        return new AdminApiPing(this);
    }

    public boolean isGalGlobal() {
        return !this.getApi(Domains.class).getSettings().getSetting().getPublicFoldersPerDomain();
    }

    public void getServerCertificates(ExtService.CertificateHandler handler) {

        try {
            ConnectCertificate certApi = getApi(ConnectCertificate.class);
            ConnectCertificate.GetResult getResult = certApi.get();
            for (com.kerio.lib.json.api.connect.admin.struct.common.Certificate connectCertificate : getResult.getCertificates()) {
                if (isValidType(connectCertificate) && handler.process(connectCertificate.getId())) {
                    String publicCertUrl = certApi.exportCertificate(connectCertificate.getId()).getFileDownload().getUrl();
                    String privateCertUrl = certApi.exportPrivateKey(connectCertificate.getId()).getFileDownload().getUrl();

                    Set<String> names = new HashSet<>();

                    for (NamedValue subject : connectCertificate.getSubject()) {
                        if (subject.getName() != null && subject.getName().equals("hostname") && subject.getValue() != null) {
                            names.add(subject.getValue());
                        }
                    }

                    for (NamedMultiValue subject : connectCertificate.getSubjectAlternativeNameList()) {
                        if (subject.getName() != null && subject.getName().equals("hostname") && subject.getValue() != null) {
                            names.addAll(Arrays.asList(subject.getValue()));
                        }
                    }
                    
                    boolean next = handler.handle(
                            connectCertificate.getId(),
                            names,
                            getFileDownload().asString(InetAddressUtil.combineURL(getHostname(), publicCertUrl)), 
                            getFileDownload().asString(InetAddressUtil.combineURL(getHostname(), privateCertUrl)),
                            connectCertificate.getType() == CertificateType.ServerCertificate,
                            connectCertificate.getValid());
                    
                    if (!next) {
                        break;
                    }
                    
                }
            }
        } catch (Exception e) {
            throw new ServiceException("Failed to get server certificate", e, Code.EXT_SERVICE_ERROR);
        }

    }
    
    private boolean isValidType(com.kerio.lib.json.api.connect.admin.struct.common.Certificate connectCertificate) {
        boolean result = (connectCertificate != null && connectCertificate.getType() != null);
        
        if (result) {
            
            CertificateType type = connectCertificate.getType();
            result = (type == CertificateType.ActiveCertificate
                    || type == CertificateType.ServerCertificate);
        }
        
        return result;
    }

    public void getAllDomains(ExtService.DomainHandler handler) {
        //LOG.log(Level.FINE, "Loading all domains using admin api");

        try {
            Domains domainsApi = getApi(Domains.class);
            SearchQuery query = new SearchQuery();

            query.setFields(
                    new String[]{"id", "name"});
            //TODO filter out aliases
            Domains.GetResult getResult = domainsApi.get(query);
            for (Domain connectDomain : getResult.getList()) {
                handler.handle(connectDomain.getId(), connectDomain.getName());
            }
        } catch (Exception ex) {
            throw new ServiceException("Loading all domains using admin api failed, received unsupported JSON response", ex, Code.EXT_SERVICE_ERROR);
        }

        //LOG.log(Level.INFO, "Loading all domains using admin api successful");
    }

    public boolean auth(String username, byte[] password, String serviceId, String connectionId, boolean secure) {

        boolean result = false;
        try {
            Users usersApi = getApi(Users.class);
            String password64 = Base64.encodeBase64String(password);
            AuthenticateConnectionFromExternalServiceResult loginResult = usersApi.authenticateConnectionFromExternalService(
                    username,
                    password64,
                    serviceId,
                    connectionId,
                    secure);

            result = loginResult.getResult() == AuthResult.AuthOK;
        } catch (Exception e) {
            throw new ServiceException("Authentication user {0} failed, received unsupported JSON response", e, Code.EXT_SERVICE_ERROR, username);
        }

        return result;
    }

    public boolean connect(String serviceId, String connectionId, String ip, int port, boolean secure) {
        try {
            ConnectFromExternalServiceResult connectFromExternalService = getApi(Users.class).connectFromExternalService(serviceId, connectionId, ip, port, secure);
            return connectFromExternalService.getResult();
        } catch (Exception e) {
            throw new ServiceException("Disconnection connection {0} failed", e, Code.EXT_SERVICE_ERROR, connectionId);
        }
    }

    public void disconnect(String serviceId, String connectionId) {
        try {
            getApi(Users.class).disconnectFromExternalService(serviceId, connectionId);
        } catch (Exception e) {
            throw new ServiceException("Disconnection connection {0} failed", e, Code.EXT_SERVICE_ERROR, connectionId);
        }
    }

    public void getAllUsers(String domainId, ExtService.UserHandler handler) {

        try {
            Users usersApi = getApi(Users.class);
            SearchQuery query = new SearchQuery();

            query.setFields(
                    new String[]{"loginName", "fullName", "isEnabled", "hasDomainRestriction", "publishInGal"});

            Users.GetResult getResult = usersApi.get(query, domainId);
            for (User user : getResult.getList()) {
                handler.handle(user.getLoginName().toLowerCase(), user.getFullName(), user.getIsEnabled(), user.getPublishInGal(), !user.getHasDomainRestriction()); //handler has to get value if user can send outsideOfOwnDomain
            }
        } catch (Exception e) {
            throw new ServiceException("Reading all users from connect failed, " + e.getMessage(), e, Code.RESOURCE_ERROR);
        }
    }

    public void getAllPorts(String serviceId, ExtService.ServiceHandler handler) {
        Services.GetIPv6Result ipv6response = getApi(Services.class).getIPv6();
        Boolean ipv6enabled = ipv6response.getIsEnabled() && !Boolean.getBoolean("java.net.preferIPv4Stack");

        Services.GetResult response = getApi(Services.class).get();
        Service[] services = response.getServices();
        AddressFilter addressFilter = AddressFilter.create(this.getApi(IpAddressGroups.class));
        for (final Service service : services) {
            if (service.getId().equals(serviceId)) {

                for (Listener listener : service.getListeners()) {
                    //pass alladdresses always
                    //pass ipv6 address/localhost only if ipv6 support is enabled
                    boolean isipv4 = InetAddressUtil.isIPv4(listener.getAddress());
                    if (listener.getType() == AddressType.AllAddresses
                            || ipv6enabled
                            || InetAddressUtil.isIPv4(listener.getAddress())) {

                        handler.port(listener.getPort(), asType(listener.getType()), listener.getAddress(), service.getName());
                    } else {
                        LOG.log(Level.FINE, "Skipping IPv6 address {0}:{1} of {2} service, IPv6 stack is globally disabled", new Object[]{listener.getAddress(), listener.getPort(), service.getName()});
                    }
                }

                //pass all configured filters
                addressFilter.apply(service, handler.getAccess(), new PortErrorHandler(handler.getAccess(), service.getName()));

                //pass connection limit
                applyLimit(service, handler.getAccess());

                //we expect only one service with desired identifier
                break;
            }
        }
    }

    public void getPortAccess(String serviceId, ExtService.AccessHandler handler) {
        Services.GetResult response = getApi(Services.class).get();
        Service[] services = response.getServices();
        AddressFilter addressFilter = AddressFilter.create(this.getApi(IpAddressGroups.class));
        for (Service service : services) {
            if (service.getId().equals(serviceId)) {
                //pass all configured filters
                addressFilter.apply(service, handler, new PortErrorHandler(handler, service.getName()));
                applyLimit(service, handler);
            }
        }
    }

    public void getSecurity(ExtService.SecurityAccessHandler handler) {
        SecurityPolicyOptions options = getApi(SecurityPolicy.class).get().getOptions();

        //tls is required without exceptions by default
        String exceptionGroupId = null;
        boolean tlsRequired = true;

        switch (options.getMode()) {
            case SPAuthenticationRequired:
                if (options.getAuthenticationExceptionGroup().getEnabled()) {
                    exceptionGroupId = options.getAuthenticationExceptionGroup().getId();
                }
                tlsRequired = true;
                break;
            case SPEncryptionRequired:
                if (options.getEncryptionExceptionGroup().getEnabled()) {
                    exceptionGroupId = options.getEncryptionExceptionGroup().getId();
                }

                tlsRequired = true;
                break;
            case SPNoRestrictions:
                tlsRequired = false;
                break;
            default:
                LOG.log(Level.WARNING, "Unsupported security option {0}, forcing TLS without exceptions", options.getMode());
        }

        handler.required(tlsRequired);

        if (tlsRequired && (exceptionGroupId != null)) {
            handler.exceptions(true);
            AddressFilter addressFilter = AddressFilter.create(this.getApi(IpAddressGroups.class));
            addressFilter.apply(exceptionGroupId, handler, new SecurityErrorHandler(handler));
        }
    }

    public ExtService.Settings getSettings() {
        GetResult settingsResult = getApi(InstantMessaging.class).get();
        ExtService.Settings result = new ExtService.Settings(settingsResult.getSettings().getSendOutsideEnabled());
        return result;

    }

    ExtService.ArchiveSettings getArchiveSettings() {
        Archive.GetResult archiveResult = getApi(Archive.class).get();
        if ((archiveResult.getOptions() != null) && (archiveResult.getOptions().getPaths() != null)) {
            return new ArchiveSettingsIml(archiveResult.getOptions().getPaths().getArchivePath(), archiveResult.getOptions().getIsXmppEnabled());
        } else {
            throw new ServiceException("Reading archive settings failed, required attributes are empty", Code.RESOURCE_ERROR);
        }
    }

    private void applyLimit(Service service, ExtService.AccessHandler handler) {
        //pass connection limit
        if (service.getConnectionLimit() != null && service.getConnectionLimit().getIsSet()) {
            handler.limit(service.getConnectionLimit().getValue().intValue());
        } else {
            handler.limit(0);
        }
    }

    private ExtService.ServiceHandler.TYPE asType(AddressType addressType) {
        ExtService.ServiceHandler.TYPE result;
        switch (addressType) {
            case AllAddresses:
                result = ExtService.ServiceHandler.TYPE.all;
                break;
            case Localhost:
                result = ExtService.ServiceHandler.TYPE.localhost;
                break;
            case RealIpAddress:
                result = ExtService.ServiceHandler.TYPE.address;
                break;
            default:
                result = ExtService.ServiceHandler.TYPE.localhost;
        }
        return result;
    }

    static class AdminApiPing implements Runnable {

        private static final Logger LOG = Logger.getLogger(AdminApiPing.class.getName());
        private final ApiClient apiClient;

        private AdminApiPing(AdminApi apiClient) {
            this.apiClient = apiClient;
        }

        @Override
        public void run() {
            LOG.log(Level.FINE, "Admin API session ping");
            this.apiClient.getApi(Session.class).whoAmI();
        }
    }
}
