/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.connect;

import com.kerio.im.bean.Certificate;
import com.kerio.im.bean.VCard;
import com.kerio.im.common.Threads;
import com.kerio.im.common.TypedProperties;
import com.kerio.im.common.error.Code;
import com.kerio.im.common.error.ServiceException;
import com.kerio.im.external.ExtService;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.jabsorb.client.ErrorResponse;

/**
 *
 * @author tskoda
 */
public class ExtServiceImpl implements ExtService {

    private static final Logger LOG = Logger.getLogger(ExtServiceImpl.class.getName());
    private static final int PING_PERIOND_DEFAULT_SECONDS = 60 * 10;
    private static final String SERVER_DEFAULT = "localhost";
    public static final int WEBMAILAPI_PORT_DEFAULT = 80;
    public static final int ADMINLAPI_PORT_DEFAULT = 4040;
    public static final boolean UNSECURE = false;
    private final AdminApi adminApi;
    private final WebmailApi webmailApi;
    private final ScheduledExecutorService scheduler;

    public ExtServiceImpl(Map<String, String> specific) {
        this.scheduler = Executors.newScheduledThreadPool(1, Threads.create("[ext service] session keep alive"));

        String username = TypedProperties.string(ConnectKey.SERVER_ADMIN_KEY, specific);
        String password = TypedProperties.string(ConnectKey.SERVER_ADMIN_PASS_KEY, specific);
        Integer interval = TypedProperties.integer(ConnectKey.PING_PERIOD_KEY, specific, PING_PERIOND_DEFAULT_SECONDS);

        this.webmailApi = new WebmailApi(
                username,
                password,
                TypedProperties.string(ConnectKey.WEBMAILAPI_IP, specific, SERVER_DEFAULT),
                TypedProperties.integer(ConnectKey.WEBMAILAPI_PORT, specific, WEBMAILAPI_PORT_DEFAULT),
                TypedProperties.bool(ConnectKey.WEBMAILAPI_SECURE, specific, UNSECURE));

        this.webmailApi.login();
        LOG.log(Level.INFO, "Connecting to Kerio Connect WEBMAIL API on {0} successful", this.webmailApi.getSession().getUrl());
        this.adminApi = new AdminApi(
                username,
                password,
                TypedProperties.string(ConnectKey.ADMINAPI_IP, specific, SERVER_DEFAULT),
                TypedProperties.integer(ConnectKey.ADMINAPI_PORT, specific, ADMINLAPI_PORT_DEFAULT),
                UNSECURE);

        this.adminApi.login();
        LOG.log(Level.INFO, "Connecting to Kerio Connect ADMIN API on {0} successful", this.adminApi.getSession().getUrl());

        this.scheduler.scheduleAtFixedRate(this.webmailApi.getPing(), interval, interval, TimeUnit.SECONDS);
        this.scheduler.scheduleAtFixedRate(this.adminApi.getPing(), interval, interval, TimeUnit.SECONDS);
    }

    @Override
    public long getAllVCards(long syncKey, String domain, VCardCache cache, VCard.VCardFactory factory) {
        long result;
        try {
            result = this.webmailApi.getAllVCards(this.adminApi.isGalGlobal(), syncKey, domain, cache, factory);
        } catch (ErrorResponse e) {
            LOG.log(Level.INFO, "Failed to read vcards, {0}", e.getMessage()); //do not log this to warning log, it is not exceptional state
            result = syncKey;
        } catch (Exception e) {
            throw new ServiceException("Failed to read vcards", e, Code.EXT_SERVICE_ERROR);
        }
        
        return result;
    }

    @Override
    public void getSecurity(SecurityAccessHandler handler) {
        this.adminApi.getSecurity(handler);
    }
    
    @Override
    public Settings getSettings() {
        return this.adminApi.getSettings();
    }
    
    @Override
    public void getServerCertificates(CertificateHandler handler) {
        this.adminApi.getServerCertificates(handler);
    }

    @Override
    public void getAllDomains(DomainHandler handler) {
        this.adminApi.getAllDomains(handler);
    }

    @Override
    public boolean auth(String username, byte[] password, String serviceId, String connectionId, boolean secure) {
        return this.adminApi.auth(username, password, serviceId, connectionId, secure);
    }

    @Override
    public boolean connect(String serviceId, String connectionId, String ip, int port, boolean secure) {
        return this.adminApi.connect(serviceId, connectionId, ip, port, secure);
    }
    
    @Override
    public void disconnect(String serviceId, String connectionId) {
        this.adminApi.disconnect(serviceId, connectionId);
    }

    @Override
    public void getAllUsers(String domainId, UserHandler handler) {
        this.adminApi.getAllUsers(domainId, handler);
    }

    @Override
    public void getAllPorts(String serviceId, ServiceHandler handler) {
        this.adminApi.getAllPorts(serviceId, handler);
    }

    @Override
    public void getPortAccess(String serviceId, AccessHandler handler) {
        this.adminApi.getPortAccess(serviceId, handler);
    }
    
    public void bye() {
        try {
            this.scheduler.shutdownNow();
        } catch (Exception e) {
            LOG.log(Level.INFO, "Failed to close scheduled tasks", e);
        }

        try {
            this.webmailApi.logout();
        } catch (Exception e) {
            LOG.log(Level.INFO, "Failed to close connection to Webmail API", e);
        }

        try {
            this.adminApi.logout();
        } catch (Exception e) {
            LOG.log(Level.INFO, "Failed to close connection to Admin API", e);
        }
    }

    @Override
    public ArchiveSettings getArchiveSettings() {
        return this.adminApi.getArchiveSettings();
    }
}
