package com.kerio.im.im;

import com.kerio.lib.json.api.client.KJsonRpcSession;
import com.kerio.lib.json.api.client.KParameterName;
import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;

import org.jabsorb.JSONRPCResult;
import org.jabsorb.JSONSerializer;
import org.jabsorb.client.ClientError;
import org.jabsorb.serializer.SerializerState;
import org.jabsorb.serializer.impl.BooleanSerializer;
import org.jabsorb.serializer.impl.DateSerializer;
import org.jabsorb.serializer.impl.NumberSerializer;
import org.jabsorb.serializer.impl.PrimitiveSerializer;
import org.jabsorb.serializer.impl.RawJSONObjectSerializer;
import org.jabsorb.serializer.impl.StringSerializer;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.kerio.lib.json.api.client.serializer.KArraySerializer;
import com.kerio.lib.json.api.client.serializer.KBeanSerializer;
import com.kerio.lib.json.api.client.serializer.KEnumSerializer;
import org.jabsorb.client.ErrorResponse;
import org.jabsorb.serializer.MarshallException;
import org.jabsorb.serializer.UnmarshallException;

public class ImJsonRpcClient implements InvocationHandler {

    private static final String JSON_RPC_VERSION = "2.0";
    private static final RequestCounter REQUEST_COUNTER = new RequestCounter();

    private final Logger logger = LoggerFactory.getLogger(ImJsonRpcClient.class);
    protected JSONSerializer serializer = new JSONSerializer();
    private KJsonRpcSession session;
    private boolean parametersInArray;
    private Integer lastRequestId;
    private boolean lastRequestIdSet = false;

    public ImJsonRpcClient(KJsonRpcSession session) {
        this.session = session;

        if (session == null) {
            throw new IllegalArgumentException("Session is not initialized.");
        }

        try {
            this.serializer.registerSerializer(new DateSerializer());
            this.serializer.registerSerializer(new StringSerializer());
            this.serializer.registerSerializer(new NumberSerializer());
            this.serializer.registerSerializer(new BooleanSerializer());
            this.serializer.registerSerializer(new PrimitiveSerializer());
            this.serializer.registerSerializer(new RawJSONObjectSerializer());

            /* problem with 'fixup' */
            // this.serializer.registerSerializer(new RawJSONArraySerializer());
            // this.serializer.registerSerializer(new DictionarySerializer());
            // this.serializer.registerSerializer(new MapSerializer());
            // this.serializer.registerSerializer(new SetSerializer());
            // this.serializer.registerSerializer(new ListSerializer());
            this.serializer.registerSerializer(new KBeanSerializer());
            this.serializer.registerSerializer(new KArraySerializer());
            this.serializer.registerSerializer(new KEnumSerializer());

            this.serializer.setMarshallClassHints(false);
            this.serializer.setMarshallNullAttributes(false);
            this.serializer.setFixupCircRefs(false);
            this.serializer.setFixupDuplicatePrimitives(false);
            this.serializer.setFixupDuplicates(false);
        } catch (Exception e) {
            throw new ClientError(e);
        }
    }

    public <T> T getProxy(Class<T> clazz) {
        Object result = java.lang.reflect.Proxy.newProxyInstance(clazz.getClassLoader(), new Class[]{clazz}, this);
        return clazz.cast(result);
    }

    // InvocationHandler
    @Override
    public Object invoke(Object proxyObj, Method method, Object[] args) throws UnmarshallException, JSONException, MarshallException {
        String methodName = method.getName();

        switch (methodName) {
            case "hashCode":
                return System.identityHashCode(proxyObj);
            case "equals":
                return (proxyObj == args[0] ? Boolean.TRUE : Boolean.FALSE);
            case "toString":
                return proxyObj.getClass().getName() + '@' + Integer.toHexString(proxyObj.hashCode());
            default:
                break;
        }

        return invokeImpl(method, args);
    }

    private Object invokeImpl(Method method, Object[] args) throws UnmarshallException, JSONException, MarshallException {
        JSONObject message = new JSONObject();

        String methodTag = method.getDeclaringClass().getSimpleName() + "." + method.getName();
        message.put("method", methodTag);

        if (this.parametersInArray) {
            if (args != null) {
                SerializerState state = new SerializerState();
                JSONArray params = (JSONArray) this.serializer.marshall(state, null, args, "params");
                message.put("params", params);
            } else {
                message.put("params", new JSONArray());
            }
        } else if (args != null) {
            JSONObject params = new JSONObject();
            List<String> paramNames = lookupParameterNames(method);

            if (paramNames.size() != args.length) {
                throw new IllegalStateException("Some parameter has no name annotation.");
            }

            SerializerState state = new SerializerState();

            for (int i = 0; i < paramNames.size(); ++i) {
                Object value = this.serializer.marshall(state, null, args[i], "result");
                if (value != null) {
                    params.put(paramNames.get(i), value);
                }

            }

            message.put("params", params);
        } else {
            message.put("params", new JSONObject());
        }

        message.put("jsonrpc", JSON_RPC_VERSION);
        message.put("id", getId());

        Object rawResult = processJsonRequest(message, method);

        Class<?> returnType = method.getReturnType();

        if (returnType.equals(Void.TYPE)) {
            return null;
        }

        return this.serializer.unmarshall(new SerializerState(), returnType, rawResult);
    }

    public Object processJsonRequest(JSONObject message, Method method) throws JSONException {
        this.logger.debug("JSON Query:\n" + message.toString());

        try {
            JSONObject responseMessage = this.session.sendAndReceive(message, method);

            this.logger.debug("JSON Response:\n" + responseMessage);

            if (!responseMessage.has("result")) {
                processException(responseMessage, 200);
            }

            Object rawResult = responseMessage.get("result");

            if (rawResult == null) {
                processException(responseMessage, 200);
            }

            return rawResult;
        } catch (JsonErrorException e) {
            processException(e.getMsg(), e.getHttpresultCode());
        }
        return new Object();
    }

    private void processException(JSONObject responseMessage, int httpResult) throws JSONException {
        JSONObject error = responseMessage.has("error") ? (JSONObject) responseMessage.get("error") : null;

        Integer code;
        String msg;
        String trace;

        if (error != null) {
            code = error.has("code") ? error.getInt("code") : 0;
            msg = error.has("message") ? error.getString("message") : null;

            if (msg == null) {
                msg = error.has("msg") ? error.getString("msg") : null;
            }

            if (error.has("data")) {
                JSONObject data = error.getJSONObject("data");

                if (data.has("messageParameters") && !data.isNull("messageParameters")) {
                    JSONObject messageParameters = data.getJSONObject("messageParameters");

                    if (messageParameters.has("positionalParameters")) {
                        JSONArray positionalParameters = messageParameters.getJSONArray("positionalParameters");

                        for (int i = 0; i < positionalParameters.length(); ++i) {
                            String param = positionalParameters.getString(i);
                            msg = msg.replaceAll(String.format("%%%d", i + 1), param);
                        }
                    }
                }
            }

            trace = error.has("trace") ? error.getString("trace") : null;
        } else {
            code = JSONRPCResult.CODE_ERR_PARSE;
            msg = "Unknown http result " + httpResult + " response:" + responseMessage.toString(2);
            trace = null;
        }

        throw new ErrorResponse(code, msg, trace);
    }

    public int generateRequestId() {
        return REQUEST_COUNTER.getId();
    }

    private int getId() {
        if (lastRequestIdSet) {
            lastRequestIdSet = false;
            return lastRequestId;
        } else {
            lastRequestId = REQUEST_COUNTER.getId();
        }
        return lastRequestId;
    }

    public void setRequestId(int id) {
        lastRequestId = id;
        lastRequestIdSet = true;
    }

    public int getLastRequestId() {
        return lastRequestId;
    }

    private List<String> lookupParameterNames(Method method) {
        List<String> ret = new ArrayList<>();
        Annotation[][] parameterAnnotations = method.getParameterAnnotations();

        for (Annotation[] annotations : parameterAnnotations) {
            for (Annotation annotation : annotations) {
                if (annotation instanceof KParameterName) {
                    KParameterName kParameterName = (KParameterName) annotation;
                    ret.add(kParameterName.value());
                }
            }
        }

        return ret;
    }

    public void setMarshallNullAttributes(boolean marshallNullAttributes) {
        this.serializer.setMarshallNullAttributes(marshallNullAttributes);
    }

    public boolean isMarshallNullAttributes() {
        return this.serializer.getMarshallNullAttributes();
    }

    public void setParametersInArray(boolean parametersInArray) {
        this.parametersInArray = parametersInArray;
    }

    public boolean isParametersInArray() {
        return this.parametersInArray;
    }

    public JSONSerializer getSerializer() {
        return serializer;
    }
}
