/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.connect;

import com.kerio.lib.json.api.connect.admin.struct.Service;
import com.kerio.im.external.ExtService;
import com.kerio.lib.json.api.connect.admin.struct.GroupLimit;
import com.kerio.lib.json.api.connect.admin.struct.common.IpAddressEntry;
import com.kerio.lib.json.api.connect.admin.struct.common.IpAddressGroup;
import com.kerio.lib.json.api.connect.admin.struct.common.IpAddressGroupType;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.junit.Before;
import org.junit.Test;
import static org.mockito.Mockito.*;

/**
 *
 * @author tskoda
 */
public class AddressFilterTest {

    private Map<String, List<IpAddressEntry>> groups;
    private AddressFilter.ErrorHandler mockErrorHandler;
    private ExtService.AccessHandler mockAccessHandler;
    private GroupLimit groupLimit;

    @Before
    public void setUp() {
        this.groups = new HashMap<String, List<IpAddressEntry>>();
        this.mockErrorHandler = mock(AddressFilter.ErrorHandler.class);
        this.mockAccessHandler = mock(ExtService.AccessHandler.class);
        this.groupLimit = new GroupLimit();
    }

    @Test
    public void testApplyServiceGroupNotUsed() {

        this.groupLimit.setIsUsed(Boolean.FALSE);
        
        Service service = new Service();
        service.setGroup(groupLimit);

        AddressFilter af = new AddressFilter(groups);
        af.apply(service, mockAccessHandler, mockErrorHandler);
        
        verifyZeroInteractions(this.mockAccessHandler, this.mockErrorHandler);
    }
    
    @Test
    public void testApplyServiceNullRoot() {
        this.groupLimit.setIsUsed(Boolean.TRUE);
        
        Service service = new Service();
        service.setGroup(groupLimit);

        AddressFilter af = new AddressFilter(groups);
        af.apply(service, mockAccessHandler, mockErrorHandler);
        
        verifyZeroInteractions(mockAccessHandler);
        verify(mockErrorHandler, times(1)).groupNotFound((String)isNull());
    }
    
    @Test
    public void testApplyServiceUnknownRoot() {
        IpAddressGroup ipGroup = new IpAddressGroup();
        ipGroup.setId("this-is-unknown-id");
        ipGroup.setName("this-is-unknown-name");
        
        this.groupLimit.setIsUsed(Boolean.TRUE);
        this.groupLimit.setIpGroup(ipGroup);
        
        Service service = new Service();
        service.setGroup(groupLimit);

        AddressFilter af = new AddressFilter(groups);
        af.apply(service, mockAccessHandler, mockErrorHandler);
        
        verifyZeroInteractions(mockAccessHandler);
        verify(mockErrorHandler, times(1)).groupNotFound(eq(ipGroup.getId()));
    }
    
    @Test
    public void testApplyServiceNullSibling() {
        
        IpAddressEntry goodEntry = new IpAddressEntry();
        goodEntry.setType(IpAddressGroupType.Host);
        goodEntry.setId("good-entry-id");
        goodEntry.setEnabled(Boolean.TRUE);
        
        IpAddressEntry childNullEntry = new IpAddressEntry();
        childNullEntry.setType(IpAddressGroupType.ChildGroup);
        childNullEntry.setId(null);
        childNullEntry.setChildGroupId(null);
        childNullEntry.setEnabled(Boolean.TRUE);
        
        IpAddressEntry anotherGoodEntry = new IpAddressEntry();
        anotherGoodEntry.setType(IpAddressGroupType.Network);
        anotherGoodEntry.setId("another-good-entry-id");
        anotherGoodEntry.setEnabled(Boolean.TRUE);
        
        IpAddressGroup parentGroup = new IpAddressGroup();
        parentGroup.setId("parent-id");
        parentGroup.setName("parent-name");
        
        this.groups.put(parentGroup.getId(), Arrays.asList(goodEntry, childNullEntry, anotherGoodEntry));
        
        this.groupLimit.setIsUsed(Boolean.TRUE);
        this.groupLimit.setIpGroup(parentGroup);
        
        Service service = new Service();
        service.setGroup(groupLimit);

        AddressFilter af = new AddressFilter(groups);
        af.apply(service, mockAccessHandler, mockErrorHandler);
        
        verify(mockAccessHandler, times(1)).filterHost(anyString(), anyBoolean());
        verify(mockAccessHandler, times(1)).filterMask(anyString(), anyString(), anyBoolean());
        verify(mockErrorHandler, times(1)).groupNotFound((String)isNull());
    }
    
    @Test
    public void testApplyServiceUnknownSibling() {
        IpAddressEntry goodEntry = new IpAddressEntry();
        goodEntry.setType(IpAddressGroupType.Host);
        goodEntry.setId("good-entry-id");
        goodEntry.setEnabled(Boolean.TRUE);
        
        IpAddressEntry childNullEntry = new IpAddressEntry();
        childNullEntry.setType(IpAddressGroupType.ChildGroup);
        childNullEntry.setId("child-id");
        childNullEntry.setChildGroupId("child-does-not-exist-id");
        childNullEntry.setEnabled(Boolean.TRUE);
        
        IpAddressEntry anotherGoodEntry = new IpAddressEntry();
        anotherGoodEntry.setType(IpAddressGroupType.Network);
        anotherGoodEntry.setId("another-good-entry-id");
        anotherGoodEntry.setEnabled(Boolean.TRUE);
        
        IpAddressGroup parentGroup = new IpAddressGroup();
        parentGroup.setId("parent-id");
        parentGroup.setName("parent-name");
        
        this.groups.put(parentGroup.getId(), Arrays.asList(goodEntry, childNullEntry, anotherGoodEntry));
        
        this.groupLimit.setIsUsed(Boolean.TRUE);
        this.groupLimit.setIpGroup(parentGroup);
        
        Service service = new Service();
        service.setGroup(groupLimit);

        AddressFilter af = new AddressFilter(groups);
        af.apply(service, mockAccessHandler, mockErrorHandler);
        
        verify(mockAccessHandler, times(1)).filterHost(anyString(), anyBoolean());
        verify(mockAccessHandler, times(1)).filterMask(anyString(), anyString(), anyBoolean());
        verify(mockErrorHandler, times(1)).groupNotFound(eq(childNullEntry.getChildGroupId()));
    }

    @Test
    public void testApplyService() {

    }

}
