/*
 * Copyright 2015 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.methods;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.kerio.im.external.IMConversation;
import com.kerio.im.external.IMEvent;
import com.kerio.im.external.IMMessage;
import com.kerio.im.external.IMPresence;
import com.kerio.im.rpcserver.DataEnvelope;
import com.kerio.im.rpcserver.IEventsConsumer;
import com.kerio.im.rpcserver.ISubscription;
import com.kerio.im.rpcserver.ISubscriptionResolver;
import com.kerio.im.rpcserver.JsonMessage;
import com.kerio.im.rpcserver.JsonMessages;
import com.kerio.im.rpcserver.RpcError;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.kerio.im.rpcserver.RpcMethod;
import com.kerio.im.rpcserver.SessionManager;
import java.io.IOException;
import java.util.Map;

/**
 *
 * @author oleopold
 */
public class Publish implements RpcMethod {

    private static final Logger LOG = Logger.getLogger(Publish.class.getName());
    private final ObjectMapper mapper;
    private final IEventsConsumer eventsConsumer;
    private final ISubscriptionResolver subscribctionsResolver;

    public Publish(ObjectMapper mapper, IEventsConsumer eventsConsumer, ISubscriptionResolver subscribctionsResolver) {
        this.mapper = mapper;
        this.eventsConsumer = eventsConsumer;
        this.subscribctionsResolver = subscribctionsResolver;
    }

    @Override
    public String getMethod() {
        return "InstantMessagingServer.publish";
    }

    private void processEvents(DataEnvelope messageData, ISubscription subscription) throws InterruptedException {
        if (messageData != null && messageData.getResult() != null && messageData.getResult().getList() != null) {
            for (Map<String, Object> result : messageData.getResult().getList()) {
                try {
                    Object event = mapper.convertValue(result, subscription.getType());
                    if (event != null) {
                        eventsConsumer.enqueue(new IMEvent<>(event, subscription.getSubscriber()));
                    }
                } catch (IllegalArgumentException e) {
                    LOG.log(Level.WARNING, "Failed to map publish data to data class: ", e);
                    LOG.log(Level.FINE, "Type: {0}", subscription.getType().toString());
                    LOG.log(Level.FINE, "Data: {0}", messageData.getResult().toString());
                }
            }
        }
    }

    @Override
    public void execute(Request data, Response res, SessionManager sessionManager) {
        try {
            if (!sessionManager.verify(data.getSessionID(), data.getToken())) {
                res.error(new RpcError(1000, "Invalid Session ID or Token!"), 403);
                return;
            }

            JsonMessages messages = data.getContent(JsonMessages.class);
            if (messages != null && messages.getMessages() != null) {
                for (JsonMessage message : messages.getMessages()) {
                    DataEnvelope messageData = mapper.readValue(message.getContent(), DataEnvelope.class);
                    ISubscription subscription = subscribctionsResolver.getSubscription(messageData.getId());
                    if (subscription == null || subscription.getType() == null) {
                        //unknown
                    } else if (subscription.getType() == IMPresence.class || subscription.getType() == IMConversation.class || subscription.getType() == IMMessage.class) {
                        processEvents(messageData, subscription);
                    } else {
                        //WTF? log somethink
                    }
                }
            }
        } catch (IllegalArgumentException | IOException ex) {
            Logger.getLogger(Publish.class.getName()).log(Level.SEVERE, null, ex);
            LOG.log(Level.FINE, "Called method \"{0}\" with data: {1}", new Object[]{data.getContent().getMethod(), data.getContent().getParams().toString()});
        } catch (InterruptedException ex) {
            Logger.getLogger(Publish.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
}
