/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.rpcserver;

import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletResponse;

/**
 *
 * @author tskoda
 */
public class RpcResponseImpl implements RpcMethod.Response {

    private static final Logger LOG = Logger.getLogger(RpcResponseImpl.class.getName());
    private static final String CONTENT_TYPE = "application/json";
    private static final String ENCODING = "utf-8";

    private static abstract class CommonResponse {

        private String jsonrpc;
        private int id;

        public CommonResponse(String jsonrpc, int id) {
            this.jsonrpc = jsonrpc;
            this.id = id;
        }

        public String getJsonrpc() {
            return jsonrpc;
        }

        public void setJsonrpc(String jsonrpc) {
            this.jsonrpc = jsonrpc;
        }

        public int getId() {
            return id;
        }

        public void setId(int id) {
            this.id = id;
        }
    }

    private static class SuccessResponse<T> extends CommonResponse {

        private T result;

        public SuccessResponse(T result, String jsonrpc, int id) {
            super(jsonrpc, id);
            this.result = result;
        }

        public T getResult() {
            return result;
        }

        public void setResult(T result) {
            this.result = result;
        }
    }

    private static class ErrorResponse extends CommonResponse {

        private RpcError error;

        public ErrorResponse(RpcError error, String jsonrpc, int id) {
            super(jsonrpc, id);
            this.error = error;
        }

        public RpcError getError() {
            return error;
        }

        public void setError(RpcError error) {
            this.error = error;
        }
    }

    private final HttpServletResponse raw;
    private final int id;
    private final ObjectMapper mapper;

    public RpcResponseImpl(HttpServletResponse raw, int id, ObjectMapper mapper) {
        this.raw = raw;
        this.id = id;
        this.mapper = mapper;
    }

    public HttpServletResponse getRaw() {
        return raw;
    }

    @Override
    public void success() {
        response(new SuccessResponse<>(new HashMap<String, String>(), RpcConstants.JSONRPC, this.id), HttpServletResponse.SC_OK);
    }

    @Override
    public void success(Object content) {
        response(new SuccessResponse<>(content, RpcConstants.JSONRPC, this.id), HttpServletResponse.SC_OK);
    }

    @Override
    public void error(RpcError error, int httpCode) {
        response(new ErrorResponse(error, RpcConstants.JSONRPC, this.id), httpCode);
    }

    @Override
    public void error(int httpCode) {
        response(httpCode);
    }

    private void response(int httpCode) {
        this.raw.setCharacterEncoding(ENCODING);
        this.raw.setStatus(httpCode);

    }

    private void response(Object response, int httpCode) {
        this.response(httpCode);
        this.raw.setContentType(CONTENT_TYPE);
        try {
            OutputStream writer = this.raw.getOutputStream();
            this.mapper.writeValue(writer, response);
            writer.flush();
        } catch (IOException ex) {
            LOG.log(Level.SEVERE, "Failed to write success response", ex);
        }
    }

    @Override
    public void setSessionID(String sessionID) {
        Cookie c = new Cookie(RpcServlet.SESSION_ID_COOKIE_NAME, sessionID);
        c.setMaxAge(sessionID == null ? 0 : -1);
        c.setPath("/admin/");
        raw.addCookie(c);
    }       
}
