/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase;

import com.kerio.im.bean.Certificate;
import com.kerio.im.common.error.Code;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.core.api.service.ConfigurationService;
import com.kerio.im.tigase.derby.DerbyChecker;
import com.kerio.im.tigase.derby.DerbyHandler;
import com.kerio.im.tigase.derby.version.Version;
import com.kerio.im.xmpp.api.XmppDomains;
import com.kerio.im.xmpp.api.XmppException;
import java.io.File;
import java.security.cert.CertificateParsingException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.script.Bindings;
import javax.script.SimpleBindings;
import tigase.db.comp.ComponentRepository;
import tigase.db.comp.ConfigRepository;
import tigase.io.SSLContextContainerIfc;
import tigase.io.TLSUtil;
import tigase.util.TigaseStringprepException;
import tigase.vhosts.VHostItem;
import tigase.vhosts.VHostManager;

/**
 *
 * @author tskoda
 */
public class DomainServiceImpl implements XmppDomains {

    private static final Logger LOG = Logger.getLogger(DomainServiceImpl.class.getName());
    private static final String PEM_CERT = "%1$s\n%2$s";
    public static final String DB_ROOT = "database";
    /**
     * This is persistent virtual domain storage.
     */
    private final ConfigRepository<VHostItem> vHostRepository;
    /**
     * This is dynamic domain database storage.
     */
    private final RepositoryManager databaseRepository;
    /**
     * Default domain owner.
     */
    private final String domainOwner;
    /**
     * Root folder for domain databases.
     */
    private final File dbsFolder;

    public DomainServiceImpl(ConfigurationService configuration, VHostManager manager, RepositoryManager repositoryManager) {
        this.vHostRepository = getVHostRepository(manager);
        this.domainOwner = configuration.get(GlobalKey.SERVICE_REMOTE_LOGIN) + "@" + configuration.get(GlobalKey.SERVICE_REMOTE_DOMAIN);
        this.dbsFolder = new File(configuration.get(GlobalKey.DATA_DIR), DB_ROOT);
        this.databaseRepository = repositoryManager;
    }

    private ConfigRepository<VHostItem> getVHostRepository(VHostManager manager) {
        Bindings bindings = new SimpleBindings();
        manager.initBindings(bindings);
        ConfigRepository<VHostItem> result = (ConfigRepository<VHostItem>) bindings.get(ComponentRepository.COMP_REPO_BIND);
        return result;
    }

    private VHostItem newVHostItem(String domain) throws TigaseStringprepException {
        VHostItem vhostItem = new VHostItem();
        vhostItem.setOwner(domainOwner);
        vhostItem.setAdmins(new String[]{domainOwner});
        vhostItem.setVHost(domain);
        vhostItem.setEnabled(true);
        vhostItem.setAnonymousEnabled(false);
        vhostItem.setRegisterEnabled(false);
        return vhostItem;
    }

    public String getDbsPath(String domain) {
        return new File(this.dbsFolder, domain).getAbsolutePath();
    }

    @Override
    public Set<String> all() {
        Set<String> result = new HashSet<String>();
        try {
            Collection<VHostItem> allDomains = this.vHostRepository.allItems();
            for (VHostItem vHostItem : allDomains) {
                result.add(vHostItem.getKey()); //domain names are always lowercase
            }
        } catch (Exception e) {
            throw new XmppException("Getting all domains failed", e, Code.RESOURCE_ERROR);
        }

        return result;
    }

    @Override
    public void update(Map<String, Certificate> serverCertificates) {
        try {
            if (serverCertificates != null) {
                serverCertificates.forEach((domain, certificate) -> update(domain, certificate));
            }

        } catch (Exception e) {
            throw new XmppException("Updating certificates failed", e, Code.RESOURCE_ERROR);
        }
    }
    
    private void update(String domain, Certificate certificate) {
        try {
            if (domain == null) {
                throw new XmppException("Updating certificate failed, name can't be null", Code.RESOURCE_ERROR);
            }
            
            TLSUtil.addCertificate(asMap(domain, certificate));
            
        } catch (CertificateParsingException e) {
            throw new XmppException("Updating certificate failed, invalid certificate", e, Code.RESOURCE_ERROR);
        } 
        
    }

    @Override
    public void remove(String domain) {

        try {
            try {
                this.databaseRepository.removeRepository(domain);
            } catch (RepositoryManager.NotFoundException e) {
                LOG.log(Level.FINE, "Cannot remove domain {0} repository, data repository not found", domain);
            }

            //VHostItem domainToDelete = newVHostItem(domain);
            if (this.vHostRepository.contains(domain)) {
                this.vHostRepository.removeItem(domain);
            } else {
                LOG.log(Level.FINE, "Cannot remove domain {0} reference, virtual domain does not exist", domain);
            }

            DerbyHandler handler = new DerbyHandler(this.dbsFolder, domain);
            if (!handler.remove()) {
                LOG.log(Level.INFO, "Cannot remove domain {0} folder, files locked or unexpected files/folders found", domain);
            }

        } catch (XmppException e) {
            throw e;
        } catch (Exception e) {
            throw new XmppException("Removing domain failed", e, Code.RESOURCE_ERROR);
        }
    }

    @Override
    public void activate(String domain) {

        // Check if database exist. If not, create new one. If yes, check it. 
        DerbyChecker checker = new DerbyChecker(this.dbsFolder, domain);
        try {
            checker.check();
        } catch (DerbyChecker.NotExistException e) {
            LOG.log(Level.FINE, "Database not found, creating new one");
            DerbyHandler handler = new DerbyHandler(this.dbsFolder, domain);
            handler.create();
        } catch (DerbyChecker.ObsoleteSchemaException e) {
            LOG.log(Level.FINE, "Database schema {0} is obsolete. Current is {1}.", new Object[]{e.getFound(), Version.CURRENT});
            DerbyHandler handler = new DerbyHandler(this.dbsFolder, domain);
            handler.upgradeSchema(e.getFound());
            checker.check(); //check it again agains the current schema
        }

        //Database exists, register specific domain repository
        try {
            this.databaseRepository.addRepository(domain);
        } catch (RepositoryManager.DuplicateException e) {
            LOG.log(Level.FINE, "Adding new repository for domain {0} - repository already exist", domain);
        }

        //Database is ready to use.
        try {

            if (!this.vHostRepository.contains(domain)) {
                this.vHostRepository.addItem(newVHostItem(domain));
            } else {
                VHostItem existing = this.vHostRepository.getItem(domain);
                existing.setEnabled(true);
                LOG.log(Level.FINE, "Domain {0} already exists", domain);
            }

        } catch (Exception e) {
            throw new XmppException("Adding virtual domain {0} item failed", e, Code.RESOURCE_ERROR, domain);
        }
    }

    @Override
    public void backup(String backupPath) {
        boolean success = true;

        for (String domain : all()) { //in the time of backup, all domains should be correctly synchronized
            DerbyHandler handler = new DerbyHandler(dbsFolder, domain);
            if (!handler.backup(backupPath) && success) {
                success = false;
            }
        }

        if (success == false) {
            throw new XmppException("XMPP store backup failed.", Code.RESOURCE_ERROR);
        }
    }

    private Map<String, String> asMap(String domain, Certificate certificate) {
        Map<String, String> result = new HashMap<String, String>(3);
        
        if (certificate != null) {
            result.put(SSLContextContainerIfc.PEM_CERTIFICATE_KEY, asPem(certificate));
        }
        
        result.put(SSLContextContainerIfc.CERT_SAVE_TO_DISK_KEY, Boolean.toString(Boolean.FALSE));
        result.put(SSLContextContainerIfc.CERT_ALIAS_KEY, domain);

        return result;
    }

    private String asPem(Certificate certificate) {
        return String.format(PEM_CERT, certificate.getPublicCert(), certificate.getPrivateKey());
    }
}
