/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase;

import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.xml.Element;
import tigase.xmpp.JID;
import tigase.xmpp.NoConnectionIdException;
import tigase.xmpp.ResourceSelector;
import tigase.xmpp.StanzaType;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.XMPPSession;

/**
 *
 * @author tskoda
 */
public class MessageResourceSelector implements ResourceSelector {

    public enum Show {

        chat,
        availible,
        away,
        xa,
        dnd,
        unknown;

        public boolean hasLowerWeightThan(Show show) {
            //the higher ordinal the lower weight
            return this.ordinal() > show.ordinal();
        }

        public static Show valueOf(Element presence) {
            if (presence == null) {
                return unknown;
            }

            String show = presence.getCData("presence/show");

            if ((show == null) || (show.isEmpty())) {
                return availible;
            }

            try {
                return Show.valueOf(show.toLowerCase());
            } catch (IllegalArgumentException e) {
                return unknown;
            }
        }

        public static Show lowestWeight() {
            return unknown;
        }
    }

    @Override
    public Set<XMPPResourceConnection> preferredResources(JID to, StanzaType type, XMPPSession session) throws CanNotDeliver {
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        if (type == null) {
            type = StanzaType.normal; //correction according to specification
        }

        if (hasResource(to)) {
            XMPPResourceConnection single = session != null ? session.getResourceForJID(to) : null;
            if ((single != null) && (single.getPriority() >= 0)) {
                result.add(single);
            } else {
                switch (type) {
                    case chat:
                        highestPrefferedResources(result, session);
                        break;
                    case normal:
                    case groupchat:
                    case headline:
                    default:
                        throw new CanNotDeliver(CanNotDeliver.Action.serviceUnavailible, "User resource " + to.toString() + " is not available anymore.");
                }
            }
        } else {
            switch (type) {
                case chat:
                case normal:
                    highestPrefferedResources(result, session);
                    break;
                case groupchat:
                    throw new CanNotDeliver(CanNotDeliver.Action.serviceUnavailible, "Recipient resource must be specified for groupchat messages.");
                case headline:
                    positivePriorityResources(result, session);
                    if (result.isEmpty()) { //ignore when no session
                        throw new CanNotDeliver(CanNotDeliver.Action.ignore, "Headline message ignored because no user resource with non-negative priority found.");
                    }
                    break;
                default:
                    throw new CanNotDeliver(CanNotDeliver.Action.serviceUnavailible, "Message has incorrect message type " + type);
            }
        }

        return result;
    }

    public void positivePriorityResources(Set<XMPPResourceConnection> result, XMPPSession session) {
        if (session != null) {
            positivePriorityResources(result, session.getActiveResources());
        }
    }

    public void positivePriorityResources(Set<XMPPResourceConnection> result, List<XMPPResourceConnection> allResources) {
        for (XMPPResourceConnection connection : allResources) {
            if (connection.getPriority() >= 0) {
                result.add(connection);
            }
        }
    }

    public void highestPrefferedResources(Set<XMPPResourceConnection> result, XMPPSession session) {
        if (session != null) {
            highestPrefferedResources(result, session.getActiveResources());
        }
    }

    public void highestPrefferedResources(Set<XMPPResourceConnection> result, List<XMPPResourceConnection> allResources) {

        //group resources by highest found show and priority
        Show highestShow = Show.lowestWeight();
        //do not include any resource with priority under 0
        int highestPriority = 0;

        XMPPResourceConnection ImConnection = null;
        for (XMPPResourceConnection connection : allResources) {
            if ((connection.getPriority() >= 0) && (connection.isAuthorized())) { //ignore all with negative priority or just authorizing
                try {
                    JID connectioID = connection.getConnectionId();
                    if (connectioID != null && connectioID.getLocalpart() != null && connectioID.getLocalpart().equals(IMBridge.IM_BRIDGE_NAME)) {
                        ImConnection = connection;
                        continue;
                    }
                } catch (NoConnectionIdException ex) {
                    Logger.getLogger(MessageResourceSelector.class.getName()).log(Level.SEVERE, null, ex);
                }
                Show current = Show.valueOf(connection.getPresence());

                if (highestShow.equals(current)) {
                    if (highestPriority == connection.getPriority()) {
                        result.add(connection);
                    } else if (highestPriority < connection.getPriority()) {

                        highestPriority = connection.getPriority();

                        result.clear();
                        result.add(connection);
                    }
                } else if (highestShow.hasLowerWeightThan(current)) {

                    highestShow = current;
                    highestPriority = connection.getPriority();

                    result.clear();
                    result.add(connection);
                }
            }
        }
        if (ImConnection != null) {
            result.add(ImConnection);
        }
    }

    private boolean hasResource(JID to) {
        String resource = to.getResource();
        return resource != null && !resource.isEmpty();
    }
}
