/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase;

import com.kerio.im.bean.Port;
import com.kerio.im.common.error.Code;
import com.kerio.im.xmpp.api.XmppException;
import com.kerio.im.xmpp.api.XmppPorts;
import java.io.IOException;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import tigase.net.ConnectionType;
import tigase.net.SocketType;
import tigase.server.ConnectionManager;
import tigase.server.xmppclient.ClientConnectionManager;
import tigase.server.xmppserver.S2SConnectionManager;

/**
 *
 * @author tskoda
 */
public class PortServiceImpl implements XmppPorts {
    
    private static final Logger LOG = Logger.getLogger(PortServiceImpl.class.getName());
    public static final String SECURITY = "security";
    public static final String REMOTE_HOST = "remote-host";
    public static final String PORT_NO = "port-no";
    public static final String IFC = "ifc";
    public static final String SOCKET = "socket";
    public static final String TYPE = "type";
    
    private final ClientConnectionManager connectionManager;
    private final S2SConnectionManager federationManager;

    public PortServiceImpl(ClientConnectionManager clientConnectionManager, S2SConnectionManager federationConnectionManager) {
        this.connectionManager = clientConnectionManager;
        this.federationManager = federationConnectionManager;
    }

    private void start(ConnectionManager manager, Port port) {
        Map<String, Object> portProps = new HashMap<String, Object>();

        portProps.put(TYPE, ConnectionType.accept);
        portProps.put(SOCKET, port.isSSL() ? SocketType.ssl : SocketType.plain);
        portProps.put(SECURITY, port.getSecurity());
        portProps.put(REMOTE_HOST, port.getHost());
        portProps.put(PORT_NO, port.getPort());
        portProps.put(IFC, new String[]{port.getIfc()});
        portProps.put(ConnectionManager.Filter.FILETR_PROP_KEY, new FilterWrapper(port));

        try {
            manager.openPort(portProps);
        } catch (IOException e) {
            throw new XmppException("Starting port failed", e, Code.INTERNAL_SERVER_ERROR);
        }
    }
    
    public void stop(ConnectionManager manager, Port port) {
        try {
            manager.closePort(port.getPort(), new String[]{port.getIfc()});
        } catch (IOException e) {
            throw new XmppException("Stopping port failed", e, Code.INTERNAL_SERVER_ERROR);
        }
    }

    @Override
    public void start(Type type, Port port) {
        switch(type) {
            case CLIENT:
                start(this.connectionManager, port);
                break;
            case SERVER:
                start(this.federationManager, port);
                break;
            default:
                throw new XmppException("Starting port failed, unsupported port type {0}", Code.INTERNAL_SERVER_ERROR, type);
        }
    }

    @Override
    public void stop(Type type, Port port) {
        switch(type) {
            case CLIENT:
                stop(this.connectionManager, port);
                break;
            case SERVER:
                stop(this.federationManager, port);
                break;
            default:
                throw new XmppException("Stopping port failed, unsupported port type {0}", Code.INTERNAL_SERVER_ERROR, type);
        }
    }
    
    private class FilterWrapper implements ConnectionManager.Filter {
        private final Port port;

        public FilterWrapper(Port port) {
            this.port = port;
        }

        @Override
        public boolean accept(InetAddress client) {
            return this.port.accept(client);
        }

        @Override
        public void connected(String remoteAddress, int remotePort) {
            this.port.connected(remoteAddress, remotePort);
        }

        @Override
        public void disconnected(String remoteAddress, int remotePort) {
            this.port.disconnected(remoteAddress, remotePort);
        }
        
        
    }
}
