 /*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase;

import com.kerio.im.bean.VCard.VCardFactory;
import com.kerio.im.common.CommandLineParser;
import com.kerio.im.common.Task;
import com.kerio.im.common.error.Code;
import com.kerio.im.common.error.ServiceException;
import com.kerio.im.core.api.CoreService;
import com.kerio.im.core.api.GlobalKey;
import com.kerio.im.tigase.configuration.ConnectConfigurator;
import com.kerio.im.tigase.configuration.task.ApplySystemProperties;
import com.kerio.im.tigase.configuration.task.CheckClassPath;
import com.kerio.im.tigase.configuration.task.InstallSasl;
import com.kerio.im.tigase.plugin.VCardWrapperFactory;
import com.kerio.im.tigase.roster.IDefaultRosterListener;
import com.kerio.im.tigase.roster.IDynamicDomainRoster;
import com.kerio.im.tigase.roster.IPersonalRosterListener;
import com.kerio.im.tigase.roster.PacketBuilder;
import com.kerio.im.tigase.roster.RosterFactory;
import com.kerio.im.tigase.roster.RosterNotify;
import com.kerio.im.tigase.user.XmppUsersStorePool;
import com.kerio.im.xmpp.api.XmppDomains;
import com.kerio.im.xmpp.api.XmppPorts;
import com.kerio.im.xmpp.api.XmppService;
import com.kerio.im.xmpp.api.XmppUsers;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.server.MessageRouterIfc;
import tigase.server.xmppclient.ClientConnectionManager;
import tigase.server.xmppserver.S2SConnectionManager;
import tigase.vhosts.VHostManager;

/**
 *
 * @author tskoda
 */
public class TigaseServerImpl implements XmppService {
    
    private static final Logger LOG = Logger.getLogger(TigaseServerImpl.class.getName());
    
    private PortServiceImpl portService;
    private DomainServiceImpl domainService;
    private ConnectConfigurator configurator;
    private SessionProvider sessions;
    private PacketSender sender;
    private IDefaultRosterListener defaultRosterListener;
    
    private XmppUsersStorePool storePool;
    
    public TigaseServerImpl() {
    }
    
    public void start(final Map<String, String> cmdProperties, final CoreService coreService) throws ServiceException {
        run(new CheckClassPath(), coreService.configuration().get(GlobalKey.SERVICE_DIR));
        run(new ApplySystemProperties(), null);

        //the rest of configuration is made in configurator (see TigaseKeys)
        tigaseMain(cmdProperties, coreService);
        
        run(new InstallSasl(), null);
    }
    
    private <T> void run(Task<T> task, T parameter) {
        task.run(parameter);
    }
    
    private void tigaseMain(Map<String, String> properties, CoreService coreService) throws ServiceException {
        try {
            LOG.log(Level.FINE, "Starting Tigase server");
            
            this.storePool = new XmppUsersStorePool();
            
            RosterFactory rosterFactory = new RosterFactory(coreService.configuration().get(GlobalKey.ROSTER_LIMIT));
            
            IDynamicDomainRoster dynamicDomainRoster = rosterFactory.newDynamicDomainRoster(this.storePool, rosterFactory.newElementMapper(coreService.groups()));
            IPersonalRosterListener personalRosterListener = rosterFactory.newPersonalRosterListener(dynamicDomainRoster, coreService.groups());
            
            ComponentLookup lookup = new ComponentLookup();
            
            this.configurator = new ConnectConfigurator(
                    coreService,
                    lookup,
                    dynamicDomainRoster,
                    personalRosterListener);
            
            this.configurator.init(CommandLineParser.toArray(properties));
            this.configurator.setName("basic-conf");
            
            String mr_class_name = configurator.getMessageRouterClassName();
            MessageRouterIfc router
                    = (MessageRouterIfc) Class.forName(mr_class_name).newInstance();
            
            router.setName("message-router");
            router.setConfig(configurator);
            router.start();
            
            this.sender = this.configurator.getConnectComponent().getPacketSender();
            this.sessions = this.configurator.getSessions();
            this.storePool.setSessions(sessions);
            
            this.portService = new PortServiceImpl(lookup.lookup(ClientConnectionManager.class), lookup.lookup(S2SConnectionManager.class));
            this.domainService = new DomainServiceImpl(coreService.configuration(), lookup.lookup(VHostManager.class), this.configurator);
            this.storePool.setDomainService(domainService);
            
            this.defaultRosterListener = rosterFactory.newDefaultRosterListener(
                    dynamicDomainRoster,
                    new RosterNotify(coreService.groups(), new PacketBuilder(), this.sessions, this.sender),
                    coreService.groups());
            
            LOG.log(Level.INFO, "Starting Tigase server successful");
            
        } catch (Exception e) {
            throw new ServiceException("Starting Tigase server failed", e, Code.INTERNAL_SERVER_ERROR);
        }
    }
    
    @Override
    public XmppPorts ports() {
        return this.portService;
    }
    
    @Override
    public XmppUsers users(String domain) {
        return new com.kerio.im.tigase.user.XmppUsers(this.storePool.get(domain), this.defaultRosterListener);
    }
    
    @Override
    public XmppDomains domains() {
        return this.domainService;
    }
    
    @Override
    public VCardFactory vcardFactory() {
        return new VCardWrapperFactory();
    }
    
}
