/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.IM;

import com.kerio.im.external.IMContactID;
import com.kerio.im.external.IMConversation;
import com.kerio.im.external.IMEvent;
import com.kerio.im.external.UserImClient;
import com.kerio.im.external.IMPresence;
import com.kerio.im.external.IMMessage;
import com.kerio.im.external.ImResponseEnvelope;
import com.kerio.im.tigase.bridge.IMSubscribctionsMamager;
import com.kerio.im.tigase.bridge.xmpp.IEventsConsumerEx;
import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import com.kerio.im.tigase.bridge.xmpp.ImMessageSenderOffline;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.util.TigaseStringprepException;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;

/**
 *
 * @author vbystricky
 */
public class IMConnection {

    private final BareJID userID;
    private IMPresence.Status presenceStatus;
    private final UserImClient imClient;
    private Integer conversationSubscribeRequestId = -1;
    private final List<Integer> messagesSubscribeRequestIds = new ArrayList<>();
    private final IEventsConsumerEx<IMEvent> eventConsumer;
    private final IMSubscribctionsMamager subscribctionsMamager;
    private final List<Integer> subscribedConversations = new ArrayList<>();
    private final ArrayList<ClientStatus> clients = new ArrayList<>();
    private static final Logger LOGGER = Logger.getLogger(IMConnection.class.getName());

    public class ClientStatus {

        private IMPresence.Status status;
        private JID clientID;

        public ClientStatus(ClientStatus clientStatus) {
            this.status = clientStatus.status;
            this.clientID = JID.jidInstanceNS(clientStatus.getClientID().getBareJID(), clientStatus.getClientID().getResource());
        }

        public ClientStatus(IMPresence.Status status, JID clientID) {
            this.status = status;
            try {
                this.clientID = JID.jidInstance(clientID.getLocalpart(), clientID.getDomain(), clientID.getResource());
            } catch (TigaseStringprepException ex) {
                LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Error creating client status.", ex);
            }
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }

            if (obj.getClass() == ClientStatus.class) {
                return clientID.equals(((ClientStatus) obj).clientID);
            } else {
                return false;
            }
        }

        public IMPresence.Status getStatus() {
            return status;
        }

        public JID getClientID() {
            return clientID;
        }

        @Override
        public String toString() {
            return "client id: " + clientID + " status: " + status;
        }
    }

    public IMConnection(BareJID userID, UserImClient imClient, IEventsConsumerEx eventConsumer, IMSubscribctionsMamager subscribctionsMamager) {
        this.userID = userID;
        this.presenceStatus = IMPresence.Status.offline;
        this.imClient = imClient;
        this.eventConsumer = eventConsumer;
        this.subscribctionsMamager = subscribctionsMamager;
    }

    public BareJID getUserId() {
        return userID;
    }

    private IMPresence.Status recomputeStatus() {
        IMPresence.Status result = IMPresence.Status.offline;
        for (ClientStatus tmp : clients) {
            if (tmp.status.ordinal() > result.ordinal()) {
                result = tmp.status;
            }
        }
        return result;
    }

    public synchronized void close(boolean restart) {
        if (!restart) {
            changePresence(IMPresence.Status.offline);
        }
        clients.clear();
    }

    public synchronized IMPresence.Status setPresence(IMPresence.Status clientPresenceStatus, JID client) throws InterruptedException {
        IMPresence.Status newPresenceStatus = processClientStatus(clientPresenceStatus, client);
        if (this.presenceStatus == newPresenceStatus) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Presence {0} already set for user {1}. Request ignored.", new Object[]{newPresenceStatus, userID});
            return presenceStatus;
        }

        if (this.presenceStatus == IMPresence.Status.offline) {
            connect(newPresenceStatus);
        } else if (newPresenceStatus == IMPresence.Status.offline) {
            disconnect();
        } else {
            changePresence(newPresenceStatus);
        }
        return this.presenceStatus;
    }

    public synchronized void sendMessage(String text, Integer conversationID, boolean markAsRead) throws ImMessageSenderOffline {
        if (!isOnline()) {
            throw new ImMessageSenderOffline("Can't send message to IM because sender " + userID + " is offline.");
        }
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "User {0} is is sending message to conversation {1}.", new Object[]{userID, conversationID});
        imClient.sendMessage(new IMMessage(
                null,
                text,
                IMMessage.MessageEvent.active,
                conversationID,
                new IMContactID(this.userID.getLocalpart(), this.userID.getDomain()),
                null),
                markAsRead
        );
    }

    public synchronized void subscribeMessages(Integer conversationId, Integer lastReadId) throws InterruptedException {
        if (!isOnline()) {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Can't subscribe messages from conversation {0} for user {1}, because user is offline.", new Object[]{conversationId, userID});
            return;
        }

        if (subscribedConversations.contains(conversationId)) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "User {0} has already subscribed messages from conversation {1}.", new Object[]{userID, conversationId});
            return;
        }

        Integer requestId = imClient.generateRequestId();
        messagesSubscribeRequestIds.add(requestId);
        IMContactID imc = new IMContactID(userID.getLocalpart(), userID.getDomain());
        subscribctionsMamager.addSubscription(requestId, IMMessage.class, imc);
        eventConsumer.lock();
        try {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Subbscribing messages from conversation {0} with last read id {2} for user {1}.", new Object[]{conversationId, userID, lastReadId});
            ImResponseEnvelope<List<IMMessage>> response = imClient.subscribeMessages(conversationId, lastReadId, requestId);
            List<IMMessage> messages = response.getResponse();

            for (int i = messages.size() - 1; i >= 0; i--) {
                eventConsumer.enqueueUnchecked(new IMEvent(messages.get(i), imc));
            }
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "{0} messages returned from subscription of conversation {1} for user {2}.", new Object[]{messages.size(), conversationId, userID});
            subscribedConversations.add(conversationId);
        } catch (RuntimeException e) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Can't subscribe messages from conversation " + conversationId + " for user " + userID + ".", e);
            subscribctionsMamager.removeSubscription(requestId);
            messagesSubscribeRequestIds.remove(requestId);
            throw e;
        } finally {
            eventConsumer.unlock();
        }
    }

    public synchronized IMConversation createConversation(IMContactID participant) {
        if (!isOnline()) {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Can't create conversation between user {0} and {1} , because user {0} is offline.", new Object[]{userID, participant});
            return null;
        }
        LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Creating conversation between user {0} and {1}.", new Object[]{userID, participant});
        List<IMContactID> participants = new ArrayList<>();
        participants.add(participant);
        return imClient.createConversation(participants);
    }

    private boolean isOnline() {
        return presenceStatus != IMPresence.Status.invisible && presenceStatus != IMPresence.Status.offline;
    }

    private ClientStatus findClientStatus(JID client) {
        for (ClientStatus tmp : clients) {
            if (tmp.clientID.equals(client)) {
                return tmp;
            }
        }
        return null;
    }

    private IMPresence.Status processClientStatus(IMPresence.Status newPresenceStatus, JID client) {
        if (newPresenceStatus == IMPresence.Status.offline) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Client {0} has disconnected from XMPP server.", client);
            ClientStatus currenClientStatus = findClientStatus(client);
            clients.remove(currenClientStatus);
        } else {
            ClientStatus currenClientStatus = findClientStatus(client);
            if (currenClientStatus != null) {
                LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Client {0} has changed status from {1} to {2}.", new Object[]{client, currenClientStatus.status, newPresenceStatus});
                currenClientStatus.status = newPresenceStatus;
            } else {
                LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Client {0} has connected to XMPP server with status {1}.", new Object[]{client, newPresenceStatus});
                clients.add(new ClientStatus(newPresenceStatus, client));
            }
        }

        return recomputeStatus();
    }

    private void connect(IMPresence.Status newPresenceStatus) throws InterruptedException {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Connecting to IM as user {0} with presence {1}.", new Object[]{userID, newPresenceStatus});
        imClient.setPresence(newPresenceStatus, "");
        presenceStatus = newPresenceStatus;
        eventConsumer.lock();
        try {
            conversationSubscribeRequestId = imClient.generateRequestId();
            IMContactID cid = new IMContactID(userID.getLocalpart(), userID.getDomain());
            subscribctionsMamager.addSubscription(conversationSubscribeRequestId, IMConversation.class, cid);
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Subscribing conversations for user {0}.", cid);
            ImResponseEnvelope<List<IMConversation>> env = imClient.subscribeConversations(conversationSubscribeRequestId);
            for (IMConversation conversation : env.getResponse()) {
                eventConsumer.enqueueUnchecked(new IMEvent<>(conversation, cid));
            }
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "{0} conversations returned by subscription for user {1}.", new Object[]{env.getResponse().size(), cid});
        } catch (Throwable e) {
            subscribctionsMamager.removeSubscription(conversationSubscribeRequestId);
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Connection to IM for user " + userID + " failed", e);
        } finally {
            eventConsumer.unlock();
        }
    }

    private void disconnect() {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Disconnecting user {0} from IM.", userID);
        subscribctionsMamager.removeSubscription(conversationSubscribeRequestId);
        conversationSubscribeRequestId = -1;
        messagesSubscribeRequestIds.stream().forEach((tmp) -> {
            subscribctionsMamager.removeSubscription(tmp);
        });
        messagesSubscribeRequestIds.clear();
        imClient.setPresence(IMPresence.Status.offline, "");
        presenceStatus = IMPresence.Status.offline;
    }

    private void changePresence(IMPresence.Status newPresenceStatus) {
        presenceStatus = newPresenceStatus;
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Changing presence for user {0} to {1}.", new Object[]{userID, newPresenceStatus});
        imClient.setPresence(newPresenceStatus, "");
    }

    public synchronized void readConversation(Integer converationId, Integer messageId) {
        if (!isOnline()) {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Can't mark message {0} in conversation {1} as read, user {2} is offfline.", new Object[]{messageId, converationId, userID});
            return;
        }
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Marking message {0} in conversation {1} as read for user {2}", new Object[]{messageId, converationId, userID});
        imClient.readConversation(converationId, messageId);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        IMConnection connection = (IMConnection) obj;
        return connection.getUserId().equals(this.getUserId());
    }

    public List<ClientStatus> getClientsCopy() {
        List<ClientStatus> copy = new ArrayList<>(clients.size());
        clients.stream().forEach((client) -> {
            copy.add(new ClientStatus(client));
        });
        return copy;
    }
}
