/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.IM;

import com.kerio.im.external.IMContactID;
import com.kerio.im.external.IMConversation;
import com.kerio.im.external.IMPresence;
import com.kerio.im.tigase.bridge.IIMSynchronizationService;
import com.kerio.im.tigase.bridge.conversation.Conversation;
import com.kerio.im.tigase.bridge.conversation.ConversationJID;
import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import com.kerio.im.tigase.bridge.xmpp.IXmppPacketProcessor;
import com.kerio.im.tigase.bridge.xmpp.ImMessageSenderOffline;
import com.kerio.im.tigase.bridge.xmpp.InvalidRecipient;
import com.kerio.im.tigase.bridge.xmpp.XmppPresence;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.server.Packet;
import tigase.xml.XMLUtils;
import tigase.xmpp.JID;
import tigase.xmpp.StanzaType;

/**
 *
 * @author vbystricky
 */
public class XmppPacketsProcessor implements IXmppPacketProcessor {

    private static final Logger LOGGER = Logger.getLogger(XmppPacketsProcessor.class.getName());
    private final IIMSynchronizationService syncService;
    private final PacketClassifier packetClassifier;

    public XmppPacketsProcessor(IIMSynchronizationService syncService) {
        this.syncService = syncService;
        packetClassifier = new PacketClassifier(syncService.DomainInfoProvider());
    }

    private String getPacketID(Packet packet) {
        return (packet.getStanzaId() == null) ? "" : packet.getStanzaId();
    }

    private void fixConversationContacts(IMConversation imConversation, String eventRecipientName, String eventRecipientDomain) {
        List<IMContactID> contacts = imConversation.getContacts();
        IMContactID recipient = new IMContactID(eventRecipientName, eventRecipientDomain);
        if (!contacts.contains(recipient)) {
            contacts.add(recipient);
        }
    }

    private boolean processMessage(Packet packet) throws InterruptedException, InvalidRecipient {
        JID sender = packet.getStanzaFrom();
        JID recipient = packet.getStanzaTo();
        if (!packetClassifier.isSupportedChatMessage(packet)) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Unsuported chat packet id {0}.", getPacketID(packet));
            return false;
        }

        String message = packet.getElemCData("message/body");
        if (message == null) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Empty message packet  with id {0} from sender {1} detected.", new Object[]{getPacketID(packet), sender});
            return false;
        }
        message = XMLUtils.unescape(message);

        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Processing chat packet {0}", packet.toStringFull());
        boolean markMessageAsRead = !packetClassifier.isPacketDeliveredOnlyToIM(packet);

        if (sender.getBareJID().equals(recipient.getBareJID())) {
            throw new InvalidRecipient("Sender is same as recipient: \"" + sender.toString() + "\", \"" + recipient.toString() + "\".");
        }
        try {
            sendMessage(sender, recipient, message, markMessageAsRead);
        } catch (ImMessageSenderOffline ex) {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Trying to deliver message to IM, but sender({0}) is offline.", sender.toString());
            sendOfflineMessage(sender, recipient, message, markMessageAsRead);
        }
        return true;
    }

    private void sendMessage(JID sender, JID recipient, String message, boolean markMessageAsRead) throws InterruptedException, ImMessageSenderOffline {
        IMConnection connection = syncService.IMConnectionManager().getConnection(sender.getBareJID());
        if (connection == null) {
            throw new ImMessageSenderOffline("Can't send message to IM because sender " + sender.getBareJID() + " is offline.");
        }
        Conversation conversation = syncService.ConversationManager().getConversation(new ConversationJID(sender.getLocalpart(), sender.getDomain()), new ConversationJID(recipient.getLocalpart(), recipient.getDomain()));
        if (conversation == null) {
            IMConversation imConversation = connection.createConversation(new IMContactID(recipient.getLocalpart(), recipient.getDomain()));
            fixConversationContacts(imConversation, sender.getLocalpart(), sender.getDomain());
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Created conversation \"{0}\".", new Object[]{imConversation});
            conversation = syncService.ConversationManager().addConversation(imConversation);
            //subscribe messages before message is send!! we can't recognize if message origins from XMPP in subscribe messages result
            connection.subscribeMessages(conversation.getConversationId(), imConversation.getReceivedLastReadId());
        } else {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Found conversation \"{0}\".", new Object[]{conversation});
        }
        connection.sendMessage(message, conversation.getConversationId(), markMessageAsRead);
        conversation.updateSenderResource(sender);
    }

    private String getPresence(Packet packet) {
        if (packet.getType() != null && packet.getType() == StanzaType.unavailable) {
            return "offline";
        }

        String presence = packet.getElemCData("presence/show");
        if (presence == null) {
            if (packet.getType() == null || packet.getType() == StanzaType.available) {
                presence = "available";
            }
        }
        return presence;
    }

    private boolean processPressence(Packet packet) throws InterruptedException {
        JID sender = packet.getStanzaFrom();
        if (!packetClassifier.isSupportedPresencePacket(packet)) {
            return false;
        }
        String presenceString = getPresence(packet);
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Processing presence packet {0}.", packet.toStringFull());
        XmppPresence presence = XmppPresence.get(presenceString);
        if (presence != null) {
            IMPresence.Status imPresence = XmppPresence.getIMPresence(presence);
            IMConnection connection = syncService.IMConnectionManager().createConnection(sender.getBareJID());
            IMPresence.Status newIMPresence;
            newIMPresence = connection.setPresence(imPresence, sender);
            if (newIMPresence == IMPresence.Status.invisible || newIMPresence == IMPresence.Status.offline) {
                LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "User {0} has gone offline.", sender.getBareJID());
                syncService.IMConnectionManager().removeConnection(sender.getBareJID());
            }
        } else {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Unsupported XMPP presence \"{0}\" detected, presence packet from {1}.", new Object[]{presenceString, sender});
        }
        return true;
    }

    @Override
    public void processPacket(Packet packet) throws InterruptedException, InvalidRecipient {
        boolean processed = false;
        if (packet.getType() != null && packet.getType() == StanzaType.chat) {
            processed = processMessage(packet);
        } else if (packetClassifier.isPresencePacket(packet)) {
            processed = processPressence(packet);
        }
        if (!processed) {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Packet {0} was not processed.", packet.toStringFull());
        }
    }

    private void sendOfflineMessage(JID sender, JID recipient, String message, boolean markMessageAsRead) {
        IMServerConnection connection = syncService.IMConnectionManager().getServerConnection();
        Conversation conversation = syncService.ConversationManager().getConversation(new ConversationJID(sender.getLocalpart(), sender.getDomain()), new ConversationJID(recipient.getLocalpart(), recipient.getDomain()));
        IMContactID imSender = new IMContactID(sender.getLocalpart(), sender.getDomain());
        if (conversation == null) {
            IMConversation imConversation = connection.createConversation(imSender, new IMContactID(recipient.getLocalpart(), recipient.getDomain()));
            fixConversationContacts(imConversation, sender.getLocalpart(), sender.getDomain());
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Created conversation \"{0}\" in offline mode.", new Object[]{imConversation});
            conversation = syncService.ConversationManager().addConversation(imConversation);
        } else {
            LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Found conversation \"{0}\" in offline mode.", new Object[]{conversation});
        }
        connection.sendMessage(imSender, message, conversation.getConversationId(), markMessageAsRead);
    }
}
