package com.kerio.im.tigase.bridge.communicator;

import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.server.Packet;
import tigase.xmpp.JID;

/**
 *
 * @author jzacek
 */
public class IMtoXmppChannel {

    private static final Logger LOGGER = Logger.getLogger(IMtoXmppChannel.class.getName());
    private final Map<Integer, PacketResponse> expectedResponses = new HashMap<>();

    private final ICommunicatorComponent comm;

    private volatile int id = 0;

    public static class ID {

        protected ID(int id) {
            this.id = id;
        }
        public final int id;

        @Override
        public String toString() {
            return Integer.toString(id);
        }
    }

    public synchronized ID newID() {
        return new ID(this.id++);
    }

    public IMtoXmppChannel(ICommunicatorComponent comm) {
        this.comm = comm;
    }

    public IMtoXmppChannel sendMessage(Packet p) {
        comm.sendPacket(p);
        return this;
    }

    public synchronized IMtoXmppChannel sleep(int millis) {
        long start;
        long spent = 0;
        do {
            try {

                start = System.currentTimeMillis();
                this.wait(Math.max(1, millis - spent));
                spent += (System.currentTimeMillis() - start);

            } catch (InterruptedException ex) {
                //no need to do anything, just sleep in next loop
            }
        } while (spent < millis);
        return this;
    }

    public synchronized IMtoXmppChannel addResponse(ID id, PacketResponse response) {
        this.expectedResponses.put(id.id, response);
        return this;
    }

    public synchronized boolean waitForResponse(ID id, int timeout) throws InterruptedException {
        PacketResponse response = this.expectedResponses.get(id.id);
        LOGGER.log(Level.FINER, IMBridge.LOG_PREFIX + "Waiting for response {0}", response);

        long start;
        long spent = 0;
        try {
            while (!response.isHit() && (spent < timeout)) {
                start = System.currentTimeMillis();
                this.wait(Math.max(1, timeout - spent));
                spent += (System.currentTimeMillis() - start);
            }
        } finally {
            this.expectedResponses.remove(id.id);
        }

        if (!response.isHit()) {
            LOGGER.log(Level.WARNING, IMBridge.LOG_PREFIX + "Didn't get expected response in time.");
            return false;
        }

        return true;
    }

    public synchronized boolean onReceivedPacket(Packet p) {
        boolean processed = false;

        for (PacketResponse tmp : this.expectedResponses.values()) {
            if (tmp.check(p)) {
                processed = true;
            }
        }
        this.notifyAll();
        return processed;
    }

    public JID getComponentId() {
        return comm.getComponentId();
    }
}
