package com.kerio.im.tigase.bridge.xmpp;

import com.kerio.im.core.api.CoreService;
import com.kerio.im.core.api.event.StartupFinishedEvent;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.tigase.bridge.IM.DisabledIMException;
import com.kerio.im.tigase.bridge.IM.IDomainInfoProvider;
import com.kerio.im.tigase.bridge.IM.IMSynchronizationService;
import com.kerio.im.tigase.bridge.IM.IMSynchronizationServiceManager;
import com.kerio.im.tigase.bridge.IM.PacketClassifier;
import com.kerio.im.tigase.bridge.communicator.IMtoXmppChannel;
import com.kerio.im.tigase.bridge.communicator.ICommunicatorComponent;
import java.lang.reflect.UndeclaredThrowableException;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.conf.Configurable;
import tigase.server.AbstractMessageReceiver;
import tigase.server.Command;
import tigase.server.DisableDisco;
import tigase.server.Packet;

/**
 *
 * @author jzacek
 */
public class IMBridge extends AbstractMessageReceiver implements DisableDisco, ICommunicatorComponent, IDomainInfoProvider {

    public static final String IM_BRIDGE_NAME = "im-bridge";
    public static final String LOG_PREFIX = "IM Bridge: ";

    private static final Logger LOGGER = Logger.getLogger(IMBridge.class.getName());
    private IMtoXmppChannel communicator;
    private final IMSynchronizationServiceManager iMSynchronizationServiceManager;
    private final PacketClassifier packetClassifier;
    private boolean startupSubscribed;

    public IMBridge() {
        packetClassifier = new PacketClassifier(this);
        iMSynchronizationServiceManager = new IMSynchronizationServiceManager();
        startupSubscribed = false;
    }

    @Override
    public boolean isLocalDomain(String domain) {
        if (domain.equals("kerio-connect.private")) {
            return false;
        }
        return vHostManager.isLocalDomain(domain);
    }

    @Override
    public void setProperties(Map<String, Object> props) {
        super.setProperties(props);
        CoreService coreService = (CoreService) props.get(CoreService.class.getName());
        if (coreService != null && !startupSubscribed) {
            startupSubscribed = true;
            coreService.bus().subscribe(StartupFinishedEvent.class, new StartupListener(iMSynchronizationServiceManager));
        }
    }

    @Override
    public void initializationCompleted() {
        super.initializationCompleted();
        if (communicator == null) {
            LOGGER.log(Level.INFO, LOG_PREFIX + "IM Bridge initialised.");
            this.communicator = new IMtoXmppChannel(this);
            iMSynchronizationServiceManager.initialize(communicator, this, Configurable.DEF_SM_NAME + "@" + getDefHostName().getDomain());
        }
    }

    @Override
    public void stop() {
        super.stop();
        try {
            iMSynchronizationServiceManager.stop();
        } catch (IllegalStateException ex) {
            LOGGER.log(Level.INFO, "Error when stopping IM service.", ex);
        } catch (DisabledIMException ex) {
            LOGGER.log(Level.INFO, "IM bridge is disabled.", ex);
        }
    }

    private void onError(Packet packet) {
        if (!packetClassifier.isSupportedChatMessage(packet)) {
            LOGGER.log(Level.INFO, "Packet is not supported message, error mesasage will not be send to sender for packet: {0}", packet.toString());
            return;
        }
        if (!packetClassifier.hasMessage(packet)) {
            LOGGER.log(Level.INFO, "Packet does not have body, error mesasage will not be send to sender for packet: {0}", packet.toString());
            return;
        }
        if (!packetClassifier.isPacketDeliveredOnlyToIM(packet)) {
            LOGGER.log(Level.INFO, "Packet was also delivered to another clinet, error mesasage will not be send to sender for packet: {0}", packet.toString());
            return;
        }
        LOGGER.log(Level.INFO, "Sending error mesasage packet: {0}", packet.toString());
        sendPacket(PacketFactory.getErrorPacket(packet));
    }

    private IMSynchronizationService getService(boolean waitForStartup) throws InterruptedException, DisabledIMException {
        IMSynchronizationService service = null;
        VariableSleeper sleeper = null;
        while (service == null) {
            service = iMSynchronizationServiceManager.service();
            if (service == null && waitForStartup) {
                if (sleeper == null) {
                    sleeper = new VariableSleeper();
                }
                sleeper.sleep();
            } else {
                break;
            }
        }
        return service;
    }

    @Override
    public void processPacket(Packet packet) {
        LOGGER.log(Level.FINE, LOG_PREFIX + "Got inbound packet: {0}", packet.toString());

        if (packet.getCommand() == Command.CHECK_USER_CONNECTION) {
            this.sendPacket(packet.okResult((String) null, 0));
            LOGGER.log(Level.FINE, LOG_PREFIX + "Received CHECK_USER_CONNECTION Packet.");
            return;
        }

        if (this.communicator.onReceivedPacket(packet)) {
            return;
        }

        VariableSleeper sleeper = null;
        try {
            while (!processPacketForIM(packet)) {
                if (sleeper == null) {
                    sleeper = new VariableSleeper();
                }
                sleeper.sleep();
            }
        } catch (InterruptedException ex) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Interrupted while processing xmpp packet.", ex);
            onError(packet);
        }
    }

    private boolean processPacketForIM(Packet packet) throws InterruptedException {
        try {
            IMSynchronizationService service = getService(true);
            if (packet.getCommand() == Command.CLOSE) {
                LOGGER.log(Level.INFO, LOG_PREFIX + "Received CLOSE Packet.");
                if (packet.getStanzaTo() != null && packet.getStanzaTo().getResource() != null) {
                    LOGGER.log(Level.INFO, LOG_PREFIX + "Recieved command to close connection {0}", packet.getStanzaTo().getResource());
                    service.closeXmppConnection(packet.getStanzaTo().getResource());
                }
                return true;
            }
            service.IXmppPacketProcessor().processPacket(packet);
        } catch (DisabledIMException ex) {
            LOGGER.log(Level.FINEST, IMBridge.LOG_PREFIX + "IM bridge is disabled, ignoring packet.", ex);
        } catch (InterruptedException ex) {
            throw ex;
        } catch (UndeclaredThrowableException | InvalidRecipient ex) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Error while processing xmpp packet.", ex);
            onError(packet);
        } catch (Throwable ex) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Error while processing xmpp packet, service will be restarted.", ex);
            iMSynchronizationServiceManager.getWatcher().signalRestartNeeded();
            return false;
        }
        return true;
    }

    @Override
    public void sendPacket(Packet p) {
        LOGGER.log(Level.FINE, LOG_PREFIX + "Sending outbound packet: {0}", p.toString());
        this.addOutPacket(p);
    }

    @Override
    public String getDiscoDescription() {
        return "Kerio Connect IM Bridge";
    }

    public IAuthServiceProvider getConnectionManager() {
        return iMSynchronizationServiceManager.authServiceProvider();
    }

    private class StartupListener implements ServiceBus.EventListener<StartupFinishedEvent> {

        private final IMSynchronizationServiceManager imSynchronizationServiceManager;

        public StartupListener(IMSynchronizationServiceManager imSynchronizationServiceManager) {
            this.imSynchronizationServiceManager = imSynchronizationServiceManager;
        }

        @Override
        public void event(StartupFinishedEvent event) {
            try {
                imSynchronizationServiceManager.start(event.getContent());
            } catch (IllegalStateException ex) {
                LOGGER.log(Level.WARNING, "IM bridge service is in illegal state.", ex);
            } catch (DisabledIMException ex) {
                LOGGER.log(Level.INFO, "IM bridge is disabled.", ex);
            }
        }
    }
}
