/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.bridge.xmpp;

import com.kerio.im.external.IMContactID;
import com.kerio.im.external.IMConversation;
import com.kerio.im.external.IMEvent;
import com.kerio.im.external.IMMessage;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.util.TigaseStringprepException;
import com.kerio.im.external.IMPresence;
import com.kerio.im.external.ISingnalHandler;
import com.kerio.im.tigase.bridge.IIMSynchronizationService;
import com.kerio.im.tigase.bridge.IM.IMConnection;
import com.kerio.im.tigase.bridge.conversation.Conversation;
import com.kerio.im.tigase.bridge.conversation.ConversationJID;
import java.lang.reflect.UndeclaredThrowableException;
import java.util.List;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;

/**
 *
 * @author vbystricky
 */
public class IMEventsProcessor implements Runnable {

    private final EventQueue<IMEvent> events;
    private final AtomicBoolean shouldRun;
    private final IIMSynchronizationService syncService;
    private static final long MAX_RESOURCE_INACTIVE_TIME = 10L * 60L * 1000000000L; //10 minutes in nanoseconds
    private static final Logger LOGGER = Logger.getLogger(IMEventsProcessor.class.getName());
    private ISingnalHandler watcher;

    Thread IMEventsWorker;

    public IMEventsProcessor(IIMSynchronizationService syncService, EventQueue<IMEvent> queue) {
        this.syncService = syncService;
        this.shouldRun = new AtomicBoolean(true);
        this.events = queue;
    }

    public synchronized void start(ISingnalHandler watcher) {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Starting IMEventsProcessor");
        this.watcher = watcher;
        shouldRun.set(true);
        IMEventsWorker = new Thread(this, "IMEventsProcessor");
        IMEventsWorker.start();
    }

    public synchronized void stop() {
        LOGGER.log(Level.FINE, IMBridge.LOG_PREFIX + "Stopping IMEventsProcessor");
        shouldRun.set(false);
        if (IMEventsWorker != null) {
            try {
                IMEventsWorker.interrupt();
                IMEventsWorker.join();
            } catch (InterruptedException ex) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Stopping IMEventsProcessor interrupted.", ex);
            }
        }
        IMEventsWorker = null;
    }

    private void processOnlineStatus(IMContactID contactId, IMPresence.Status presence) throws InterruptedException {
        try {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Processing online status: {0} for user: {1}", new Object[]{presence, contactId});
            XmppPresence xmppPresence = XmppPresence.getXmppPresence(presence);
            IXmppConnection connection = syncService.XmppConnectionManager().getConnectionByUserName(contactId.getFullName());
            if (connection == null) {
                connection = syncService.XmppConnectionManager().createConnection(contactId.getUserName(), contactId.getDomain());
            }
            connection.setPresence(xmppPresence, "");
        } catch (TigaseStringprepException | XMPPConnectionOpenFailed ex) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Error while processing online status.", ex);
        }
    }

    private void processOfflineStatus(IMContactID contactId) {
        LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Processing offline status for user: {0}", contactId);
        IXmppConnection connection = syncService.XmppConnectionManager().getConnectionByUserName(contactId.getFullName());
        if (connection != null) {
            syncService.XmppConnectionManager().destroyConnection(connection);
        }
    }

    private void processPresence(IMPresence presenceEvent) throws InterruptedException {
        if (presenceEvent == null || presenceEvent.getStatus() == null || presenceEvent.getContactId() == null) {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Recieved invalid IMPresence {0}", presenceEvent);
            return;
        }

        switch (presenceEvent.getStatus()) {
            case available:
            case dnd:
            case away:
                processOnlineStatus(presenceEvent.getContactId(), presenceEvent.getStatus());
                break;
            case offline:
            case invisible:
                processOfflineStatus(presenceEvent.getContactId());
                break;
        }
    }

    private void fixConversationContacts(IMConversation imConversation, String eventRecipientName, String eventRecipientDomain) {
        List<IMContactID> contacts = imConversation.getContacts();
        IMContactID recipient = new IMContactID(eventRecipientName, eventRecipientDomain);
        if (!contacts.contains(recipient)) {
            contacts.add(recipient);
        }
    }

    private void processConversation(IMConversation imConversation, IMContactID eventRecipient) throws InterruptedException {
        if (imConversation.getEvent() == IMConversation.ConversationEvent.Created && imConversation.getConversationId() != null) {
            LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Recieved conversation (id:{0}) create event for user {1}", new Object[]{imConversation.getConversationId(), eventRecipient});
            try {
                fixConversationContacts(imConversation, eventRecipient.getUserName(), eventRecipient.getDomain());
                syncService.ConversationManager().addConversation(imConversation);
                IMConnection imConnection = syncService.IMConnectionManager().getConnection(BareJID.bareJIDInstance(eventRecipient.getUserName(), eventRecipient.getDomain()));
                if (imConnection != null) {
                    imConnection.subscribeMessages(imConversation.getConversationId(), imConversation.getReceivedLastReadId());
                }
            } catch (TigaseStringprepException e) {
                LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Error while processing conversation.", e);
            }
        }
    }

    boolean resourceExpired(ConversationJID receiverCID) {
        long currentTime = System.nanoTime();
        return (currentTime - receiverCID.getLastetResourceUsedTime()) > MAX_RESOURCE_INACTIVE_TIME;
    }

    private void processMessages(IMMessage message, IMContactID eventRecipient) {
        try {
            if (eventRecipient.equals(message.getFrom())) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "IM message recipient {0} is same as sender.", eventRecipient);
                return;
            }

            Conversation conversation = syncService.ConversationManager().getConversation(message.getTo());
            if (conversation == null) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Can't find conversation for sender {0} and recipient {1}", new Object[]{message.getFrom(), message.getTo()});
                return;
            }

            ConversationJID receiverCID = conversation.getOther(message.getFrom());
            if (receiverCID.getResource() != null && resourceExpired(receiverCID)) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Resource in contact {0} from conversation {1} has expired.", new Object[]{receiverCID, conversation});
                receiverCID.setResource(null);
            }

            JID receiver = receiverCID.toJID();
            IMConnection imConnection = syncService.IMConnectionManager().getConnection(receiver.getBareJID());
            if (imConnection == null) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Can't find connection for sender {0}.", receiver.getBareJID());
                return;
            }

            IXmppConnection connection = syncService.XmppConnectionManager().getConnectionByUserName(message.getFrom().getFullName());
            if (connection == null) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "Sending message from {0} to {1} in offline mode.", new Object[]{message.getFrom(), message.getTo()});
                connection = syncService.XmppConnectionManager().getOfllineDeliveryConnection(message.getFrom().getUserName(), message.getFrom().getDomain());
            }

            connection.sendMessage(receiver, message.getText());
            imConnection.readConversation(message.getTo(), message.getMessageId());
        } catch (TigaseStringprepException ex) {
            LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "Error while processing messages.", ex);
        }
    }

    @Override
    public void run() {
        while (shouldRun.get()) {
            try {
                Object event = events.dequeue();

                if (event == null || !(event instanceof IMEvent<?>)) {
                    continue;
                }

                IMEvent<?> imEvent = (IMEvent<?>) event;
                if (imEvent.getEvent() instanceof IMPresence) {
                    processPresence((IMPresence) imEvent.getEvent());
                } else if (imEvent.getEvent() instanceof IMConversation) {
                    processConversation((IMConversation) imEvent.getEvent(), imEvent.getRecipient());
                } else if (imEvent.getEvent() instanceof IMMessage) {
                    processMessages((IMMessage) imEvent.getEvent(), imEvent.getRecipient());
                }
            } catch (InterruptedException ex) {
                LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "IM Events Processor interrupted.", ex);
            } catch (UndeclaredThrowableException ex) {
                LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "IM Events Processor ended with error.", ex);
            } catch (Throwable ex) {
                LOGGER.log(Level.SEVERE, IMBridge.LOG_PREFIX + "IM Events Processor ended with error, service will be restarted.", ex);
                watcher.signalRestartNeeded();
                break;
            }
        }
        LOGGER.log(Level.INFO, IMBridge.LOG_PREFIX + "IM Events Processor is stopping.");
    }
}
