/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.plugin;

import com.kerio.im.core.api.service.ConversationService;
import com.kerio.im.core.api.service.ConversationService.FileInfo;
import com.kerio.im.core.api.service.ConversationService.MessageInfo;
import com.kerio.im.core.api.service.ConversationService.MessageType;
import java.time.LocalDateTime;
import java.util.EnumMap;
import java.util.logging.Logger;
import tigase.server.Packet;
import tigase.server.PacketFilterIfc;
import tigase.server.QueueType;
import tigase.stats.StatisticsList;
import tigase.xml.Element;
import tigase.xmpp.BareJID;
import tigase.xmpp.StanzaType;

/**
 * Inherit only from XMPPPreprocessorIfc as we want to archive/put in history
 * only packets received from clients not generated on server.
 *
 * @author tskoda
 */
public class ConversationListenerFilter implements PacketFilterIfc {

    private static final Logger LOG = Logger.getLogger(ConversationListenerFilter.class.getName());
    public static Filter BLOCK_DELAYED = new BlockDelayed();
    public static Filter BLOCK_ON_BEHALF_OF_RECIPIENT = new BlockOnBehalfOfRecipient();

    public static abstract class Notifier {
	
	private final Filter[] filters;
	
	public Notifier(Filter... filters) {
	    this.filters = filters;
	}

	abstract StanzaType handles();
	abstract void notifyImpl(BareJID from, BareJID to, Element content, ConversationService listener);

	public final void notify(BareJID from, BareJID to, Packet packet, ConversationService listener) {
	    if (this.filters != null) {
		for (Filter filter: this.filters) {
		    if (! filter.pass(packet)) {
			return;
		    }
		}
	    }
	    
	    notifyImpl(from, to, packet.getElement(), listener);
	}
    }

    public interface Filter {

	boolean pass(Packet packet);
    }

    public static class BlockDelayed implements Filter {

	@Override
	public boolean pass(Packet packet) {
	    boolean result = true;
	    if ((packet.getElement().findChild("message/delay")) != null) {
		result = false;
	    }
	    return result;
	}
    }
    
    public static class BlockOnBehalfOfRecipient implements Filter {

	@Override
	public boolean pass(Packet packet) {
	    boolean result = true;
	    if (packet.getPacketTo() != null) {
		result = false;
	    }
	    return result;
	}
    }

    private static class Void extends Notifier {

	@Override
	public StanzaType handles() {
	    return null;
	}

	@Override
	void notifyImpl(BareJID from, BareJID to, Element content, ConversationService listener) {
	    //do nothing
	}
    }

    public static class GroupChatMessage extends ChatMessage {

	public GroupChatMessage(Filter... filters) {
	    super(filters);
	}
	
	@Override
	public StanzaType handles() {
	    return StanzaType.groupchat;
	}

	@Override
	public MessageInfo info(final String content) {

	    return new MessageInfo() {

		@Override
		public String content() {
		    return content;
		}

		@Override
		public MessageType type() {
		    return MessageType.group;
		}
	    };
	}
    }

    public static class ChatMessage extends Notifier {

	public ChatMessage(Filter... filters) {
	    super(filters);
	}

	@Override
	public StanzaType handles() {
	    return StanzaType.chat;
	}

	@Override
	public void notifyImpl(BareJID from, BareJID to, Element content, ConversationService listener) {
	    String chatContent = getContent(content);
	    if (chatContent != null) { //filter out chat states etc.
            listener.message(from.toString(), to.toString(), info(chatContent), LocalDateTime.now(), content.toString());
	    }
	}

	private String getContent(Element messageElement) {
	    String result = null;
	    if (messageElement != null) {
		result = messageElement.getCData("message/body/");
	    }

	    return result;
	}

	public MessageInfo info(final String content) {

	    return new MessageInfo() {

		@Override
		public String content() {
		    return content;
		}

		@Override
		public MessageType type() {
		    return MessageType.chat;
		}

	    };
	}
    }

    public static class FileTransfer extends Notifier {

	public FileTransfer(Filter... filters) {
	    super(filters);
	}
	
	@Override
	public StanzaType handles() {
	    return StanzaType.set;
	}

	@Override
	public void notifyImpl(BareJID from, BareJID to, Element content, ConversationService listener) {
	    Element fileTransferElement = getFileElement(content);
	    if (fileTransferElement != null) { // is this file transfer?
		ConversationService.FileInfo fileInfo = getFileInfo(fileTransferElement);
		if (fileInfo != null) {
		    listener.file(from.toString(), to.toString(), fileInfo, LocalDateTime.now(), content.toString());
		} else {
		    //log debug log warning
		}
	    }
	}

	private Element getFileElement(Element iqSetElement) {
	    Element result = null;

	    if (iqSetElement != null) {
		result = iqSetElement.findChild("iq/si/file");
	    }

	    return result;
	}

	private ConversationService.FileInfo getFileInfo(Element element) {
	    ConversationService.FileInfo result = null;

	    if (element != null) {
		String fileName = element.getAttribute("name");
		if (fileName != null) {
		    String fileSizeString = element.getAttribute("size");
		    if (fileSizeString != null) {
			try {
			    long fileSize = Long.valueOf(fileSizeString);
			    result = info(fileName, fileSize);
			} catch (NumberFormatException e) {
			    result = info(fileName, 0);
			}
		    }
		}
	    }
	    return result;
	}

	public FileInfo info(final String name, final long size) {

	    return new FileInfo() {

		@Override
		public String name() {
		    return name;
		}

		@Override
		public long size() {
		    return size;
		}
	    };
	}
    }

    private class NotifierSelect {

	private final EnumMap<StanzaType, Notifier> notifiers = new EnumMap<StanzaType, Notifier>(StanzaType.class);
	private final Void DEFAULT = new Void();

	public NotifierSelect(Notifier... notifiers) {
	    for (Notifier n : notifiers) {
		this.notifiers.put(n.handles(), n);
	    }
	}

	public Notifier get(Packet packet) {
	    Notifier result = this.notifiers.get(packet.getType());
	    return result != null ? result : DEFAULT;
	}
    }

    private final ConversationService conversationService;
    private final NotifierSelect notifierSelect;
    private QueueType type;
    private String name;

    public ConversationListenerFilter(ConversationService conversationService, Notifier... notifiers) {
	this.conversationService = conversationService;
	this.notifierSelect = new NotifierSelect(notifiers);
    }
    
    @Override
    public void init(String name, QueueType qType) {
	this.type = qType;
	this.name = name;
    }

    @Override
    public Packet filter(Packet packet) {
//	if ((this.name.equals("sess-man") && this.type == QueueType.OUT_QUEUE) || this.name.equals("s2s")) {
//	    PacketLogger.log(this.name + "(" + this.type + ")", packet);
//	}

	BareJID jidFrom;
        BareJID jidTo;
	
	if (((jidFrom = packet.getStanzaFrom() != null ? packet.getStanzaFrom().getBareJID() : null) != null)
                    && ((jidTo = packet.getStanzaTo() != null ? packet.getStanzaTo().getBareJID() : null) != null)) {
	    
	    this.notifierSelect.get(packet).notify(jidFrom, jidTo, packet, this.conversationService);
	}
	return packet;
    }

    @Override
    public void getStatistics(StatisticsList list) {
    }

//CODE SNIPET FOR RESOLVING packet type from user point of view    
//    private boolean sentFromLocal(XMPPResourceConnection session, BareJID stanzaFrom, JID packetFrom) throws NoConnectionIdException, NotAuthorizedException {
//	return session != null && session.isUserId(stanzaFrom) && session.getConnectionId().equals(packetFrom);
//    }
//
//    //messages sent from foreign user to local user
//    private boolean sentFromRemote(Packet packet) throws NoConnectionIdException, NotAuthorizedException {
//	return packet.getPermissions() == Permissions.REMOTE;
//    }
//
//    //messages sent by server
//    private boolean sentToUser(Packet packet) {
//	return packet.getPacketFrom() == null;
//    }
}
