/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.plugin;

import com.kerio.im.core.api.service.AccessPolicyService;
import com.kerio.im.core.api.service.PacketType;
import com.kerio.im.tigase.PacketResolver;
import java.util.ArrayDeque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.db.NonAuthUserRepository;
import tigase.db.TigaseDBException;
import tigase.server.Message;
import tigase.server.Packet;
import tigase.util.TigaseStringprepException;
import tigase.xml.Element;
import tigase.xmpp.Authorization;
import tigase.xmpp.JID;
import tigase.xmpp.PacketErrorTypeException;
import tigase.xmpp.StanzaType;
import tigase.xmpp.XMPPPacketFilterIfc;
import tigase.xmpp.XMPPPreprocessorIfc;
import tigase.xmpp.XMPPProcessor;
import tigase.xmpp.XMPPResourceConnection;

/**
 *
 * @author tskoda
 */
public class DomainFilterPlugin extends XMPPProcessor
        implements XMPPPacketFilterIfc, XMPPPreprocessorIfc {

    private static final Logger LOG = Logger.getLogger(DomainFilterPlugin.class.getName());
    public static final String ID = "connect-domain-filter";
    
    private static final String[] ELEMENTS = {ALL};
    private static final String[] XMLNSS = {ALL};
    private static final boolean ALLOW = false;
    private static final boolean DENY = true;
    private AccessPolicyService accessPolicy;

    public DomainFilterPlugin() {
    }

    @Override
    public void init(Map<String, Object> settings) throws TigaseDBException {
        super.init(settings);
        this.accessPolicy = ((AccessPolicyService) settings.get(AccessPolicyService.class.getName()));
    }

    @Override
    public String id() {
        return ID;
    }

    private Packet presenceResponse(Packet source, JID from, JID to) {
        Element item = new Element(PacketResolver.PRESENCE);
        item.setXMLNS(CLIENT_XMLNS);
        item.setAttribute("from", from.toString());
        item.setAttribute("to", to.toString());
        item.setAttribute("type", StanzaType.unavailable.toString());


        Packet result = source.swapFromTo(item, source.getStanzaTo(), source.getStanzaFrom());
        //Packet result = new Presence(item, from, to);
        result.processedBy(ID);
        return result;
    }

    private Packet messageResponse(Packet source, JID from, JID to, String message) {
        Element messageElement = new Element("message", new Element[]{new Element("body", message)}, null, null);
        messageElement.setXMLNS(CLIENT_XMLNS);
        messageElement.addAttribute("from", from.toString());
        messageElement.addAttribute("to", to.toString());
        messageElement.addAttribute("type", StanzaType.chat.toString());

        Packet result = source.swapFromTo(messageElement, source.getStanzaTo(), source.getStanzaFrom());
        //Packet result = Message.getMessage(from, to, StanzaType.chat, , null, null, null);
        result.processedBy(ID);
        return result;
    }

    private Packet commonResponse(Packet source, String message) throws PacketErrorTypeException {
        Packet result = Authorization.RECIPIENT_UNAVAILABLE.getResponseMessage(source, message, true);
        result.processedBy(ID);
        return result;
    }

    private void deny(Packet source, Queue<Packet> errors, String message, PacketType type) {
        source.processedBy(ID);

        if ((source.getStanzaFrom() != null) && (source.getStanzaTo() != null)) {
            switch (type) {
                case ERROR:
                    //nothing
                    break;
                case MESSAGE:
                    //and also send unavailible presence
                    errors.add(messageResponse(source, source.getStanzaFrom(), source.getStanzaTo(), message));
                    errors.add(presenceResponse(source, source.getStanzaFrom(), source.getStanzaTo()));
                    break;
                case PRESENCE:
                    //errors.add(presenceResponse(source.getStanzaFrom(), source.getStanzaTo()));
                    //errors.add(presenceResponse(source.getStanzaTo(), source.getStanzaFrom()));
                    //break;
                default:
                    try {
                        errors.add(commonResponse(source, message));
                    } catch (PacketErrorTypeException e) {
                        //nothing to do
                    }
            }
        }
    }

    private void deny(Packet source, Queue<Packet> errors, String message, PacketType type, Iterator<Packet> it) {
        if (it != null) {
            it.remove();
        }

        deny(source, errors, message, type);
    }

    /*
     * Processing outgoing packets.
     */
    @Override
    public void filter(Packet origin, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> outgoingPackets) {
        if ((session != null) && (outgoingPackets != null) && (!outgoingPackets.isEmpty())) {
            Queue<Packet> errors = new ArrayDeque<Packet>(1);

            for (Iterator<Packet> it = outgoingPackets.iterator(); it.hasNext();) {
                Packet outgoingPacket = it.next();
                if (outgoingPacket.wasProcessedBy(ID)) {
                    continue;
                }
                //send to access policy null values as well and let it to decide
                JID stanzaFrom = outgoingPacket.getStanzaFrom();
                String from = stanzaFrom != null ? stanzaFrom.getLocalpart() : null;
                String fromDomain = stanzaFrom != null ? stanzaFrom.getDomain() : null;

                JID stanzaTo = outgoingPacket.getStanzaTo();
                String to = stanzaTo != null ? stanzaTo.getLocalpart() : null;
                String toDomain = stanzaTo != null ? stanzaTo.getDomain() : null;

//                LOG.log(Level.INFO,
//                        "[outgoing] Session owner {0}, session id {1}, stanza from {2} to {3}, content {4}",
//                        new Object[]{getOwner(session), getConnectionId(session), stanzaFrom, stanzaTo, outgoingPacket.getElemName()});

                PacketType type = PacketResolver.resolve(outgoingPacket);
                
                if (!this.accessPolicy.outgoing(to, toDomain, from, fromDomain, type)) {
                    deny(outgoingPacket, errors, this.accessPolicy.outgoingResponse(), type, it);
                }
            }

            outgoingPackets.addAll(errors);
        }
    }

    /*
     * Processing incoming packets
     */
    @Override
    public boolean preProcess(Packet incomingPacket, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) {
        boolean result = ALLOW;

        if ((session != null) && (!session.isServerSession()) && !incomingPacket.wasProcessedBy(ID)) {

//            BareJID fromBareJID = getFrom(session);
//            String from = fromBareJID != null ? fromBareJID.getLocalpart() : null;
//            String fromDomain = fromBareJID != null ? fromBareJID.getDomain() : null;
//
//            BareJID toBareJID = getRecipient(session, packet);
//            String to = toBareJID != null ? toBareJID.getLocalpart() : null;
//            String toDomain = toBareJID != null ? toBareJID.getDomain() : null;


            //send to access policy null values as well and let it to decide
            //should we read 'from' from authorized session? 
            JID stanzaFrom = incomingPacket.getStanzaFrom();
            String from = stanzaFrom != null ? stanzaFrom.getLocalpart() : null;
            String fromDomain = stanzaFrom != null ? stanzaFrom.getDomain() : null;

            JID stanzaTo = incomingPacket.getStanzaTo();
            String to = stanzaTo != null ? stanzaTo.getLocalpart() : null;
            String toDomain = stanzaTo != null ? stanzaTo.getDomain() : null;
            
            PacketType type = PacketResolver.resolve(incomingPacket);

            if (!this.accessPolicy.outgoing(to, toDomain, from, fromDomain, type)) {
                deny(incomingPacket, results, this.accessPolicy.outgoingResponse(), type);
                result = DENY;
            } else if (!this.accessPolicy.incoming(from, fromDomain, to, toDomain, type)) {
                deny(incomingPacket, results, this.accessPolicy.incomingResponse(), type);
                result = DENY;
            }
        }
        return result;
    }

    @Override
    public String[] supElements() {
        return ELEMENTS;
    }

    @Override
    public String[] supNamespaces() {
        return XMLNSS;
    }
}
