/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.plugin;

import com.kerio.im.bean.VCard;
import com.kerio.im.core.api.CoreService;
import com.kerio.im.core.api.event.StartupFinishedEvent;
import com.kerio.im.core.api.service.PacketType;
import com.kerio.im.core.api.service.ServiceBus;
import com.kerio.im.core.api.service.VCardService;
import com.kerio.im.tigase.ClientResolver;
import com.kerio.im.tigase.PacketResolver;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.db.NonAuthUserRepository;
import tigase.db.TigaseDBException;
import tigase.server.Packet;
import tigase.xml.Element;
import tigase.xml.XMLNodeIfc;
import tigase.xmpp.JID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPPreprocessorIfc;
import tigase.xmpp.XMPPProcessor;
import tigase.xmpp.XMPPResourceConnection;

/**
 *
 * @author tskoda
 */
public class VCardFilterPlugin extends XMPPProcessor implements XMPPPreprocessorIfc {

    private static final Logger LOG = Logger.getLogger(DomainFilterPlugin.class.getName());
    public static final String PLUGIN_ID = "connect-vcard-fix";
    private static final String[] ELEMENTS = {"x"};
    private static final String[] XMLNS = {"vcard-temp:x:update"};
    private static final boolean PASS = false;
    private static final String PHOTO = "photo";
    private VCardService<VCardWrapper> vcardService;

    public VCardFilterPlugin() {
    }

    @Override
    public String id() {
        return PLUGIN_ID;
    }

    @Override
    public String[] supElements() {
        return ELEMENTS;
    }

    @Override
    public String[] supNamespaces() {
        return XMLNS;
    }

    @Override
    public void init(Map<String, Object> settings) throws TigaseDBException {
        super.init(settings);
        CoreService coreService = init(CoreService.class.getName(), settings);
        coreService.bus().subscribe(StartupFinishedEvent.class, new VCardFilterPlugin.StartupListener(coreService));
    }

    @Override
    public boolean preProcess(Packet packet, XMPPResourceConnection session, NonAuthUserRepository repo, Queue<Packet> results, Map<String, Object> settings) {
        JID from;
        if (iChatPacket(session) && isPresence(packet) && ((from = packet.getStanzaFrom()) != null) && isOnBehalfOnSender(from, session)) {
            Element xUpdate;
            if ((xUpdate = xUpdate(packet.getElemChildren(PacketResolver.PRESENCE))) != null) {
                if (initialized()) {
                    VCard vcard;
                    if (((vcard = this.vcardService.get(from.getLocalpart(), from.getDomain())) != null)) {
                        List<XMLNodeIfc> photos = new ArrayList<XMLNodeIfc>(1);
                        if (vcard.photoSHA1() != null) {    
                            photos.add(new Element(PHOTO, vcard.photoSHA1()));   
                        } else {
                            photos.add(new Element(PHOTO));   
                        }
                        xUpdate.setChildren(photos);
                        packet.wasProcessedBy(PLUGIN_ID);
                    }
                } else {
                    LOG.log(Level.FINE, "Presence packet update failed, vcard filter not initialized yet");
                }
            }
        }

        return PASS;
    }

    private <T> T init(String id, Map<String, Object> settings) throws TigaseDBException {
        Object vcardManagerObject = settings.get(id);
        if (vcardManagerObject != null) {
            return (T) vcardManagerObject;
        } else {
            throw new TigaseDBException("VCardPlugin initialization failed, " + id + " not found");
        }
    }

    private Element xUpdate(List<Element> xElements) {
        Element result = null;
        if (xElements != null) {
            for (Element xElement : xElements) {
                if (ELEMENTS[0].equals(xElement.getName()) && XMLNS[0].equals(xElement.getXMLNS())) {
                    result = xElement;
                    break;
                }
            }
        }
        return result;
    }

    private boolean isOnBehalfOnSender(JID from, XMPPResourceConnection session) {
        boolean result;
        try {
            result = session != null && session.isUserId(from.getBareJID());
        } catch (NotAuthorizedException e) {
            result = false;
        }

        return result;
    }

    private boolean isPresence(Packet packet) {
        return PacketResolver.resolve(packet) == PacketType.PRESENCE;
    }

    private boolean iChatPacket(XMPPResourceConnection session) {
        boolean result = false;
        Object clientTypeObject;
        if ((session != null) && ((clientTypeObject = session.getSessionData(ClientResolver.CLIENT_TYPE)) != null)) {
            result = ((ClientResolver.ClientType) clientTypeObject) == ClientResolver.ClientType.ichat;
        }
        return result;
    }

    private boolean initialized() {
        synchronized (this) {
            return this.vcardService != null && this.vcardService.enabled();
        }
    }

    private class StartupListener implements ServiceBus.EventListener<StartupFinishedEvent> {

        private final CoreService coreService;

        public StartupListener(CoreService coreService) {
            this.coreService = coreService;
        }

        @Override
        public void event(StartupFinishedEvent event) {
            synchronized (VCardFilterPlugin.this) {
                VCardFilterPlugin.this.vcardService = coreService.vcards();
            }
        }
    }
}
