/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.tigase.Conversion;
import com.kerio.im.xmpp.api.XmppException;
import com.kerio.im.xmpp.api.XmppUsers;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.xmpp.BareJID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.impl.JabberIqRoster;
import tigase.xmpp.impl.roster.RosterAbstract;
import static tigase.xmpp.impl.roster.RosterAbstract.ROSTERHASH;
import tigase.xmpp.impl.roster.RosterElement;
import tigase.xmpp.impl.roster.RosterFactory;

/**
 *
 * @author tskoda
 */
//not that new user connections are blocked by auth method
public class DefaultRosterListener implements IDefaultRosterListener {

    private static final Logger LOG = Logger.getLogger(DefaultRosterListener.class.getName());

    class AddedFilter implements RosterNotify.Filter<PacketBuilder.AddUserPacket> {

        @Override
        public PacketBuilder.AddUserPacket pass(BareJID recipient, PacketBuilder.AddUserPacket packet) {
            //do not nofify as there is no change for this domainUser in his roster
            if (recipient.equals(packet.getAdded()) || DefaultRosterListener.this.personalRoster.of(recipient).managed(packet.getAdded())) {
                return null; //block
            } else {
                return packet; //pass
            }
        }
    }

    class RemovedFilter implements RosterNotify.Filter<PacketBuilder.RemoveUserPacket> {

        @Override
        public PacketBuilder.RemoveUserPacket pass(BareJID recipient, PacketBuilder.RemoveUserPacket packet) {
            //do not nofify as there is no change for this domainUser in his roster
            if (recipient.equals(packet.getRemoved()) || DefaultRosterListener.this.personalRoster.of(recipient).managed(packet.getRemoved())) {
                return null; //block
            } else {
                return packet; //pass
            }
        }
    }

    class UpdateFilter implements RosterNotify.Filter<PacketBuilder.AddUserPacket> {

        @Override
        public PacketBuilder.AddUserPacket pass(BareJID recipient, PacketBuilder.AddUserPacket packet) {

            if (recipient.equals(packet.getAdded())) {
                packet = null;
            } else {
                //we should add personal name or personal groups if such exists    
                PersonalRoster.Buddy rosterUser = DefaultRosterListener.this.personalRoster.of(recipient).find(packet.getAdded());
                if (rosterUser != null) {
                    if (rosterUser.isDeleted() || rosterUser.isAdded()) {
                        packet = null;
                    } else {
                        packet = packet.merge(rosterUser);
                    }
                } //otherwise we pass the default value
            }
            return packet;
        }
    }

    class RosterHashCache implements RosterNotify.HashProvider {

        private final RosterAbstract rosterControl;

        private BareJID lastUser;
        private String lastHash;

        public RosterHashCache() {
            this.rosterControl = RosterFactory.getRosterImplementation(true);
        }

        /*
         * This method could be called multiple times for single barejid (in one row). 
         */
        @Override
        public String calculate(XMPPResourceConnection connection) {
            String result = null;
            try {
                BareJID user = connection.getBareJID();
                if ((lastUser == null) || (!lastUser.equals(user))) {
                    try {
                        lastUser = user;
                        String roster = JabberIqRoster.calculateRoster(
                                rosterControl,
                                connection,
                                defaultRoster.roster(user.getDomain()).asList(user, personalRoster.of(user).asMap()));

                        //update only once for single user across all his resources
                        rosterControl.updateRosterHash(roster, connection);

                        //read computed roster hash
                        lastHash = (String) connection.getCommonSessionData(ROSTERHASH);

                    } catch (Exception e) {
                        LOG.log(Level.INFO, "Calculation of roster hash failed", e);
                        lastUser = null;
                        lastHash = null;
                    }
                }

                result = lastHash;
            } catch (NotAuthorizedException e) {
                LOG.log(Level.INFO, "Calculation of roster hash failed, user is not autorized yet");
            }

            return result;
        }

    }

    private final DefaultRoster defaultRoster;
    private final PersonalRoster personalRoster;
    private final RosterNotify notify;
    private final GroupService groups;

    public DefaultRosterListener(DefaultRoster defaultRoster, PersonalRoster personalRoster, RosterNotify notify, GroupService groups) {
        this.defaultRoster = defaultRoster;
        this.personalRoster = personalRoster;
        this.notify = notify;
        this.groups = groups;
    }

    @Override
    public void add(User userToAdd) {

        //only if user is enabled and visible
        if (userToAdd.isEnabled() && userToAdd.isVisible()) {
            //only if user not existed yet
            if (this.defaultRoster.roster(userToAdd.getDomain()).add(userToAdd)) {
                notify.aboutAdd(
                        userToAdd,
                        this.defaultRoster.roster(userToAdd.getDomain()).all(),
                        new AddedFilter(),
                        new RosterHashCache());
            }
        }
    }

    @Override
    public void remove(User userToRemove) {
        //we do not markRemoved userToRemove from personal rosters
        if (this.defaultRoster.roster(userToRemove.getDomain()).remove(Conversion.toBareJID(userToRemove))) {
            this.notify.aboutRemove(
                    userToRemove,
                    this.defaultRoster.roster(userToRemove.getDomain()).all(),
                    new RemovedFilter(),
                    new RosterHashCache());
        }
    }

    @Override
    public void update(final User user, final EnumSet<XmppUsers.Changes> changes) {
        //invisible -> visible && enabled and he is online we should pass updates of all users to him if he is online
        //invisible -> visible && disabled - cant be online
        //disabled - > enabled && visible || invisible - this user werent online, filled with default once he become online
        BareJID userJID = Conversion.toBareJID(user);

        if (user.isEnabled()) {
            if (user.isVisible()) {
                //user is enabled and visible
                this.defaultRoster.roster(user.getDomain()).add(user); //replaces existing value

                //user is enabled and visible and visibility was changed
                if (changes.contains(XmppUsers.Changes.VISIBILITY)) {
                    //add default contacts to updated user roster
                    this.notify.aboutAdd(
                            this.defaultRoster.roster(user.getDomain()).asList(userJID, this.personalRoster.of(userJID).asMap()),
                            userJID,
                            new RosterHashCache());
                }

                //broadcast user update to all other
                this.notify.aboutAdd(
                        user,
                        this.defaultRoster.roster(user.getDomain()).all(),
                        new UpdateFilter(),
                        new RosterHashCache());

            } else {
                //remove from default roster
                this.defaultRoster.roster(user.getDomain()).remove(Conversion.toBareJID(user));

                //user is enabled and invisible and visibility was changed
                if (changes.contains(XmppUsers.Changes.VISIBILITY)) {
                    //remove all default contacts from updated user roster
                    this.notify.aboutRemove(
                            this.defaultRoster.roster(user.getDomain()).asArray(userJID, this.personalRoster.of(userJID).asMap()),
                            userJID,
                            new RosterHashCache());
                }

                //broadcast this change to all other
                this.notify.aboutRemove(
                        user,
                        this.defaultRoster.roster(user.getDomain()).all(),
                        new RemovedFilter(),
                        new RosterHashCache());
            }
        } else {
            //remove from default roster
            this.defaultRoster.roster(user.getDomain()).remove(Conversion.toBareJID(user));

            //broadcast this change to all other
            this.notify.aboutRemove(
                    user,
                    this.defaultRoster.roster(user.getDomain()).all(),
                    new RemovedFilter(),
                    new RosterHashCache());
        }
    }

    //we dont need to repair other rosters as they are not modified (if user B delets A from its roster, B is not deleted from A roster)
    //returns true if roster was modified
    @Override
    public void repair(User user, Map<BareJID, RosterElement> roster) {
        // read personal roster and remove all records
        try {
            this.personalRoster.of(Conversion.toBareJID(user)).clear();
        } catch (XmppException e) {
            LOG.log(Level.INFO, "Clearing modifications of " + user.getPrimaryEmail() + "'s roster failed", e);
        }

        // read real user roster - delete all contact from its domain
        for (Iterator<BareJID> it = roster.keySet().iterator(); it.hasNext();) {
            BareJID record = it.next();

            if (user.getDomain().equals(record.getDomain())) {
                it.remove();
            }
        }

        LOG.log(Level.INFO, "Clearing modifications of {0}''s roster successful, automatic content is applied now", user.getPrimaryEmail());
    }

    //called right after login
    @Override
    public void check(User user, Map<BareJID, RosterElement> roster) {
        BareJID ownerJID = Conversion.toBareJID(user);

        int total = roster.size();
        int personal = 0;
        int dynamic = 0;
        int marked = 0;
        int skipped = 0;

        RosterContent defaultContent = this.defaultRoster.roster(ownerJID.getDomain());
        boolean ownerFound = defaultContent.find(ownerJID) != null;

        for (Iterator<Map.Entry<BareJID, RosterElement>> it = roster.entrySet().iterator(); it.hasNext();) {
            Map.Entry<BareJID, RosterElement> record = it.next();

            //we found contact in default roster    
            RosterContent.Buddy defaultBuddy = defaultContent.find(record.getKey());
            PersonalRoster.Buddy personalBuddy = personalRoster.of(ownerJID).find(record.getKey());
            if ((ownerFound) && (defaultBuddy != null)) {
                //if owner found (is visible) and buddy found (is visible) and buddy is not added manually

                if (personalBuddy == null || !personalBuddy.isAdded()) {
                    String nameToUpdate = null;
                    if ((record.getValue().getName() != null) && (!record.getValue().getName().equals(defaultBuddy.getUser().getFullName()))) {
                        nameToUpdate = record.getValue().getName();
                    }

                    String[] groupsToUpdate = null;
                    if (record.getValue().getGroups() == null) {
                        groupsToUpdate = new String[]{null};
                    } else if (!Arrays.equals(record.getValue().getGroups(), new String[]{this.groups.group(record.getKey().toString())})) {
                        groupsToUpdate = record.getValue().getGroups();
                    }

                    if (nameToUpdate != null || groupsToUpdate != null) {
                        //if there are changes against default
                        this.personalRoster.of(ownerJID).update(record.getKey(), nameToUpdate, groupsToUpdate);
                        personal++;
                    } else {
                        this.personalRoster.of(ownerJID).remove(record.getKey());
                        dynamic++;
                    }

                    it.remove();
                } else {
                    //we keep it in static roster and keep added flag in personal
                }
            } else {
                //only if the same doman
                if (ownerJID.getDomain().equals(record.getKey().getDomain())) {
                    //if not in default group mark as added and leave it in static roster
                    this.personalRoster.of(ownerJID).markAdded(record.getKey());
                    marked++;
                } else {
                    skipped++;
                }

            }
        }

        LOG.log(Level.INFO, "Merging {0} contacts from {1}''s roster successful, {2} identical to dynamic, {3} changed saved in personal, {4} marked as added, {5} skipped", new Object[]{total, user.getPrimaryEmail(), dynamic, personal, marked, skipped});
    }

    public void bye() {

    }
}
