/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.tigase.SessionProvider;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;
import tigase.server.Packet;
import tigase.xml.Element;
import tigase.xml.XMLUtils;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.NoConnectionIdException;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.StanzaType;
import tigase.xmpp.impl.roster.RosterAbstract;

/**
 *
 * @author tskoda
 */
public class PacketBuilder {

    private static final Logger LOG = Logger.getLogger(PacketBuilder.class.getName());

    public interface ChangePacket {

        Packet toPacket(String nextStanzaId, final SessionProvider.UserResource conn, final SessionProvider.UserData userData, final String hash) throws NotAuthorizedException, NoConnectionIdException;
    }

    public static abstract class MultiChangePacket implements ChangePacket {

        abstract List<Element> toElements();

        @Override
        public Packet toPacket(String nextStanzaId, final SessionProvider.UserResource conn, final SessionProvider.UserData userData, final String hash) throws NotAuthorizedException, NoConnectionIdException {

            Element update = new Element("iq");
            update.setXMLNS(RosterAbstract.CLIENT_XMLNS);
            update.setAttribute("type", StanzaType.set.toString());

            Element query = new Element("query");
            query.setXMLNS(RosterAbstract.ROSTER_XMLNS);

            if (hash != null) {
                query.addAttribute(RosterAbstract.VER_ATT, hash);
            }

            query.addChildren(toElements());

            update.addChild(query);
            update.setAttribute("to", conn.getBareJID().toString());
            update.setAttribute("id", "rst" + nextStanzaId);

            Packet result = Packet.packetInstance(update, null, conn.getJID());
            result.setPacketTo(conn.getConnectionId());
            return result;
        }
    }

    public static abstract class SingleChangePacket implements ChangePacket {

        abstract Element toElement();

        @Override
        public Packet toPacket(String nextStanzaId, final SessionProvider.UserResource conn, final SessionProvider.UserData userData, final String hash) throws NotAuthorizedException, NoConnectionIdException {

            Element update = new Element("iq");
            update.setXMLNS(RosterAbstract.CLIENT_XMLNS);
            update.setAttribute("type", StanzaType.set.toString());

            Element query = new Element("query");
            query.setXMLNS(RosterAbstract.ROSTER_XMLNS);

            if (hash != null) {
                query.addAttribute(RosterAbstract.VER_ATT, hash);
            }

            query.addChild(toElement());

            update.addChild(query);
            update.setAttribute("to", conn.getBareJID().toString());
            update.setAttribute("id", "rst" + nextStanzaId);

            Packet result = Packet.packetInstance(update, null, conn.getJID());
            result.setPacketTo(conn.getConnectionId());
            return result;
        }
    }

    public static class UsersMultiPacket extends MultiChangePacket {

        private final List<Element> elements;

        public UsersMultiPacket(List<Element> elements) {
            this.elements = elements;
        }

        public void put(SingleChangePacket packet) {
            this.elements.add(packet.toElement());
        }

        @Override
        public List<Element> toElements() {
            return this.elements;
        }
    }

    public static class AddUserPacket extends SingleChangePacket {

        private final BareJID added;
        private final RosterAbstract.SubscriptionType subscription;
        private final String[] groups;
        private final String name;

        public AddUserPacket(BareJID added, RosterAbstract.SubscriptionType subscription, String name, String[] groups) {
            this.added = added;
            this.subscription = subscription;
            this.name = name;
            this.groups = groups;
        }

        public BareJID getAdded() {
            return added;
        }

        public AddUserPacket merge(PersonalRoster.Buddy rosterUser) {
            AddUserPacket result = this;
            if (rosterUser.isNameManaged()) {
                result = result.setName(rosterUser.getName());
            }
            if (rosterUser.isGroupsManaged()) {
                result = result.setGroups(rosterUser.getGroups());
            }
            return result;
        }

        public AddUserPacket setGroups(String[] groups) {
            return new AddUserPacket(this.added, this.subscription, this.name, groups);
        }

        public AddUserPacket setName(String name) {
            return new AddUserPacket(this.added, this.subscription, name, this.groups);
        }

        @Override
        Element toElement() {

            Element item = new Element("item");
            item.setAttribute("jid", this.added.toString());
            item.setAttribute("subscription", this.subscription.name());
            item.setAttribute("name", XMLUtils.escape(this.name));

            for (String gr : this.groups) {
                if (gr != null) {
                    Element group = new Element("group");
                    group.setCData(XMLUtils.escape(gr));
                    item.addChild(group);
                }
            }

            return item;
        }

        @Override
        public int hashCode() {
            int hash = 5;
            hash = 19 * hash + (this.added != null ? this.added.hashCode() : 0);
            hash = 19 * hash + (this.subscription != null ? this.subscription.hashCode() : 0);
            hash = 19 * hash + Arrays.deepHashCode(this.groups);
            hash = 19 * hash + (this.name != null ? this.name.hashCode() : 0);
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final AddUserPacket other = (AddUserPacket) obj;
            if (this.added != other.added && (this.added == null || !this.added.equals(other.added))) {
                return false;
            }
            if (this.subscription != other.subscription) {
                return false;
            }
            if (!Arrays.deepEquals(this.groups, other.groups)) {
                return false;
            }
            if ((this.name == null) ? (other.name != null) : !this.name.equals(other.name)) {
                return false;
            }
            return true;
        }

    }

    public static class RemoveUserPacket extends SingleChangePacket {

        private final BareJID removed;

        public RemoveUserPacket(BareJID removed) {
            this.removed = removed;
        }

        public BareJID getRemoved() {
            return removed;
        }

        @Override
        Element toElement() {

            Element item = new Element("item");
            item.setAttribute("jid", removed.toString());
            item.setAttribute("subscription", "remove");

            return item;
        }

        @Override
        public int hashCode() {
            int hash = 3;
            hash = 37 * hash + (this.removed != null ? this.removed.hashCode() : 0);
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final RemoveUserPacket other = (RemoveUserPacket) obj;
            if (this.removed != other.removed && (this.removed == null || !this.removed.equals(other.removed))) {
                return false;
            }
            return true;
        }

    }

    public PacketBuilder() {
    }

    public UsersMultiPacket added(List<Element> elements) {
        return new UsersMultiPacket(elements);
    }

    public UsersMultiPacket removed(JID[] jids) {
        //empty list at the beginning
        UsersMultiPacket result = new UsersMultiPacket(new LinkedList<Element>());
        for (JID jid : jids) {
            result.put(new RemoveUserPacket(jid.getBareJID()));
        }
        return result;
    }

    public AddUserPacket added(BareJID added, RosterAbstract.SubscriptionType subscription, String name, String[] groups) {
        return new AddUserPacket(added, subscription, name, groups);
    }

    public RemoveUserPacket removed(BareJID removed) {
        return new RemoveUserPacket(removed);
    }

    public AddUserPacket renamed(BareJID added, RosterAbstract.SubscriptionType subscription, String name, String[] groups) {
        return new AddUserPacket(added, subscription, name, groups);
    }

}
