/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.google.common.base.Joiner;
import com.google.common.base.Splitter;
import com.google.common.collect.Iterables;
import com.kerio.im.tigase.user.XmppUsersStore;
import com.kerio.im.tigase.user.XmppUsersStorePool;
import com.kerio.im.xmpp.api.XmppException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import tigase.xmpp.BareJID;

/**
 *
 * @author tskoda
 */
public class PersonalRoster {

    private static final Logger LOG = Logger.getLogger(PersonalRoster.class.getName());
    
    public static class Buddy {
        
        public static String groups(String[] groups) {
            if ((groups == null) || (groups.length == 0)) {
                return null;
            } else {
                return Joiner.on(",").useForNull("null").join(groups);
            }
        }
        
        public static String[] groups(String groups) {
            if ((groups == null) || (groups.isEmpty())) {
                return null;
            } else {
                String[] result = groups.split(",");
                for (int i = 0; i < result.length; i++) {
                    if ("null".equals(result[i])) {
                        result[i] = null;
                    }
                }
                return result;
            }
        }
        
        public enum FLAG {
            updated,
            deleted,
            added
        }

        //those not managed must be null
        private final BareJID buddy;
        private final String name;
        private final String[] groups;
        private final FLAG flag;

        public Buddy(BareJID buddy, String name, String[] groups, FLAG flag) {
            this.buddy = buddy;
            this.name = name;
            this.groups = groups;
            this.flag = flag;
        }
      
        public boolean isDeleted() {
            return this.flag == FLAG.deleted;
        }

        public boolean isAdded() {
            return this.flag == FLAG.added;
        }

        public BareJID getBuddy() {
            return buddy;
        }

        public String getName() {
            return name;
        }

        public boolean isNameManaged() {
            return this.name != null && !this.name.isEmpty();
        }

        public String[] getGroups() {
            return groups;
        }

        public FLAG getFlag() {
            return flag;
        }
        
        public boolean isGroupsManaged() {
            return this.groups != null;
        }

        @Override
        public int hashCode() {
            int hash = 5;
            hash = 71 * hash + (this.buddy != null ? this.buddy.hashCode() : 0);
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final Buddy other = (Buddy) obj;
            if (this.buddy != other.buddy && (this.buddy == null || !this.buddy.equals(other.buddy))) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return "Buddy{" + "buddy=" + buddy + ", name=" + name + ", groups=" + Arrays.toString(groups) + ", flag=" + flag + '}';
        }

    }

    public static class PersonalUserRoster {

        private final BareJID owner;
        private final Map<BareJID, Buddy> content;
        private final XmppUsersStore store;

        public PersonalUserRoster(BareJID owner, XmppUsersStore store, Map<BareJID, Buddy> content) {
            this.owner = owner;
            this.store = store;
            this.content = content;           
        }
        
        public int size() {
            return this.content.size();
        }

        public Map<BareJID, Buddy> asMap() {
            return new HashMap<BareJID, Buddy>(this.content);
        }

        public synchronized void markRemoved(BareJID buddyJID) {
            if (this.owner.equals(buddyJID)) {
                return;
            }
            Buddy buddy = this.content.get(buddyJID);
            Buddy updatedBuddy;

            if (buddy != null) {
                updatedBuddy = new Buddy(buddyJID, buddy.getName(), buddy.getGroups(), Buddy.FLAG.deleted);
            } else {
                updatedBuddy = new Buddy(buddyJID, null, null, Buddy.FLAG.deleted);
            }

            this.store.personalUpdate(this.owner, updatedBuddy); //update cache only if success
            this.content.put(buddyJID, updatedBuddy);
        }

        public synchronized void clear() {
            this.store.personalRemove(this.owner); //update cache only if success
            this.content.clear();
        }

        //returns true if owner of this roster has manually added/removed the user passed as parameter
        //it should return false when user just changed name or group
        public synchronized boolean managed(BareJID buddy) {
            boolean result = false;

            Buddy buddyData = this.content.get(buddy);
            if (buddyData != null) {
                result = buddyData.isAdded() || buddyData.isDeleted();
            }

            return result;
        }

        public synchronized Buddy find(BareJID buddy) {
            return this.content.get(buddy);
        }

        public synchronized void update(BareJID buddyJID, String nameToUpdate, String[] groupsToUpdate) {
            if (this.owner.equals(buddyJID)) {
                return;
            }

            String[] groupsToUpdateList = groupsToUpdate != null && groupsToUpdate.length > 0? groupsToUpdate : null;

            Buddy updatedBuddy = new Buddy(buddyJID, nameToUpdate, groupsToUpdateList, Buddy.FLAG.updated);

            this.store.personalUpdate(this.owner, updatedBuddy); //update cache only if success
            this.content.put(buddyJID, updatedBuddy);
        }

        public synchronized void remove(BareJID buddyJID) {
            this.store.personalRemove(this.owner, buddyJID); //update cache only if success
            this.content.remove(buddyJID);
        }

        public synchronized void markAdded(BareJID buddyJID) {
            if (this.owner.equals(buddyJID)) {
                return;
            }

            Buddy buddy = this.content.get(buddyJID);
            Buddy updatedBuddy;

            if (buddy != null) {
                updatedBuddy = new Buddy(buddyJID, buddy.getName(), buddy.getGroups(), Buddy.FLAG.added);
            } else {
                updatedBuddy = new Buddy(buddyJID, null, null, Buddy.FLAG.added);
            }

            this.store.personalUpdate(this.owner, updatedBuddy); //update cache only if success
            this.content.put(buddyJID, updatedBuddy);
        }
    }

    private final Map<BareJID, PersonalUserRoster> rosterMap;
    private final XmppUsersStorePool storePool;

    public PersonalRoster(XmppUsersStorePool storePool) {
        this(new HashMap<BareJID, PersonalUserRoster>(), storePool);
    }

    PersonalRoster(Map<BareJID, PersonalUserRoster> map, XmppUsersStorePool storePool) {
        this.rosterMap = map;
        this.storePool = storePool;
    }

    //never returns null
    public PersonalUserRoster of(BareJID owner) {
        PersonalUserRoster roster = this.rosterMap.get(owner);
        if (roster == null) {
            synchronized (this.rosterMap) {
                if ((roster = this.rosterMap.get(owner)) == null) {
                    XmppUsersStore store = this.storePool.get(owner.getDomain());
                    //try delete all on error, log it and start with "reseted roster"
                    try {
                        roster = new PersonalUserRoster(owner, store, store.personalAll(owner));
                        LOG.log(Level.INFO, "Reading modifications of {0}''s automatic roster successful, found {1} modifications", new Object[]{owner, roster.size()});
                    } catch (XmppException eAll) {
                        LOG.log(Level.WARNING, "Reading modifications of {0}''s automatic roster failed, reseting to default. Reason: {1}", new Object[]{owner, eAll.getMessage()});
                        LOG.log(Level.INFO, "Reading modifications of " + owner + "'s automatic roster failed, reseting to default. Reason: ", eAll);
                        
                        try {
                            store.personalRemove(owner);
                        } catch (XmppException eRemove) {
                            LOG.log(Level.INFO, "Removing all modifications of " + owner + "'s automatic roster failed", eRemove);
                        }
                        
                        roster = new PersonalUserRoster(owner, store, new HashMap<BareJID, Buddy>());
                    }
                    this.rosterMap.put(owner, roster);
                }
            }
        }
        
        return roster;
    }

}
