/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.bean.User;
import com.kerio.im.tigase.Conversion;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.xml.Element;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;

/**
 *
 * @author tskoda
 */
public class RosterContent {

    private static final Logger LOG = Logger.getLogger(RosterContent.class.getName());

    public static void warnAboutMaximumReached(int maximum, String domain) {
        LOG.log(
                Level.WARNING,
                "Maximum size ({0}) of the Instant Messaging contact list in domain {1} reached. Users exceeding the limit will not have the automatic contact list available and they will not be populated to the automatic contact lists of other users. See http://kb.kerio.com/1403 for details.",
                new Object[]{maximum, domain});
    }

    public static class Buddy {

        private final Element element;
        private final User user;

        public Buddy(Element element, User user) {
            this.element = element;
            this.user = user;
        }

        public Element getElement() {
            return element;
        }

        public User getUser() {
            return user;
        }

        @Override
        public int hashCode() {
            int hash = 5;
            hash = 59 * hash + (this.user != null ? this.user.hashCode() : 0);
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final Buddy other = (Buddy) obj;
            if (this.user != other.user && (this.user == null || !this.user.equals(other.user))) {
                return false;
            }
            return true;
        }

        @Override
        public String toString() {
            return "Buddy{" + "user=" + user + '}';
        }
    }

    private final Map<BareJID, Buddy> content;
    private final ElementMapper mapper;
    private final int maximum;
    private boolean maximumAnnounced;

    public RosterContent(Map<BareJID, Buddy> content, ElementMapper elementMapper, int maximum, boolean maximumAnnounced) {
        //TODO content could hold more than maximum!
        this.maximum = maximum;
        this.mapper = elementMapper;
        this.content = content;
        this.maximumAnnounced = maximumAnnounced;
    }

    public Buddy find(BareJID bareJID) {
        return this.content.get(bareJID);
    }

    public boolean add(User user) {
        boolean result = false;
        if (this.content.size() <= this.maximum) { //'<=' because roster size = content.size - roster owner

            if (user != null) {
                result = this.content.put(Conversion.toBareJID(user), this.mapper.value(user)) == null;
            }
        } else {
            LOG.log(
                    Level.INFO,
                    "Adding new user to {0} dynamic roster skipped, maximum size ({1}) of Instant Messaging contact list reached",
                    new Object[]{user.getDomain(), this.maximum});

            if (!maximumAnnounced) {
                warnAboutMaximumReached(maximum, user.getDomain());
                maximumAnnounced = true;
            }
        }

        return result;
    }

    /**
     * Returns true if user was removed from roster content.
     *
     * @param userJID user to remove
     * @return true if roster content contained this user, otherwise false
     */
    public boolean remove(BareJID userJID) {
        return this.content.remove(userJID) != null;
    }

    public List<? extends BareJID> all() {
        return new LinkedList<BareJID>(this.content.keySet());
    }

    public JID[] asArray(BareJID owner, Map<BareJID, PersonalRoster.Buddy> personal) {
        Set<BareJID> copy = new HashSet<BareJID>(this.content.keySet());
        copy.remove(owner);

        if (personal != null) {
            for (Map.Entry<BareJID, PersonalRoster.Buddy> personalEntry : personal.entrySet()) {
                if (personalEntry.getValue().getFlag() != PersonalRoster.Buddy.FLAG.updated) {
                    copy.remove(personalEntry.getKey());
                }
            }
        }

        JID[] result = new JID[copy.size()];
        int index = 0;

        for (BareJID bJID : copy) {
            result[index++] = JID.jidInstance(bJID);
        }

        return result;
    }

    public List<Element> asList(BareJID owner, Map<BareJID, PersonalRoster.Buddy> personal) {
        Map<BareJID, Buddy> copy = new HashMap<BareJID, Buddy>(this.content);
        copy.remove(owner);

        LinkedList<Element> result = new LinkedList<Element>();
        for (Map.Entry<BareJID, Buddy> entry : copy.entrySet()) {
            if (personal != null) {
                PersonalRoster.Buddy personalBuddy = personal.get(entry.getKey());

                if (personalBuddy != null) {
                    //ignore those added or removed
                    if (personalBuddy.getFlag() == PersonalRoster.Buddy.FLAG.updated) {
                        result.add(this.mapper.value(entry.getValue().getUser(), personalBuddy));
                    } else {
                        //do not add it to the result list
                    }
                } else {
                    //add as is
                    result.add(entry.getValue().getElement());
                }
            } else {
                result.add(entry.getValue().getElement());
            }
        }
        return result;
    }

    public int size() {
        return this.content.size();
    }

}
