/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.tigase.Conversion;
import com.kerio.im.tigase.PacketSender;
import com.kerio.im.tigase.SessionProvider;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;
import tigase.server.Packet;
import tigase.xml.Element;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.impl.roster.RosterAbstract;

/**
 *
 * @author tskoda
 */
public class RosterNotify {

    private static final Logger LOG = Logger.getLogger(RosterNotify.class.getName());

    public interface Filter<T extends PacketBuilder.SingleChangePacket> {

        //if null then stop, if not null then pass, returned instance could be modified packet or the original one
        T pass(BareJID recipient, T packet);
    }
    
    public interface HashProvider {
        /**
         * 
         * @param connection
         * @return null on error, or hash of roster
         */
        String calculate(XMPPResourceConnection connection);
    }

    private final GroupService groups;
    private final PacketBuilder packetBuilder;
    private final SessionProvider sessions;
    private final PacketSender packetSender;

    public RosterNotify(GroupService groups, PacketBuilder packetBuilder, SessionProvider sessions, PacketSender packetSender) {
        this.groups = groups;
        this.packetBuilder = packetBuilder;
        this.sessions = sessions;
        this.packetSender = packetSender;
    }

    public PacketBuilder.AddUserPacket added(User user) {
        return this.packetBuilder.added(
                Conversion.toBareJID(user),
                RosterAbstract.SubscriptionType.both,
                user.getFullName(),
                new String[]{(groups.group(user.getPrimaryEmail()))});
    }

    public void aboutAdd(List<Element> rosterElements, BareJID recipient, HashProvider hashProvider) {
        PacketBuilder.MultiChangePacket packet = this.packetBuilder.added(rosterElements);
        if (isUserOnline(recipient)) {
            notifyAllResources(recipient, packet, hashProvider);
        }
    }

    public void aboutRemove(JID[] rosterElements, BareJID recipient, HashProvider hashProvider) {
        PacketBuilder.MultiChangePacket packet = this.packetBuilder.removed(rosterElements);
        if (isUserOnline(recipient)) {
            notifyAllResources(recipient, packet, hashProvider);
        }
    }

    public void aboutAdd(User userToAdd, Collection<? extends BareJID> recipients, Filter<PacketBuilder.AddUserPacket> filter, HashProvider hashProvider) {
        PacketBuilder.AddUserPacket defaultPacket = added(userToAdd);

        for (BareJID recipient : recipients) {
            if (isUserOnline(recipient)) {
                PacketBuilder.AddUserPacket addPacket = filter.pass(recipient, defaultPacket);
                if (addPacket != null) {
                    notifyAllResources(recipient, addPacket, hashProvider);
                }
            }
        }
    }

    public void aboutRemove(User userToRemove, Collection<? extends BareJID> recipients, Filter<PacketBuilder.RemoveUserPacket> filter, HashProvider hashProvider) {
        BareJID removedUserJID = Conversion.toBareJID(userToRemove);
        PacketBuilder.RemoveUserPacket defaultPacket = this.packetBuilder.removed(removedUserJID);

        for (BareJID recipient : recipients) {
            if (isUserOnline(recipient)) {
                PacketBuilder.RemoveUserPacket removePacket = filter.pass(recipient, defaultPacket);
                if (removePacket != null) {
                    notifyAllResources(recipient, removePacket, hashProvider);
                }
            }
        }
    }

    private void notifyAllResources(BareJID user, PacketBuilder.ChangePacket lazyPacket, HashProvider hashProvider) {

        Queue<Packet> outPackets = new LinkedList<Packet>();

        try {
            for (SessionProvider.UserResource resource : this.sessions.getUserResources(user)) {
                if (resource.isAuthorized()) {
                    try {
                        outPackets.offer(lazyPacket.toPacket(resource.nextStanzaId(), resource, this.sessions.getUserData(user), hashProvider.calculate(resource.raw())));
                    } catch (tigase.xmpp.XMPPException e) {
                        LOG.log(Level.FINE, "Attemped to update roster on not yet authorized connection", e);
                    }
                }
            }
        } catch (NotAuthorizedException e) {
            LOG.log(Level.INFO, "Attemped to update roster on not yet authorized connection", e);
        }
        this.packetSender.processOutPacket(outPackets);
    }

    private boolean isUserOnline(BareJID userJID) {
        try {
            return this.sessions.isOnline(userJID);
        } catch (NotAuthorizedException e) {
            return false;
        }
    }

}
