/*
 * Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.user;

import com.kerio.im.bean.User;
import com.kerio.im.common.error.Code;
import com.kerio.im.tigase.Conversion;
import com.kerio.im.tigase.SessionProvider;
import com.kerio.im.tigase.derby.DerbyUtil;
import com.kerio.im.tigase.roster.PersonalRoster;
import com.kerio.im.xmpp.api.XmppException;
import com.kerio.im.xmpp.api.XmppUsers;
import java.io.StringReader;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.derby.jdbc.EmbeddedDataSource40;
import tigase.db.TigaseDBException;
import tigase.db.jdbc.JDBCRepository;
import tigase.xmpp.BareJID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.impl.roster.RosterAbstract;
import tigase.xmpp.impl.roster.RosterElement;
import tigase.xmpp.impl.roster.RosterFlat;

/**
 *
 * @author tskoda
 */
public class XmppUsersStore {

    private static final Logger LOG = Logger.getLogger(XmppUsersStore.class.getName());
    private static final int AVERAGE_USER_COUNT = 100;
    private static final String ALL_USERS_QUERY = ""
            + "select  u.USER_ID as barejid, "
            + "       pf.PVAL as fullname, "
            + "       u.K_STATUS as status, "
            + "       u.K_GAL as visible "
            + "  from APP.TIG_USERS as u "
            + "  left join APP.TIG_NODES as n "
            + "    on u.uid = n.UID and n.NODE = 'root' "
            + "  left join APP.TIG_PAIRS as pf "
            + "    on n.nid = pf.NID and pf.PKEY = 'fullname' "
            + " where LOCATE('@' || ? , u.USER_ID) > 0"
            + " order by u.acc_create_time";
    public static final String CREATE_USER
            = "{ call kAddUser(?, ?, ?, ?) }";
    public static final String USER_UPDATE_STATUS
            = " update APP.TIG_USERS t"
            + "    set t.k_status = ?"
            + " where t.uid = ?";
    public static final String USER_UPDATE_VISIBLE
            = " update APP.TIG_USERS t"
            + "    set t.k_gal = ?"
            + " where t.uid = ?";
    public static final String GET_ROSTER = ""
            + " select p.pval as roster "
            + "   from app.TIG_PAIRS as p "
            + "  where p.uid=? and p.nid=? and p.pkey='roster'";
    public static final String SET_ROSTER
            = JDBCRepository.UPDATE_PAIRS_QUERY;
    public static final String PERSONAL_ROSTER_GET_ALL = ""
            + " select r.buddy, r.name, r.groups, r.flag"
            + "  from app.PERSONAL_ROSTER r"
            + " where r.owner=?";
    public static final String PERSONAL_ROSTER_EXIST = ""
            + " select count(r.owner)"
            + "  from app.PERSONAL_ROSTER r"
            + " where r.owner=? and r.buddy=?";
    public static final String PERSONAL_ROSTER_INSERT = ""
            + " insert into app.PERSONAL_ROSTER (owner, buddy, name, groups, flag)"
            + " values (?, ?, ?, ?, ?)";
    public static final String PERSONAL_ROSTER_UPDATE = ""
            + " update app.PERSONAL_ROSTER r"
            + "    set r.name = ?, "
            + "        r.groups = ?, "
            + "        r.flag = ? "
            + "  where r.owner=? and r.buddy=?";
    public static final String PERSONAL_ROSTER_REMOVE_ALL = ""
            + " delete from app.PERSONAL_ROSTER r"
            + "  where r.owner=?";
    public static final String PERSONAL_ROSTER_REMOVE = ""
            + " delete from app.PERSONAL_ROSTER r"
            + "  where r.owner=? and r.buddy=?";

    public interface Mapper<K, T> {

        K key(BareJID userBareJID);

        T value(BareJID userBareJID, String fullName, boolean status, boolean visible);
    }

    public interface Filter {

        boolean pass(boolean status, boolean visible);
    }

    public static class Roster {

        private final BareJID ownerBareJID;
        private final long uid;
        private final long nid;
        private final Map<BareJID, RosterElement> roster;

        public Roster(BareJID ownerBareJID, long uid, long nid, Map<BareJID, RosterElement> roster) {
            this.ownerBareJID = ownerBareJID;
            this.uid = uid;
            this.nid = nid;
            this.roster = roster;
        }

        public Map<BareJID, RosterElement> getRoster() {
            return roster;
        }

        public BareJID getOwnerBareJID() {
            return ownerBareJID;
        }

        public long getNid() {
            return nid;
        }

        public long getUid() {
            return uid;
        }

        public String asString() {
            StringBuilder result = new StringBuilder();
            for (RosterElement element : roster.values()) {
                result.append(element.getRosterElement().toString());
            }
            return result.toString();
        }
    }

    private final String domain;
    private final EmbeddedDataSource40 dataSource;
    private final Connection connection;
    private final SessionProvider sessions;

    public XmppUsersStore(String domain, EmbeddedDataSource40 dataSource, SessionProvider sessions) throws SQLException {
        this.domain = domain;
        this.dataSource = dataSource;
        this.connection = dataSource.getConnection();
        this.connection.setAutoCommit(false);
        this.sessions = sessions;
    }

    public boolean add(User user) {
        boolean result = false;
        BareJID userJID = Conversion.toBareJID(user);
        try {
            if (getUserUid(userJID) == -1) {
                long uid = create(userJID, user.isEnabled(), user.isVisible());
                long nid = getRootNode(uid);
                setName(uid, nid, user.getFullName());
                setEmail(uid, nid, Conversion.toBareJID(user).toString());
                setRoster(uid, nid);
                commit();
                result = true;
            }
        } catch (XmppException e) {
            rollback(e);
        }
        return result;
    }

    public boolean remove(User user) {
        boolean result = false;
        BareJID userJID = Conversion.toBareJID(user);
        try {
            if (getUserUid(userJID) != -1) {
                removeImpl(userJID);
                commit();
                result = true;
            }
        } catch (XmppException e) {
            rollback(e);
        } finally {
            disconnect(userJID);
        }
        return result;
    }

    public boolean exist(String userName) {
        BareJID user = BareJID.bareJIDInstanceNS(userName, this.domain);
        return getUserUid(user) != -1;
    }

    public boolean update(User user, EnumSet<XmppUsers.Changes> changes) {
        boolean result = false;
        BareJID userJID = Conversion.toBareJID(user);

        try {
            if (getUserUid(userJID) != -1) {
                long uid = getUserUid(userJID);
                long nid = getRootNode(uid);

                for (XmppUsers.Changes change : changes) {
                    switch (change) {
                        case NAME:
                            setName(uid, nid, user.getFullName());
                            break;
                        case STATUS:
                            /*
                             * always keep roster synchronized
                             * disabled: disable, remove from all rosters
                             * enabled: enable, add to all rosters
                             */
                            setStatus(userJID, uid, user.isEnabled());
                            if (!user.isEnabled()) {
                                disconnect(userJID);
                            }
                            break;
                        case VISIBILITY:
                            /**
                             * always keep roster synchronized visible: set
                             * visible, add to all rosters, set subscription to
                             * both invisible: set invisible, remove from all
                             * rosters, set subscription to to all rosters
                             */
                            setVisible(userJID, uid, user.isVisible());
                            break;
                        default:
                            throw new XmppException("Received unsupported user change {0}", Code.INTERNAL_SERVER_ERROR, change);
                    }
                }
                commit();
                result = true;
            }
        } catch (XmppException e) {
            rollback(e);
        }

        return result;
    }

    public Roster readRoster(User user) {
        Map<BareJID, RosterElement> rosterContent = new HashMap<BareJID, RosterElement>();
        BareJID userBare = Conversion.toBareJID(user);
        long uid = getUserUid(userBare);
        long nid = getRootNode(uid);
        try {
            PreparedStatement getRosterStatement = this.connection.prepareStatement(GET_ROSTER);
            try {
                getRosterStatement.setLong(1, uid);
                getRosterStatement.setLong(2, nid);
                ResultSet rs = getRosterStatement.executeQuery();
                try {
                    if (rs.next()) {
                        String rosterFlat = rs.getString("roster");
                        RosterFlat.parseRosterUtil(rosterFlat, rosterContent, null);
                    }
                } finally {
                    DerbyUtil.close(rs);
                }
            } finally {
                DerbyUtil.close(getRosterStatement);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to read roster of {0}", e, Code.RESOURCE_ERROR, userBare);
        }

        return new Roster(userBare, uid, nid, rosterContent);
    }

    public void saveRoster(Roster roster) {
        try {
            PreparedStatement setRosterStatement = this.connection.prepareStatement(SET_ROSTER);
            try {
                setRosterStatement.setLong(1, roster.getNid());
                setRosterStatement.setLong(2, roster.getUid());
                setRosterStatement.setString(3, "roster");
                setRosterStatement.setClob(4, new StringReader(roster.asString()));
                setRosterStatement.executeUpdate();
                commit();
            } finally {
                DerbyUtil.close(setRosterStatement);
            }
        } catch (SQLException e) {
            rollback(new XmppException("Failed to save roster of {0}", e, Code.RESOURCE_ERROR, roster.getOwnerBareJID()));
        }
    }

    public void bye() {
        try {
            byeImpl();
        } catch (Exception e) {
            LOG.log(Level.INFO, "Failed to shutdown connection to {0} database", this.domain);
        }
    }

    public Map<String, User> all() {
        Map<String, User> result = new HashMap<String, User>(AVERAGE_USER_COUNT);
        all(result, new UserMapper(), new AllUsersFilter(), -1);
        return result;
    }

    public <K, T> boolean all(Map<K, T> result, Mapper<K, T> mapper, Filter filter, int maximum) {

        boolean hasMore = false;

        try {
            PreparedStatement allUsersStatement
                    = connection.prepareStatement(ALL_USERS_QUERY);

            try {

                // Load all user ids from database
                allUsersStatement.setString(1, this.domain);
                ResultSet rs = allUsersStatement.executeQuery();
                try {
                    while (rs.next()) {
                        
                        BareJID userBareJID = BareJID.bareJIDInstanceNS(rs.getString("barejid"));
                        String fullName = rs.getString("fullname");
                        boolean status = rs.getBoolean("status");
                        boolean visible = rs.getBoolean("visible");

                        if (filter.pass(status, visible)) {
                            result.put(mapper.key(userBareJID), mapper.value(userBareJID, fullName, status, visible));
                        }

                        if ((maximum > 0) && (result.size() >= maximum)) {
                            hasMore = true;
                            break;
                        }
                    }
                } finally {
                    DerbyUtil.close(rs);
                }
            } finally {
                DerbyUtil.close(allUsersStatement);
            }
        } catch (Exception e) {
            throw new XmppException("Loading list of users from tigase repository failed", e, Code.RESOURCE_ERROR);
        }

        return hasMore;
    }

    private void commit() {
        try {
            this.connection.commit();
        } catch (SQLException commitException) {
            LOG.log(Level.INFO, "Failed to commit user changes", commitException);
        }
    }

    private void rollback(XmppException e) {
        try {
            this.connection.rollback();
        } catch (SQLException rollbackException) {
            LOG.log(Level.INFO, "Failed to rollback user changes", rollbackException);
        } finally {
            throw e;
        }
    }

    private void removeImpl(BareJID user) {
        try {
            PreparedStatement removeUser = this.connection.prepareStatement(JDBCRepository.REMOVE_USER_QUERY);
            try {
                removeUser.setString(1, user.toString());
                removeUser.executeUpdate();

            } finally {
                DerbyUtil.close(removeUser);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to remove user", e, Code.RESOURCE_ERROR);
        }
    }

    public void disconnect(BareJID userJID) {
        try {
            this.sessions.disconnect(userJID);
        } catch (NotAuthorizedException e) {
            LOG.log(Level.FINE, "Can not disconnect user {0}", userJID);
        }
    }

    public void personalUpdate(BareJID owner, PersonalRoster.Buddy updatedBuddy) {
        try {

            if (personalExist(owner, updatedBuddy.getBuddy())) {
                personalUpdateImpl(owner, updatedBuddy);
            } else {
                personalAdd(owner, updatedBuddy);
            }

        } catch (SQLException e) {
            throw new XmppException("Failed to update {0} in {1}'s roster", e, Code.RESOURCE_ERROR, updatedBuddy.getBuddy(), owner);
        }
    }

    private void personalAdd(BareJID owner, PersonalRoster.Buddy updatedBuddy) throws SQLException {
        PreparedStatement insertPS = this.connection.prepareStatement(PERSONAL_ROSTER_INSERT);
        try {
            insertPS.setString(1, owner.toString());
            insertPS.setString(2, updatedBuddy.getBuddy().toString());

            if (updatedBuddy.getName() != null) {
                insertPS.setString(3, updatedBuddy.getName());
            } else {
                insertPS.setNull(3, Types.VARCHAR);
            }

            if ((updatedBuddy.getGroups() != null) && (updatedBuddy.getGroups().length > 0)) {
                insertPS.setString(4, PersonalRoster.Buddy.groups(updatedBuddy.getGroups()));
            } else {
                insertPS.setNull(4, Types.VARCHAR);
            }

            insertPS.setString(5, updatedBuddy.getFlag().toString());

            int created = insertPS.executeUpdate();
            if (created != 1) {
                throw new XmppException("Created {0} records instead of 1", Code.RESOURCE_ERROR, created);
            }
            
            commit();
        } catch (XmppException e) {
            rollback(e);
        } catch (SQLException e) {
            rollback(new XmppException("Failed to add {1} to {0}'s personal roster", e, Code.RESOURCE_ERROR, owner, updatedBuddy.getBuddy()));
        } finally {
            DerbyUtil.close(insertPS);
        }
    }

    private void personalUpdateImpl(BareJID owner, PersonalRoster.Buddy updatedBuddy) throws SQLException {
        PreparedStatement updatePS = this.connection.prepareStatement(PERSONAL_ROSTER_UPDATE);
        try {
            if (updatedBuddy.getName() != null) {
                updatePS.setString(1, updatedBuddy.getName());
            } else {
                updatePS.setNull(1, Types.VARCHAR);
            }

            if ((updatedBuddy.getGroups() != null) && (updatedBuddy.getGroups().length > 0)) {
                updatePS.setString(2, PersonalRoster.Buddy.groups(updatedBuddy.getGroups()));
            } else {
                updatePS.setNull(2, Types.VARCHAR);
            }

            updatePS.setString(3, updatedBuddy.getFlag().toString());

            updatePS.setString(4, owner.toString());
            updatePS.setString(5, updatedBuddy.getBuddy().toString());
            int updated = updatePS.executeUpdate();
            if (updated != 1) {
                throw new XmppException("Updated {0} records instead of 1", Code.RESOURCE_ERROR, updated);
            }
            
            commit();
        } catch (XmppException e) {
            rollback(e);
        } catch (SQLException e) {
            rollback(new XmppException("Failed to update contact {0} in {1}'s personal roster", e, Code.RESOURCE_ERROR, updatedBuddy.getBuddy(), owner));
        } finally {
            DerbyUtil.close(updatePS);
        }
    }

    private boolean personalExist(BareJID owner, BareJID buddy) throws SQLException {
        boolean result;
        PreparedStatement existPS = this.connection.prepareStatement(PERSONAL_ROSTER_EXIST);
        try {
            existPS.setString(1, owner.toString());
            existPS.setString(2, buddy.toString());
            ResultSet existRS = existPS.executeQuery();
            try {
                if (existRS.next()) {
                    if (existRS.getInt(1) == 0) {
                        result = false;
                    } else {
                        result = true;
                    }
                } else {
                    throw new XmppException("Unable to check {0} existence", Code.RESOURCE_ERROR, buddy);
                }
            } finally {
                DerbyUtil.close(existRS);
            }
        } finally {
            DerbyUtil.close(existPS);
        }
        return result;
    }

    public void personalRemove(BareJID owner) {
        try {
            PreparedStatement removePS = this.connection.prepareStatement(PERSONAL_ROSTER_REMOVE_ALL);
            try {
                removePS.setString(1, owner.toString());
                removePS.executeUpdate();
                commit();
            } finally {
                DerbyUtil.close(removePS);
            }
        } catch (SQLException e) {
            rollback(new XmppException("Failed to remove all from {0}'s roster", e, Code.RESOURCE_ERROR, owner));
        }
    }

    public void personalRemove(BareJID owner, BareJID buddyJID) {
        try {
            PreparedStatement removePS = this.connection.prepareStatement(PERSONAL_ROSTER_REMOVE);
            try {
                removePS.setString(1, owner.toString());
                removePS.setString(2, buddyJID.toString());
                removePS.executeUpdate();
                commit();
            } finally {
                DerbyUtil.close(removePS);
            }
        } catch (SQLException e) {
            rollback(new XmppException("Failed to remove {0} from {1}'s roster", e, Code.RESOURCE_ERROR, buddyJID, owner));
        }
    }

    public Map<BareJID, PersonalRoster.Buddy> personalAll(BareJID owner) {
        Map<BareJID, PersonalRoster.Buddy> result = new HashMap<BareJID, PersonalRoster.Buddy>();

        try {
            PreparedStatement allPS = this.connection.prepareStatement(PERSONAL_ROSTER_GET_ALL);
            try {
                allPS.setString(1, owner.toString());
                ResultSet allRS = allPS.executeQuery();
                try {
                    while (allRS.next()) {
                        BareJID buddy = BareJID.bareJIDInstanceNS(allRS.getString("buddy"));
                        String name = allRS.getString("name");
                        String groups = allRS.getString("groups");
                        PersonalRoster.Buddy.FLAG flag = PersonalRoster.Buddy.FLAG.valueOf(allRS.getString("flag"));

                        result.put(buddy, new PersonalRoster.Buddy(buddy, name, PersonalRoster.Buddy.groups(groups), flag));
                    }
                } finally {
                    DerbyUtil.close(allRS);
                }
            } finally {
                DerbyUtil.close(allPS);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to read personal roster for user", e, Code.RESOURCE_ERROR, owner);
        }

        return result;
    }

    private void byeImpl() throws SQLException, TigaseDBException {

        this.dataSource.setShutdownDatabase("shutdown");
        try {
            this.dataSource.getConnection();
            throw new TigaseDBException("Can not shutdown database");
        } catch (SQLException ex) {
            String engineShutdownSucc = "XJ015";
            String databaseShutdownSucc = "08006";
            if (engineShutdownSucc.equals(ex.getSQLState()) || (databaseShutdownSucc.equals(ex.getSQLState()))) {
                LOG.info("Database successfuly closed");
            } else {
                LOG.log(Level.SEVERE, "Can not close database", ex);
            }
        } catch (Exception ex) {
            LOG.log(Level.SEVERE, "Can not close database", ex);
        }
    }

    private long create(BareJID user, boolean enabled, boolean visible) {
        long result = -1;
        try {
            PreparedStatement addUser = this.connection.prepareStatement(CREATE_USER);
            try {
                addUser.setString(1, user.toString());
                addUser.setNull(2, Types.VARCHAR);
                addUser.setBoolean(3, enabled);
                addUser.setBoolean(4, visible);
                ResultSet rs = addUser.executeQuery();
                try {
                    if (rs.next()) {
                        result = rs.getLong(1);
                    } else {
                        throw new XmppException("Failed to add user {0}", Code.RESOURCE_ERROR, user.toString());
                    }
                } finally {
                    DerbyUtil.close(rs);
                }
            } finally {
                DerbyUtil.close(addUser);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to add user", e, Code.RESOURCE_ERROR);
        }

        return result;
    }

    private long getUserUid(BareJID user) {
        long result = -1;
        try {
            PreparedStatement userExists = this.connection.prepareStatement(JDBCRepository.GET_USER_DB_UID_QUERY);
            try {
                userExists.setString(1, user.toString());
                ResultSet rs = userExists.executeQuery();
                try {
                    if (rs.next()) {
                        result = rs.getLong(1);
                    }
                } finally {
                    DerbyUtil.close(rs);
                }
            } finally {
                DerbyUtil.close(userExists);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to check if user exists", e, Code.RESOURCE_ERROR);
        }

        return result;
    }

    private long getRootNode(long uid) {
        long result = -1;
        try {
            Statement rootNode = this.connection.createStatement();
            try {
                String query = JDBCRepository.buildNodeQuery(uid, null);

                ResultSet rs = rootNode.executeQuery(query);
                try {
                    if (rs.next()) {
                        result = rs.getLong(1);
                    } else {
                        throw new XmppException("Failed to get root node", Code.RESOURCE_ERROR);
                    }
                } finally {
                    DerbyUtil.close(rs);
                }
            } finally {
                DerbyUtil.close(rootNode);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to add user", e, Code.RESOURCE_ERROR);
        }

        return result;
    }

    private void setStatus(BareJID userJID, long uid, boolean status) {
        try {
            PreparedStatement updateStatus = this.connection.prepareStatement(USER_UPDATE_STATUS);
            try {
                updateStatus.setBoolean(1, status);
                updateStatus.setLong(2, uid);
                updateStatus.executeUpdate();
            } finally {
                DerbyUtil.close(updateStatus);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to update user {0} status", e, Code.RESOURCE_ERROR, userJID);
        }
    }

    private void setVisible(BareJID userJID, long uid, boolean visible) {
        try {
            PreparedStatement updateVisibility = this.connection.prepareStatement(USER_UPDATE_VISIBLE);
            try {
                updateVisibility.setBoolean(1, visible);
                updateVisibility.setLong(2, uid);
                updateVisibility.executeUpdate();
            } finally {
                DerbyUtil.close(updateVisibility);
            }
        } catch (SQLException e) {
            throw new XmppException("Failed to update user {0} visibility", e, Code.RESOURCE_ERROR, userJID);
        }
    }

    private void setName(long uid, long nid, String fullName) {
        try {
            this.setData(uid, nid, "fullname", fullName);
        } catch (SQLException e) {
            throw new XmppException("Failed to set user {0} name", e, Code.RESOURCE_ERROR);
        }
    }

    private void setEmail(long uid, long nid, String email) {
        try {
            this.setData(uid, nid, "email", email);
        } catch (SQLException e) {
            throw new XmppException("Failed to set user {0} email", e, Code.RESOURCE_ERROR);
        }
    }

    private void setRoster(long uid, long nid) {

        try {
            this.setData(uid, nid, RosterAbstract.ROSTER, "");
        } catch (SQLException e) {
            throw new XmppException("Failed to set user {0} roster", e, Code.RESOURCE_ERROR);
        }
    }

    private void setData(long uid, long nid, String key, String value) throws SQLException {
        PreparedStatement setData = this.connection.prepareStatement(JDBCRepository.UPDATE_PAIRS_QUERY);
        try {
            setData.setLong(1, nid);
            setData.setLong(2, uid);
            setData.setString(3, key);
            setData.setClob(4, new StringReader(value));
            setData.executeUpdate();

        } finally {
            DerbyUtil.close(setData);
        }
    }
}
