/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase;

import static org.hamcrest.CoreMatchers.is;
import static org.junit.Assert.*;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import tigase.xmpp.JID;
import tigase.xmpp.StanzaType;
import tigase.xmpp.XMPPResourceConnection;
import tigase.xmpp.XMPPSession;
import static com.kerio.im.tigase.ResourceSelectorTestUtil.*;
import java.util.Set;

/**
 *
 * @author tskoda
 */
public class MessageResourceSelectorNoResourceTest {

    @Rule
    public ExpectedException exception = ExpectedException.none();
    private MessageResourceSelector selector = new MessageResourceSelector();
    private JID jid = JID.jidInstanceNS("user@test");

    @Test
    public void testEmpty_allNegative() throws Exception {
        StanzaType type = null; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(-100, MessageResourceSelector.Show.availible),
                with(-1000, MessageResourceSelector.Show.chat),
                with(-1, MessageResourceSelector.Show.dnd));

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(0)); //offline delivery
    }

    @Test
    public void testNormal_allNegative() throws Exception {
        StanzaType type = StanzaType.normal; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(-100, MessageResourceSelector.Show.availible),
                with(-1000, MessageResourceSelector.Show.chat),
                with(-1, MessageResourceSelector.Show.dnd));

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(0)); //offline delivery
    }

    @Test
    public void testChat_allNegative() throws Exception {
        StanzaType type = StanzaType.chat; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(-100, MessageResourceSelector.Show.availible),
                with(-1000, MessageResourceSelector.Show.chat),
                with(-1, MessageResourceSelector.Show.dnd));

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(0)); //offline delivery
    }

    @Test
    public void testGroupchat_allNegative() throws Exception {
        StanzaType type = StanzaType.groupchat; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(-100, MessageResourceSelector.Show.availible),
                with(-1000, MessageResourceSelector.Show.chat),
                with(-1, MessageResourceSelector.Show.dnd));

        exception.expect(serviceUnavailible());
        selector.preferredResources(jid, type, mockSession);
    }

    @Test
    public void testHeadline_allNegative() throws Exception {
        StanzaType type = StanzaType.headline; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(-100, MessageResourceSelector.Show.availible),
                with(-1000, MessageResourceSelector.Show.chat),
                with(-1, MessageResourceSelector.Show.dnd));

        exception.expect(ignore());
        selector.preferredResources(jid, type, mockSession);
    }

    @Test
    public void testEmpty_allPositive() throws Exception {
        StanzaType type = null; // null == normal message
        XMPPResourceConnection active = with(10, MessageResourceSelector.Show.availible);
        XMPPSession mockSession = mockSession(
                jid,
                0,
                active,
                with(100, MessageResourceSelector.Show.availible),
                with(1000, MessageResourceSelector.Show.chat),
                with(1, MessageResourceSelector.Show.dnd));

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(1));
    }

    @Test
    public void testNormal_allPositive() throws Exception {
        StanzaType type = StanzaType.normal; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(100, MessageResourceSelector.Show.availible),
                with(1000, MessageResourceSelector.Show.chat),
                with(1, MessageResourceSelector.Show.dnd));

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(1));
    }

    @Test
    public void testChat_allPositive() throws Exception {
        StanzaType type = StanzaType.chat; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(100, MessageResourceSelector.Show.availible),
                with(1000, MessageResourceSelector.Show.chat),
                with(1, MessageResourceSelector.Show.dnd));

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(1));
    }

    @Test
    public void testGroupchat_allPositive() throws Exception {
        StanzaType type = StanzaType.groupchat; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(100, MessageResourceSelector.Show.availible),
                with(1000, MessageResourceSelector.Show.chat),
                with(1, MessageResourceSelector.Show.dnd));

        exception.expect(serviceUnavailible());
        selector.preferredResources(jid, type, mockSession);
    }

    @Test
    public void testHeadline_allPositive() throws Exception {
        StanzaType type = StanzaType.headline; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                with(100, MessageResourceSelector.Show.availible),
                with(1000, MessageResourceSelector.Show.chat),
                with(1, MessageResourceSelector.Show.dnd));

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(3)); //deliver to all
    }

    @Test
    public void testHeadline_mixed() throws Exception {
        StanzaType type = StanzaType.headline; // null == normal message
        XMPPResourceConnection positive = with(10, MessageResourceSelector.Show.availible);
        XMPPResourceConnection anotherPositive = with(1, MessageResourceSelector.Show.dnd);
        XMPPSession mockSession = mockSession(
                jid,
                -1,
                positive,
                with(-100, MessageResourceSelector.Show.availible),
                with(-1000, MessageResourceSelector.Show.chat),
                anotherPositive);

        Set<XMPPResourceConnection> result = selector.preferredResources(jid, type, mockSession);
        assertThat(result.size(), is(2)); //deliver to possitive ones
        assertTrue(selector.preferredResources(jid, type, mockSession).contains(positive));
        assertTrue(selector.preferredResources(jid, type, mockSession).contains(anotherPositive));
    }

    @Test
    public void testEmpty_empty() throws Exception {
        StanzaType type = null; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1);

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(0)); //offline delivery
    }

    @Test
    public void testNormal_empty() throws Exception {
        StanzaType type = StanzaType.normal; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1);

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(0)); //offline delivery
    }

    @Test
    public void testChat_empty() throws Exception {
        StanzaType type = StanzaType.chat; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1);

        assertThat(selector.preferredResources(jid, type, mockSession).size(), is(0)); //offline delivery
    }

    @Test
    public void testGroupchat_empty() throws Exception {
        StanzaType type = StanzaType.groupchat; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1);

        exception.expect(serviceUnavailible());
        selector.preferredResources(jid, type, mockSession);
    }

    @Test
    public void testHeadline_empty() throws Exception {
        StanzaType type = StanzaType.headline; // null == normal message
        XMPPSession mockSession = mockSession(
                jid,
                -1);

        exception.expect(ignore());
        selector.preferredResources(jid, type, mockSession);
    }

}
