/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase;

import com.kerio.im.tigase.MessageResourceSelector.Show;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import org.junit.Test;
import static org.hamcrest.CoreMatchers.*;
import static org.mockito.Mockito.*;
import static org.junit.Assert.*;
import tigase.xml.Element;
import tigase.xml.XMLNodeIfc;
import tigase.xmpp.XMPPResourceConnection;
import static com.kerio.im.tigase.ResourceSelectorTestUtil.*;

/**
 *
 * @author tskoda
 */
public class MessageResourceSelectorTest {

    public MessageResourceSelectorTest() {
    }

    @Test
    public void testShowValueOf() {
        Element presence = new Element("presence");

        List<XMLNodeIfc> children = new LinkedList<XMLNodeIfc>();
        children.add(new Element("show", "chat"));
        presence.setChildren(children);

        assertThat(Show.valueOf(presence), is(Show.chat));

        children.clear();
        children.add(new Element("show", "CHAT"));
        presence.setChildren(children);

        assertThat(Show.valueOf(presence), is(Show.chat));

        children.clear();
        children.add(new Element("show", "dnd"));
        presence.setChildren(children);

        assertThat(Show.valueOf(presence), is(Show.dnd));

        children.clear();
        children.add(new Element("show", "another"));
        presence.setChildren(children);

        assertThat(Show.valueOf(presence), is(Show.unknown));

        children.clear();
        presence.setChildren(children);

        assertThat(Show.valueOf(presence), is(Show.availible));
    }

    @Test
    public void testPreferredResourcesNegative() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        all.add(with(-1, Show.unknown));
        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(0));

        all.add(with(-1, Show.chat));
        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(0));

        all.add(with(-100, Show.availible));
        all.add(with(-100, Show.away));
        all.add(with(-100, Show.dnd));
        all.add(with(-100, Show.xa));
        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(0));
    }

    @Test
    public void testPreferredResourceEmpty() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(0));
    }

    @Test
    public void testPreferredResourceSingle() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        XMPPResourceConnection away = with(0, Show.away);
        all.add(away);

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(1));
        assertTrue(result.contains(away));
    }

    @Test
    public void testPreferredResourceSingleNotAuthenticated() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        XMPPResourceConnection away = with(0, Show.away);
        when(away.isAuthorized()).thenReturn(Boolean.FALSE);
        all.add(away);

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(0));
    }

    @Test
    public void testPreferredResourceSingleLowestPriority() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        XMPPResourceConnection chat = with(0, Show.chat); //highest status has precedence
        all.add(chat);

        XMPPResourceConnection availible = with(1, Show.availible);
        all.add(availible);

        XMPPResourceConnection away = with(1, Show.away);
        all.add(away);

        XMPPResourceConnection dnd = with(5, Show.dnd);
        all.add(dnd);

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(1));
        assertTrue(result.contains(chat));

    }

    @Test
    public void testPreferredResourceSingleLowestPriorityBackwards() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        XMPPResourceConnection availible = with(1, Show.availible);
        all.add(availible);

        XMPPResourceConnection away = with(1, Show.away);
        all.add(away);

        XMPPResourceConnection dnd = with(5, Show.dnd);
        all.add(dnd);

        XMPPResourceConnection chat = with(0, Show.chat); //highest status has presedence
        all.add(chat);

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(1));
        assertTrue(result.contains(chat));

    }

    @Test
    public void testPreferredResourceSingleWithHighestPriority() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        XMPPResourceConnection chat = with(-1, Show.chat);
        all.add(chat);

        XMPPResourceConnection chat1 = with(0, Show.chat);
        all.add(chat1);

        XMPPResourceConnection chat2 = with(2, Show.chat);
        all.add(chat2);

        XMPPResourceConnection chat3 = with(3, Show.chat);
        all.add(chat3);

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(1));
        assertTrue(result.contains(chat3));

    }

    @Test
    public void testPreferredResourceSingleWithHighestPriorityBackwards() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        XMPPResourceConnection chat = with(3, Show.chat);
        all.add(chat);

        XMPPResourceConnection chat1 = with(2, Show.chat);
        all.add(chat1);

        XMPPResourceConnection chat2 = with(-1, Show.chat);
        all.add(chat2);

        XMPPResourceConnection chat3 = with(0, Show.chat);
        all.add(chat3);

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(1));
        assertTrue(result.contains(chat));

    }

    @Test
    public void testPreferredResourceMany() throws Exception {
        MessageResourceSelector selector = new MessageResourceSelector();
        List<XMPPResourceConnection> all = new ArrayList<XMPPResourceConnection>();
        Set<XMPPResourceConnection> result = new HashSet<XMPPResourceConnection>();

        XMPPResourceConnection chat = with(3, Show.chat);
        all.add(chat);

        XMPPResourceConnection chat2 = with(3, Show.chat);
        all.add(chat2);

        XMPPResourceConnection chat3 = with(4, Show.chat);
        all.add(chat3);

        XMPPResourceConnection chat4 = with(4, Show.chat);
        all.add(chat4);

        XMPPResourceConnection availible = with(100, Show.availible);
        all.add(availible);

        selector.highestPrefferedResources(result, all);
        assertThat(result.size(), is(2));
        assertTrue(result.contains(chat3));
        assertTrue(result.contains(chat4));
    }
}
