/*
 * Copyright 2014 Kerio Technologies s.r.o. All rights reserved
 */
package com.kerio.im.tigase.plugin;

import com.kerio.im.tigase.bridge.IMBridgePlugin;
import com.kerio.im.tigase.bridge.xmpp.IMBridge;
import java.util.LinkedList;
import java.util.Queue;
import org.junit.Assert;
import org.junit.Test;
import tigase.server.Packet;
import tigase.xml.Element;
import static tigase.xml.Element.PROCESSED_BY_IM_BRIDGE_SYNC;
import tigase.xmpp.JID;
import tigase.xmpp.StanzaType;
import tigase.xmpp.XMPPResourceConnection;

/**
 *
 * @author jzacek
 */
public class IMPluginTest {

    private final IMBridgePlugin bridge = new IMBridgePlugin();
    private final XMPPResourceConnection con = new FakeResourceConnection("d.com");

    private static Packet createPacket(JID pFrom, JID pTo, JID sFrom, JID sTo, StanzaType type, String element, String body) {
        return createPacket(pFrom, pTo, sFrom, sTo, type, element, body, false);
    }

    private static Packet createPacket(JID pFrom, JID pTo, JID sFrom, JID sTo, StanzaType type, String element, String body, boolean alreadySychronized) {
        Element e = new Element(element);

        e.setAttribute("type", type.toString());

        if (body != null) {
            Element b = new Element("body");
            b.setCData(body);
            e.addChild(b);
        }

        Packet res = Packet.packetInstance(e, sFrom, sTo);
        res.setPacketFrom(pFrom);
        res.setPacketTo(pTo);
        if (alreadySychronized) {
            res.setAttribute(PROCESSED_BY_IM_BRIDGE_SYNC, "1");
        }
        return res;
    }

    private static JID jid(String j) {
        return JID.jidInstanceNS(j);
    }

    @Test
    public void testIgnoredPackets() {
        Queue<Packet> input = new LinkedList<>();

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                null));

        input.add(createPacket(jid("sess-man@d.com"),
                jid(IMBridge.IM_BRIDGE_NAME + "@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa1"));

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.available,
                "presence",
                "aaa2",
                true));

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@db.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa3"));

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@db.com"),
                StanzaType.chat,
                "message",
                "aaa4"));

        int expected = input.size();

        bridge.filter(null, con, null, input);

        Assert.assertEquals("IMPlugin filter synchronized unwanted packet", expected, input.size());

        input.stream().forEach((p) -> {
            if (!p.getPacketTo().getLocalpart().equals(IMBridge.IM_BRIDGE_NAME) && (p.getType() != StanzaType.available)) {
                Assert.assertNull("Packet should not have been processed by filter.", p.getAttribute(Element.PROCESSED_BY_IM_BRIDGE_SYNC));
                Assert.assertNull("Packet should not have RECIEVED_BY_MULTIPLE_CLIENTS.", p.getAttribute(Element.RECIEVED_BY_MULTIPLE_CLIENTS));
            } else {
                Assert.assertEquals("Packet should have been processed by filter.", "1", p.getAttribute(Element.PROCESSED_BY_IM_BRIDGE_SYNC));
                Assert.assertNull("Packet should not have RECIEVED_BY_MULTIPLE_CLIENTS.", p.getAttribute(Element.RECIEVED_BY_MULTIPLE_CLIENTS));
            }
        });
    }

    @Test
    public void testSynchronizationSynchronize() {
        Queue<Packet> input = new LinkedList<>();

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa"));

        input.add(createPacket(
                jid("s2s@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa"));

        int expected = input.size() + 1;

        bridge.filter(null, con, null, input);

        Assert.assertEquals("IMPlugin filter synchronized unwanted packet or did not synchronize expected packets", expected, input.size());
    }

    @Test
    public void testSynchronizationDoNotSynchronize() {
        Queue<Packet> input = new LinkedList<>();

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa"));

        input.add(createPacket(
                jid("s2s@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa"));

        input.add(createPacket(jid("s2s@d.com"),
                jid(IMBridge.IM_BRIDGE_NAME + "@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa"));

        int expected = input.size();

        bridge.filter(null, con, null, input);

        Assert.assertEquals("IMPlugin filter synchronized unwanted packet", expected, input.size());
    }

    @Test
    public void testSynchronizationSynchronizeMultiGroup() {
        Queue<Packet> input = new LinkedList<>();

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa"));

        input.add(createPacket(
                jid("s2s@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "bbb"));

        input.add(createPacket(
                jid("s2s@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "aaa"));

        input.add(createPacket(
                jid("sess-man@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "bbb"));

        input.add(createPacket(
                jid("s3s@d.com"),
                jid("c2s@d.com"),
                jid("test@d.com"),
                jid("test1@d.com"),
                StanzaType.chat,
                "message",
                "bbb"));

        int expectedAAA = 2 + 1;
        int expectedBBB = 3 + 1;

        bridge.filter(null, con, null, input);

        long actualAAA = input.stream().filter((p) -> p.getElement().getChild("body").getCData().equals("aaa")).count();
        long actualBBB = input.stream().filter((p) -> p.getElement().getChild("body").getCData().equals("bbb")).count();

        Assert.assertEquals("IMPlugin filter synchronized unwanted packet or did not synchronize expected packets", expectedAAA, actualAAA);
        Assert.assertEquals("IMPlugin filter synchronized unwanted packet or did not synchronize expected packets", expectedBBB, actualBBB);
    }

}
