/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.tigase.roster.RosterContent.Buddy;
import com.kerio.im.tigase.user.RosterElementUtil;
import com.kerio.im.tigase.user.UserMapper;
import com.kerio.im.tigase.user.UserUtil;
import com.kerio.im.tigase.user.XmppUsersStore;
import com.kerio.im.xmpp.api.XmppUsers;
import java.util.Arrays;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import org.hamcrest.CoreMatchers;
import static org.hamcrest.CoreMatchers.is;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import static org.junit.Assert.*;
import static org.mockito.Mockito.*;
import tigase.xmpp.BareJID;
import tigase.xmpp.JID;
import tigase.xmpp.impl.roster.RosterAbstract;
import tigase.xmpp.impl.roster.RosterElement;

/**
 *
 * @author tskoda
 */
public class DefaultRosterListenerTest {

    private static final UserUtil as = new UserUtil();
    private static final RosterElementUtil re = new RosterElementUtil();
    private static final BuddyUtil bd = new BuddyUtil();

    private ElementMapper mapper;
    private GroupService groups;
    private RosterNotify rosterNotify;

    private DefaultRoster defaultRoster;
    private Map<BareJID, Buddy> defaultRosterContent;

    private PersonalRoster personalRoster;
    private PersonalRoster.PersonalUserRoster personalRosterRoster;

    public DefaultRosterListenerTest() {
    }

    @Before
    public void setUp() {

        this.groups = new TestGroupService();

        this.rosterNotify = mock(RosterNotify.class);
        this.mapper = new ElementMapper(this.groups, new UserMapper());

        this.defaultRoster = mock(DefaultRoster.class);
        this.defaultRosterContent = new LinkedHashMap<BareJID, Buddy>(); //we need to preserve insertion order
        when(this.defaultRoster.roster(any(String.class))).thenReturn(new RosterContent(this.defaultRosterContent, this.mapper, 10, false));

        this.personalRoster = mock(PersonalRoster.class);
        this.personalRosterRoster = mock(PersonalRoster.PersonalUserRoster.class);
        when(this.personalRoster.of(any(BareJID.class))).thenReturn(this.personalRosterRoster);

    }

    @After
    public void tearDown() {
    }

    private void toDefault(User... users) {
        for (User user : users) {
            this.defaultRosterContent.put(as.bare(user), mapper.value(user));
        }
    }

    private void toStatic(Map<BareJID, RosterElement> staticRosterContent, User... users) {
        for (User user : users) {
            staticRosterContent.put(as.bare(user), new RosterElement(as.jid(user), user.getFullName(), new String[]{groups.group(user.getPrimaryEmail())}, null));
        }
    }

    private void toStaticUpdate(Map<BareJID, RosterElement> staticRosterContent, BareJID user, String name, String[] groups) {
        RosterElement el = staticRosterContent.get(user);
        if (el != null) {
            el.setGroups(groups);
            el.setName(name);
        }
    }

    @Test
    public void testAddDisabledInvisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User disabledInvisible = as.user(1, false, false);

        listener.add(disabledInvisible);

        assertThat(this.defaultRosterContent.isEmpty(), is(true));

        verifyZeroInteractions(this.rosterNotify);
    }

    @Test
    public void testAddDisabledVisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User disabledVisible = as.user(1, false, true);

        listener.add(disabledVisible);

        assertThat(this.defaultRosterContent.isEmpty(), is(true));

        verifyZeroInteractions(this.rosterNotify);
    }

    @Test
    public void testAddEnabledInvisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User enabledInvisible = as.user(1, true, false);

        listener.add(enabledInvisible);

        assertThat(this.defaultRosterContent.isEmpty(), is(true));

        verifyZeroInteractions(this.rosterNotify);
    }

    @Test
    public void testAddKnown() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User added = as.user(1); //visible and enabled

        //make the user already present in default roster
        this.defaultRosterContent.put(as.bare(added), this.mapper.value(added));
        //make the user NOT managed by private roster
        when(this.personalRosterRoster.managed(as.bare(added))).thenReturn(Boolean.FALSE);

        listener.add(added);

        assertThat(this.defaultRosterContent.size(), is(1));
        //no notification is produced as this user is already in default roster
        verifyZeroInteractions(this.rosterNotify);

    }

    @Test
    public void testAddEmptyDefault() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User added = as.user(1); //visible and enabled

        //by default zero users in default roster
        //make the user NOT managed by private roster
        when(this.personalRosterRoster.managed(as.bare(added))).thenReturn(Boolean.FALSE);

        listener.add(added);

        //verify user is added to default roster
        assertThat(this.defaultRosterContent.size(), is(1));
        assertThat(this.defaultRosterContent.get(as.bare(added)), is(this.mapper.value(added)));

        //to capture recipients
        ArgumentCaptor<List> recipientCaptor = ArgumentCaptor.forClass(List.class);

        //verify packet notifier is called just once
        verify(this.rosterNotify, times(1)).aboutAdd(eq(added), recipientCaptor.capture(), any(RosterNotify.Filter.class), any(RosterNotify.HashProvider.class));

        //recipients contains only freshly added user
        assertThat(recipientCaptor.getValue().size(), is(1));
        assertThat((BareJID) recipientCaptor.getValue().get(0), is(as.bare(added)));
    }

    @Test
    public void testAddTwoRecipientsNoPersonal() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User added = as.user(1); //visible and enabled

        //by default zero users in default roster
        this.defaultRosterContent.put(as.bare(2), this.mapper.value(as.user(2)));
        this.defaultRosterContent.put(as.bare(3), this.mapper.value(as.user(3)));

        //make the user NOT managed by private roster
        when(this.personalRosterRoster.managed(as.bare(added))).thenReturn(Boolean.FALSE);

        listener.add(added);

        //verify user is added to default roster
        assertThat(this.defaultRosterContent.size(), is(3));
        assertThat(this.defaultRosterContent.get(as.bare(added)), is(this.mapper.value(added)));

        //to capture recipients 
        ArgumentCaptor<List> recipientCaptor = ArgumentCaptor.forClass(List.class);

        //verify packet notifier is called just once
        verify(this.rosterNotify, times(1)).aboutAdd(eq(added), recipientCaptor.capture(), any(RosterNotify.Filter.class), any(RosterNotify.HashProvider.class));

        //recipients contains three users
        assertThat(recipientCaptor.getValue().size(), is(3));

    }

    @Test
    public void testAddedFilter() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        DefaultRosterListener.AddedFilter filter = listener.new AddedFilter();

        BareJID added = as.bare(1);
        PacketBuilder.AddUserPacket packet = new PacketBuilder.AddUserPacket(added, RosterAbstract.SubscriptionType.both, "Custom Name", new String[]{"Custom Group"});

        //block added user itself
        assertNull(filter.pass(added, packet));

        //pass packet 'as is' if no personal info found
        BareJID customRecipient = as.bare(2);
        when(personalRosterRoster.managed(added)).thenReturn(Boolean.FALSE);
        assertThat(filter.pass(customRecipient, packet), is(packet));

        //block packet if personal info found
        when(personalRosterRoster.managed(added)).thenReturn(Boolean.TRUE);
        assertNull(filter.pass(added, packet));
    }

    @Test
    public void testRemovenonExistent() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User disabledInvisible = as.user(1, false, false);
        User disabledVisible = as.user(1, false, true);
        User enabledInvisible = as.user(1, true, false);
        User enabledVisible = as.user(1, true, true);

        listener.remove(disabledInvisible);
        listener.remove(disabledVisible);
        listener.remove(enabledInvisible);
        listener.remove(enabledVisible);

        verifyZeroInteractions(this.rosterNotify);
    }

    @Test
    public void testRemoveExists() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User removed = as.user(0);

        this.defaultRosterContent.put(as.bare(removed), this.mapper.value(removed));
        this.defaultRosterContent.put(as.bare(as.user(1)), this.mapper.value(as.user(1)));
        this.defaultRosterContent.put(as.bare(as.user(2)), this.mapper.value(as.user(2)));

        listener.remove(removed);

        //to capture recipients 
        ArgumentCaptor<List> recipientCaptor = ArgumentCaptor.forClass(List.class);

        //verify packet notifier is called just once
        verify(this.rosterNotify, times(1)).aboutRemove(eq(removed), recipientCaptor.capture(), any(RosterNotify.Filter.class), any(RosterNotify.HashProvider.class));

        //recipients contains three users
        assertThat(recipientCaptor.getValue().size(), is(2));

    }

    @Test
    public void testRemoveFilter() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        DefaultRosterListener.RemovedFilter filter = listener.new RemovedFilter();

        BareJID removed = as.bare(1);
        PacketBuilder.RemoveUserPacket packet = new PacketBuilder.RemoveUserPacket(removed);

        //block added user itself
        assertNull(filter.pass(removed, packet));

        //pass packet 'as is' if no personal info found
        BareJID customRecipient = as.bare(2);
        when(personalRosterRoster.managed(removed)).thenReturn(Boolean.FALSE);
        assertThat(filter.pass(customRecipient, packet), is(packet));

        //block packet if personal info found
        when(personalRosterRoster.managed(removed)).thenReturn(Boolean.TRUE);
        assertNull(filter.pass(removed, packet));
    }

    @Test
    public void testUpdateName() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User user = as.user(1); //enabled and visible, other cases are in testUpdateInvisibleOrDisabled
        User userChangedName = as.renamed(user, "this is new name");

        assertThat(this.defaultRosterContent.containsKey(as.bare(user)), is(false));

        listener.update(user, EnumSet.of(XmppUsers.Changes.NAME));

        assertThat(this.defaultRosterContent.get(as.bare(user)), is(this.mapper.value(userChangedName)));
        verify(this.rosterNotify, times(1)).aboutAdd(eq(userChangedName), any(List.class), any(DefaultRosterListener.UpdateFilter.class), any(RosterNotify.HashProvider.class));

    }

    @Test
    public void testUpdateEnabledAndVisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User user = as.user(1); //enabled and visible, other cases are in testUpdateInvisibleOrDisabled
        this.defaultRosterContent.put(as.bare(user), this.mapper.value(user));

        listener.update(user, EnumSet.of(XmppUsers.Changes.NAME));

        assertThat(this.defaultRosterContent.get(as.bare(user)), is(this.mapper.value(user)));
        verify(this.rosterNotify, times(1)).aboutAdd(eq(user), any(List.class), any(DefaultRosterListener.UpdateFilter.class), any(RosterNotify.HashProvider.class));
    }

    @Test
    public void testUpdateFilter() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        DefaultRosterListener.UpdateFilter filter = listener.new UpdateFilter();

        User updated = as.user(1);
        PacketBuilder.AddUserPacket packet = new PacketBuilder.AddUserPacket(as.bare(updated), RosterAbstract.SubscriptionType.both, updated.getFullName(), new String[]{"groups"});

        //block if recipient is the same as updated user
        assertNull(filter.pass(as.bare(updated), packet));

        //pass as is if user is not managed by personal roster
        User recipient = as.user(2);
        assertThat(filter.pass(as.bare(recipient), packet), is(packet));

        //pass merged if user is managed by personal roster and not deleted  - neither name nor group is managed by personal roster
        PersonalRoster.Buddy personalBuddy = new PersonalRoster.Buddy(as.bare(updated), null, null, PersonalRoster.Buddy.FLAG.updated);
        when(this.personalRosterRoster.find(as.bare(updated))).thenReturn(personalBuddy);

        assertThat(filter.pass(as.bare(recipient), packet), is(packet)); //neither name nor group is managed by personal roster
        verify(this.personalRosterRoster, times(2)).find(eq(as.bare(updated)));

        //pass merged if user is managed by personal roster and not deleted - name and group changed
        PersonalRoster.Buddy personalBuddyChanged = new PersonalRoster.Buddy(as.bare(updated), "changed name", new String[]{"changed group"}, PersonalRoster.Buddy.FLAG.updated);
        PacketBuilder.AddUserPacket changedPacket = packet.setName("changed name").setGroups(new String[]{"changed group"});
        when(this.personalRosterRoster.find(as.bare(updated))).thenReturn(personalBuddyChanged);

        assertThat(filter.pass(as.bare(recipient), packet), is(changedPacket)); //user is deleted
        verify(this.personalRosterRoster, times(3)).find(eq(as.bare(updated)));

        //block if user is managed by personal roster and deleted
        PersonalRoster.Buddy personalBuddyDeleted = new PersonalRoster.Buddy(as.bare(updated), null, null, PersonalRoster.Buddy.FLAG.deleted);
        when(this.personalRosterRoster.find(any(BareJID.class))).thenReturn(personalBuddyDeleted);

        assertNull(filter.pass(as.bare(recipient), packet)); //user is deleted by personal roster
        verify(this.personalRosterRoster, times(4)).find(as.bare(updated));

    }

    @Test
    public void testUpdateInvisibleOrDisabled_otherRosters() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User disabledInvisible = as.user(1, false, false);
        User disabledVisible = as.user(2, false, true);
        User enabledInvisible = as.user(3, true, false);

        //none of these users should be in default roster and on each of this change it should try to markRemoved it from all rosters
        //enabled -> disabled, user is invisible
        listener.update(disabledInvisible, EnumSet.of(XmppUsers.Changes.STATUS));
        verify(this.rosterNotify, times(1)).aboutRemove(eq(disabledInvisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //visible -> invisible, user is disabled
        listener.update(disabledInvisible, EnumSet.of(XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(2)).aboutRemove(eq(disabledInvisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //visible -> invisible && enabled -> disabled
        listener.update(disabledInvisible, EnumSet.of(XmppUsers.Changes.STATUS, XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(3)).aboutRemove(eq(disabledInvisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //enabled -> disabled, user is visible
        listener.update(disabledVisible, EnumSet.of(XmppUsers.Changes.STATUS));
        verify(this.rosterNotify, times(1)).aboutRemove(eq(disabledVisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //invisible -> visible, user is disabled
        listener.update(disabledVisible, EnumSet.of(XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(2)).aboutRemove(eq(disabledVisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //invisible -> visible && enabled -> disabled
        listener.update(disabledVisible, EnumSet.of(XmppUsers.Changes.STATUS, XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(3)).aboutRemove(eq(disabledVisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //disabled -> enabled, user is invisible
        listener.update(enabledInvisible, EnumSet.of(XmppUsers.Changes.STATUS));
        verify(this.rosterNotify, times(1)).aboutRemove(eq(enabledInvisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //visible -> invisible, user is enabled
        listener.update(enabledInvisible, EnumSet.of(XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(2)).aboutRemove(eq(enabledInvisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        //visible -> invisible && disabled -> enabled
        listener.update(enabledInvisible, EnumSet.of(XmppUsers.Changes.STATUS, XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(3)).aboutRemove(eq(enabledInvisible), any(List.class), any(DefaultRosterListener.RemovedFilter.class), any(RosterNotify.HashProvider.class));

        verify(this.rosterNotify, never()).aboutAdd(any(User.class), any(List.class), any(RosterNotify.Filter.class), any(RosterNotify.HashProvider.class));
    }

    @Test
    public void testUpdateInvisibleOrDisabled_updatedRoster() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        User disabledInvisible = as.user(1, false, false);
        User disabledVisible = as.user(2, false, true);
        User enabledInvisible = as.user(3, true, false);
        User enabledVisible = as.user(4); //marked as added
        User normal5 = as.user(5); //marked as added
        User normal6 = as.user(6); //marked as removed
        User normal7 = as.user(7); //marked as updated

        toDefault(enabledVisible, normal5, normal6, normal7);

        Map<BareJID, PersonalRoster.Buddy> map = new HashMap<BareJID, PersonalRoster.Buddy>();
        map.put(as.bare(normal5), new PersonalRoster.Buddy(as.bare(normal5), null, null, PersonalRoster.Buddy.FLAG.added));
        map.put(as.bare(normal6), new PersonalRoster.Buddy(as.bare(normal6), null, null, PersonalRoster.Buddy.FLAG.deleted));
        map.put(as.bare(normal7), new PersonalRoster.Buddy(as.bare(normal7), "name", null, PersonalRoster.Buddy.FLAG.updated));
        when(this.personalRosterRoster.asMap()).thenReturn(map);

        listener.update(disabledInvisible, EnumSet.of(XmppUsers.Changes.NAME)); //nothing changes for this user considering roster content
        listener.update(disabledInvisible, EnumSet.of(XmppUsers.Changes.STATUS)); //user is going to be offline until enabled, not doing anything
        listener.update(disabledInvisible, EnumSet.of(XmppUsers.Changes.VISIBILITY)); //user was disabled already, cant be online 
        listener.update(disabledVisible, EnumSet.of(XmppUsers.Changes.NAME)); //nothing changes for this user considering roster content
        listener.update(disabledVisible, EnumSet.of(XmppUsers.Changes.STATUS)); //user is going to be offline until enabled, not doing anything
        listener.update(disabledVisible, EnumSet.of(XmppUsers.Changes.VISIBILITY)); ////user was disabled already, cant be online 
        listener.update(enabledInvisible, EnumSet.of(XmppUsers.Changes.NAME)); //nothing changes for this user considering roster content
        listener.update(enabledInvisible, EnumSet.of(XmppUsers.Changes.STATUS)); ////user was disabled so cant be online yet
        listener.update(enabledVisible, EnumSet.of(XmppUsers.Changes.NAME)); //nothing changes for this user considering roster content
        listener.update(enabledVisible, EnumSet.of(XmppUsers.Changes.STATUS)); //user was disabled so cant be online yet

        verify(this.rosterNotify, times(0)).aboutAdd(any(List.class), any(BareJID.class), any(RosterNotify.HashProvider.class));
        verify(this.rosterNotify, times(0)).aboutRemove(any(JID[].class), any(BareJID.class), any(RosterNotify.HashProvider.class));

        listener.update(enabledInvisible, EnumSet.of(XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(0)).aboutAdd(any(List.class), any(BareJID.class), any(RosterNotify.HashProvider.class));
        verify(this.rosterNotify, times(1)).aboutRemove(eq(new JID[]{as.jid(enabledVisible), as.jid(normal7)}), eq(as.bare(enabledInvisible)), any(RosterNotify.HashProvider.class));

        listener.update(enabledVisible, EnumSet.of(XmppUsers.Changes.VISIBILITY));
        verify(this.rosterNotify, times(1)).aboutAdd(eq(Arrays.asList(mapper.value(normal7, map.get(as.bare(normal7))))), eq(as.bare(enabledVisible)), any(RosterNotify.HashProvider.class)); //normal7
        verify(this.rosterNotify, times(1)).aboutRemove(any(JID[].class), any(BareJID.class), any(RosterNotify.HashProvider.class)); //1 call from previous call, nothing changed
    }

//    public void testResetSingleUserNotBroken() throws SQLException {
//    public void testResetSingleUserMissing() throws SQLException {
//    public void testResetSingleUserExtra() throws SQLException {
//    public void testInvalidGroup() throws SQLException {
//    public void testResetEmptyGroup() throws SQLException {
//    public void testResetOwnerEmptyGroup() throws SQLException {
//    public void testSingleUserInvisibleDisabledExtra() throws SQLException {
//    public void testResetSingleUserSubscriptionUserOnly() throws SQLException {
//    public void testResetSingleUserSubscriptionAlsoOthers() throws SQLException {
//    public void testResetSingleUserSubscriptionAlsoOthersDisabledAndInvisible() throws SQLException {
//    public void testResetManagedStatus() throws SQLException {
//    public void testResetMultiple() throws SQLException {
    //TODO do not forget to test PersonalRoster throwing xmppexception on clear
    @Test
    public void testRepair() {
    }

    @Test
    public void testCheck_StaticEqualsDynamic() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //default roster - 4 users
        toDefault(as.user(1), as.user(2), as.user(3), as.user(4));

        //personal roster - empty
        //static roster - 4 users, equals to default without the owner
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(staticRosterContent, as.user(2), as.user(3), as.user(4));

        listener.check(as.user(1), staticRosterContent);

        //all static removed as present in default
        assertThat(staticRosterContent.size(), is(0));

        //no change detected against default, no personal entries
        verify(this.personalRosterRoster, times(1)).remove(eq(as.bare(2)));
        verify(this.personalRosterRoster, times(1)).remove(eq(as.bare(3)));
        verify(this.personalRosterRoster, times(1)).remove(eq(as.bare(4)));
        verify(this.personalRosterRoster, times(0)).markAdded(any(BareJID.class));
        verify(this.personalRosterRoster, times(0)).update(any(BareJID.class), any(String.class), any(String[].class));
        verify(this.personalRosterRoster, times(0)).markRemoved(any(BareJID.class));
    }

    @Test
    public void testCheck_PersonalNone() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                as.user(1),
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //4, 5 should be left in static roster
        assertThat(staticRosterContent.size(), is(2));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));

        //2, 6 left as is in default (we do not mark removed those who are not in static roster)
        //4, 5, 7 stored to personal
        assertThat(personalRosterContent.size(), is(3));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), CoreMatchers.allOf(bd.name("custom name 7"), bd.groups((String) null)));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4, 5 should be left in static roster
        assertThat(staticRosterContent.size(), is(2));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));

        //2, 6 left as is in default (we do not mark removed those who are not in static roster)
        //4, 5, 7 stored to personal
        assertThat(personalRosterContent.size(), is(3));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), CoreMatchers.allOf(bd.name("custom name 7"), bd.groups((String) null)));
    }

    @Test
    public void testCheck_PersonalDeleted() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                as.user(1),
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        personalRosterContent.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(7), new PersonalRoster.Buddy(as.bare(7), null, null, PersonalRoster.Buddy.FLAG.deleted));

        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //4, 5 should be left in static
        assertThat(staticRosterContent.size(), is(2));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));

        assertThat(personalRosterContent.size(), is(5)); //2 (not in static), 3(not in static), 4,5,7 (see below)

        //2 left as is, 6 removed from personal (as it is not removed from static)
        assertFalse(personalRosterContent.containsKey(as.bare(6)));
        //4, 5, 7 stored to personal
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), CoreMatchers.allOf(bd.name("custom name 7"), bd.groups((String) null)));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4, 5 should be left in static
        assertThat(staticRosterContent.size(), is(2));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));

        assertThat(personalRosterContent.size(), is(5));

        //2 left as is, 6 removed from personal (as it is not removed from static)
        assertFalse(personalRosterContent.containsKey(as.bare(6)));
        //4, 5, 7 stored to personal
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), CoreMatchers.allOf(bd.name("custom name 7"), bd.groups((String) null)));
    }

    @Test
    public void testCheck_PersonalUpdated() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                as.user(1),
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        personalRosterContent.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), "u name 2", new String[]{"u group 2"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), "u name 3", new String[]{"u group 3"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), "u name 4", new String[]{"u group 4"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), "u name 5", new String[]{"u group 5"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), "u name 6", new String[]{"u group 6"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(7), new PersonalRoster.Buddy(as.bare(7), "u name 7", new String[]{"u group 7"}, PersonalRoster.Buddy.FLAG.updated));

        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //4, 5 should be left in static
        assertThat(staticRosterContent.size(), is(2));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));

        assertThat(personalRosterContent.size(), is(5)); //2 (not in static), 3(not in static), 4,5,7 (see below)

        //2 left as is, 6 removed from personal (as it is not removed from static)
        assertFalse(personalRosterContent.containsKey(as.bare(6)));
        //4, 5, 7 stored to personal
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), CoreMatchers.allOf(bd.name("custom name 7"), bd.groups((String) null)));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4, 5 should be left in static
        assertThat(staticRosterContent.size(), is(2));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));

        assertThat(personalRosterContent.size(), is(5)); //2 (not in static), 3(not in static), 4,5,7 (see below)

        //2 left as is, 6 removed from personal (as it is not removed from static)
        assertFalse(personalRosterContent.containsKey(as.bare(6)));
        //4, 5, 7 stored to personal
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), CoreMatchers.allOf(bd.name("custom name 7"), bd.groups((String) null)));
    }

    @Test
    public void testCheck_PersonalAdded() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                as.user(1),
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        personalRosterContent.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), "u name 2", new String[]{"u group 2"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), "u name 3", new String[]{"u group 3"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), "u name 4", new String[]{"u group 4"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), "u name 5", new String[]{"u group 5"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), "u name 6", new String[]{"u group 6"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(7), new PersonalRoster.Buddy(as.bare(7), "u name 7", new String[]{"u group 7"}, PersonalRoster.Buddy.FLAG.added));

        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //all managed (in personal) left in static
        assertThat(staticRosterContent.size(), is(4));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));
        assertThat(staticRosterContent.get(as.bare(6)), CoreMatchers.allOf(re.name(as.user(6).getFullName()), re.group(groups.group(as.user(6).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(7)), CoreMatchers.allOf(re.name("custom name 7"), re.group(null)));

        assertThat(personalRosterContent.size(), is(6)); //all managed (added or modified) are left as is

        //all in personal left as is
        assertThat(personalRosterContent.get(as.bare(2)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(3)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));

        //no further change on second run
        listener.check(as.user(1), staticRosterContent);

        //all managed (in personal) left in static
        assertThat(staticRosterContent.size(), is(4));
        assertThat(staticRosterContent.get(as.bare(4)), CoreMatchers.allOf(re.name(as.user(4).getFullName()), re.group(groups.group(as.user(4).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(5)), CoreMatchers.allOf(re.name(as.user(5).getLogin()), re.group("group1")));
        assertThat(staticRosterContent.get(as.bare(6)), CoreMatchers.allOf(re.name(as.user(6).getFullName()), re.group(groups.group(as.user(6).getPrimaryEmail()))));
        assertThat(staticRosterContent.get(as.bare(7)), CoreMatchers.allOf(re.name("custom name 7"), re.group(null)));

        assertThat(personalRosterContent.size(), is(6)); //all managed (added or modified) are left as is

        //all in personal left as is
        assertThat(personalRosterContent.get(as.bare(2)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(3)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));
    }

    @Test
    public void testCheck_StaticDifferentDomain() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                as.user(1),
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.domain(4, "custom"),
                as.domain(5, "custom"),
                as.domain(6, "custom"),
                as.domain(7, "custom"));

        toStaticUpdate(staticRosterContent, as.bare(as.domain(5, "custom")), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(as.domain(7, "custom")), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //no change to static roster at all
        //no change to personal roster at all
        assertThat(staticRosterContent.size(), is(4));
        assertThat(personalRosterContent.size(), is(0));

        listener.check(as.user(1), staticRosterContent);

        //no change to static roster at all
        //no change to personal roster at all
        assertThat(staticRosterContent.size(), is(4));
        assertThat(personalRosterContent.size(), is(0));

    }

    @Test
    public void testInsertOverLimit() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        listener.add(as.user(1));
        listener.add(as.user(2));
        listener.add(as.user(3));
        listener.add(as.user(4));
        listener.add(as.user(5));
        listener.add(as.user(6));
        listener.add(as.user(7));
        listener.add(as.user(8));
        listener.add(as.user(9));
        listener.add(as.user(10));

        assertThat(this.defaultRosterContent.size(), is(10));

        listener.add(as.user(11));
        assertThat(this.defaultRosterContent.size(), is(11)); //roster of the length 10 means we need to store 10 + 1 owner of contact list who is not counted

        listener.add(as.user(12));
        assertThat(this.defaultRosterContent.size(), is(11));

        listener.remove(as.user(12));
        assertThat(this.defaultRosterContent.size(), is(11));

        listener.remove(as.user(11));
        assertThat(this.defaultRosterContent.size(), is(10));

        listener.remove(as.user(10));
        assertThat(this.defaultRosterContent.size(), is(9));

    }

    @Test
    public void testCheck_PersonalNone_emptyGroup() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);
        toDefault(as.user(1), as.user(2), as.user(3));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(2),
                as.user(3));

        //empty personal
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        //null groups
        toStaticUpdate(staticRosterContent, as.bare(2), null, null);
        toStaticUpdate(staticRosterContent, as.bare(3), "custom name 3", null);

        listener.check(as.user(1), staticRosterContent);

        assertThat(staticRosterContent.size(), is(0));
        assertThat(personalRosterContent.size(), is(2));
        assertThat(personalRosterContent.get(as.bare(2)), CoreMatchers.allOf(bd.groups((String) null), bd.notManaged(), bd.name(as.bare(2).getLocalpart())));
        assertThat(personalRosterContent.get(as.bare(3)), CoreMatchers.allOf(bd.groups((String) null), bd.notManaged(), bd.name("custom name 3")));

        listener.check(as.user(1), staticRosterContent);

        assertThat(staticRosterContent.size(), is(0));
        assertThat(personalRosterContent.size(), is(2));
        assertThat(personalRosterContent.get(as.bare(2)), CoreMatchers.allOf(bd.groups((String) null), bd.notManaged(), bd.name(as.bare(2).getLocalpart())));
        assertThat(personalRosterContent.get(as.bare(3)), CoreMatchers.allOf(bd.groups((String) null), bd.notManaged(), bd.name("custom name 3")));

    }

    @Test
    public void testCheck_PersonalNone_ownerInvisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner, invisible (not present in default)
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(4));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(4));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));
    }

    @Test
    public void testCheck_PersonalDeleted_ownerInvisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner - invisible, not in default
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                /* as.user(1),*/
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        personalRosterContent.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), null, null, PersonalRoster.Buddy.FLAG.deleted));
        personalRosterContent.put(as.bare(7), new PersonalRoster.Buddy(as.bare(7), null, null, PersonalRoster.Buddy.FLAG.deleted));

        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(6));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(6));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));
    }

    @Test
    public void testCheck_PersonalUpdated_ownerInvisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner, invisible
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                /*as.user(1), */
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        personalRosterContent.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), "u name 2", new String[]{"u group 2"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), "u name 3", new String[]{"u group 3"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), "u name 4", new String[]{"u group 4"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), "u name 5", new String[]{"u group 5"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), "u name 6", new String[]{"u group 6"}, PersonalRoster.Buddy.FLAG.updated));
        personalRosterContent.put(as.bare(7), new PersonalRoster.Buddy(as.bare(7), "u name 7", new String[]{"u group 7"}, PersonalRoster.Buddy.FLAG.updated));

        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(6));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(6));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));
    }

    @Test
    public void testCheck_PersonalAdded_ownerInvisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner, invisible
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                /*as.user(1), */
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.user(4),
                as.user(5),
                as.user(6),
                as.user(7));

        toStaticUpdate(staticRosterContent, as.bare(5), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(7), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        personalRosterContent.put(as.bare(2), new PersonalRoster.Buddy(as.bare(2), "u name 2", new String[]{"u group 2"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(3), new PersonalRoster.Buddy(as.bare(3), "u name 3", new String[]{"u group 3"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(4), new PersonalRoster.Buddy(as.bare(4), "u name 4", new String[]{"u group 4"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(5), new PersonalRoster.Buddy(as.bare(5), "u name 5", new String[]{"u group 5"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(6), new PersonalRoster.Buddy(as.bare(6), "u name 6", new String[]{"u group 6"}, PersonalRoster.Buddy.FLAG.added));
        personalRosterContent.put(as.bare(7), new PersonalRoster.Buddy(as.bare(7), "u name 7", new String[]{"u group 7"}, PersonalRoster.Buddy.FLAG.added));

        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(6));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));

        //on second run the state must not change!
        listener.check(as.user(1), staticRosterContent);

        //4,5,6,7 should be left in static roster
        assertThat(staticRosterContent.size(), is(4));

        //4,5,6,7 should be marked as added
        assertThat(personalRosterContent.size(), is(6));
        assertThat(personalRosterContent.get(as.bare(4)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(5)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(6)), is(bd.added()));
        assertThat(personalRosterContent.get(as.bare(7)), is(bd.added()));
    }

    @Test
    public void testCheck_StaticDifferentDomain_ownerInvisible() {
        DefaultRosterListener listener = new DefaultRosterListener(this.defaultRoster, this.personalRoster, this.rosterNotify, this.groups);

        //1 - roster owner - invisible
        //2 - dynamic yes, static no (we do not remove removed)
        //3 - dynamic no, static no
        //4 - dynamic no, static yes, as default
        //5 - dynamic no, static yes edited (nullvalues should not be updated by default)
        //6 - dynamic yes, static yes as default
        //7 - dynamic yes, static yes, edited (nullvalues should not be updated by default)
        toDefault(
                /*as.user(1), */
                as.user(2),
                as.user(6),
                as.user(7));

        //static roster
        Map<BareJID, RosterElement> staticRosterContent = new HashMap<BareJID, RosterElement>();
        toStatic(
                staticRosterContent,
                as.domain(4, "custom"),
                as.domain(5, "custom"),
                as.domain(6, "custom"),
                as.domain(7, "custom"));

        toStaticUpdate(staticRosterContent, as.bare(as.domain(5, "custom")), null, new String[]{"group1"});
        toStaticUpdate(staticRosterContent, as.bare(as.domain(7, "custom")), "custom name 7", null);

        //personal changes
        Map<BareJID, PersonalRoster.Buddy> personalRosterContent = new HashMap<BareJID, PersonalRoster.Buddy>();
        XmppUsersStore store = mock(XmppUsersStore.class);
        when(this.personalRoster.of(as.bare(1))).thenReturn(new PersonalRoster.PersonalUserRoster(as.bare(1), store, personalRosterContent));

        listener.check(as.user(1), staticRosterContent);

        //no change to static roster at all
        //no change to personal roster at all
        assertThat(staticRosterContent.size(), is(4));
        assertThat(personalRosterContent.size(), is(0));

        listener.check(as.user(1), staticRosterContent);

        //no change to static roster at all
        //no change to personal roster at all
        assertThat(staticRosterContent.size(), is(4));
        assertThat(personalRosterContent.size(), is(0));

    }

    @Test
    public void testBye() {
    }

}
