/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.tigase.roster.PersonalRoster.Buddy;
import com.kerio.im.tigase.user.TestPacketSender;
import com.kerio.im.tigase.user.TestSessionProvider;
import com.kerio.im.tigase.user.UserMapper;
import com.kerio.im.tigase.user.UserUtil;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import static org.hamcrest.CoreMatchers.is;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.mockito.Mockito.*;
import tigase.xmpp.BareJID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.XMPPResourceConnection;
import static org.junit.Assert.*;
import tigase.xml.Element;
import tigase.xmpp.JID;
import tigase.xmpp.impl.roster.RosterAbstract;

/**
 *
 * @author tskoda
 */
public class DynamicDomainRosterTest {

    private static final UserUtil as = new UserUtil();
    private static final PacketUtil element = new PacketUtil();

    private static final BuddyUtil buddy = new BuddyUtil();

    private Map<BareJID, RosterContent.Buddy> defaultRosterContentMap;
    private RosterContent defaultRosterContent;

    private PersonalRoster.PersonalUserRoster personalRosterContent;
    private PersonalRoster personalRoster;
    private DefaultRoster defaultRoster;
    private GroupService groupService;
    private SessionUtil sessionUtil;

    private ElementMapper mapper;

    public DynamicDomainRosterTest() {
    }

    @Before
    public void setUp() {
        this.groupService = new TestGroupService();
        this.mapper = new ElementMapper(groupService, new UserMapper());
        
        this.defaultRosterContentMap = new HashMap<BareJID, RosterContent.Buddy>();
        this.defaultRosterContent = new RosterContent(defaultRosterContentMap, mapper, 10, false);
        this.defaultRoster = mock(DefaultRoster.class);
        when(this.defaultRoster.roster(any(String.class))).thenReturn(defaultRosterContent);

        this.personalRosterContent = mock(PersonalRoster.PersonalUserRoster.class);
        this.personalRoster = mock(PersonalRoster.class);
        when(this.personalRoster.of(any(BareJID.class))).thenReturn(personalRosterContent);
        
        this.sessionUtil = mock(SessionUtil.class);
        
    }

    @After
    public void tearDown() {
    }

    private XMPPResourceConnection session(final BareJID jid) throws NotAuthorizedException {
        XMPPResourceConnection result = mock(XMPPResourceConnection.class);
        when(this.sessionUtil.ownerOf(any(XMPPResourceConnection.class))).thenReturn(jid);
        sessionAllLocal();

        return result;
    }

    private void sessionAllRemote() throws NotAuthorizedException {
        when(this.sessionUtil.localDomain(any(XMPPResourceConnection.class))).thenReturn(false);
    }

    private void sessionAllLocal() throws NotAuthorizedException {
        when(this.sessionUtil.localDomain(any(XMPPResourceConnection.class))).thenReturn(true);
    }

    private void toDefault(User user) {
        defaultRosterContentMap.put(as.bare(user), mapper.value(user));
    }

    private static Map<BareJID, Element> toMap(List<Element> list) {
        Map<BareJID, Element> map = new HashMap<BareJID, Element>();
        for (Element el : list) {
            map.put(BareJID.bareJIDInstanceNS(el.getAttribute("jid")), el);
        }
        return map;
    }

    @Test
    public void testManagesEmptyPrivate() throws Exception {
        DynamicDomainRoster ddr = new DynamicDomainRoster(defaultRoster, personalRoster, sessionUtil, mapper);

        BareJID owner1 = as.bare(1);
        JID buddy2 = as.jid(2, 0);
        XMPPResourceConnection session1 = session(owner1);

        //checking if user 1 hase user 2 in his roster
        //empty roster
        assertFalse(ddr.manages(session1, buddy2));

        //only user 3 is in default roster
        toDefault(as.user(3));
        assertFalse(ddr.manages(session1, buddy2));

        //both user 1 and user 3 are in default roster
        toDefault(as.user(1));
        assertFalse(ddr.manages(session1, buddy2));

        //user 2 from different domain is in defalut roster
        toDefault(as.domain(2, "another-domain"));
        assertFalse(ddr.manages(session1, buddy2));

        //user 2 considered as from remote domian
        sessionAllRemote();
        assertFalse(ddr.manages(session1, buddy2));

        //user 2 is present id default roster
        sessionAllLocal();
        toDefault(as.user(2));
        assertTrue(ddr.manages(session1, buddy2));
    }

    @Test
    public void testManagesPrivate() throws Exception {
        DynamicDomainRoster ddr = new DynamicDomainRoster(defaultRoster, personalRoster, sessionUtil, mapper);
        BareJID owner1 = as.bare(1);
        toDefault(as.user(1));
        XMPPResourceConnection session1 = session(owner1);

        JID buddy2 = as.jid(2, 0);
        toDefault(as.user(2));

        //only updated in private
        Buddy updated = new Buddy(buddy2.getBareJID(), "custom buddy 2 name", new String[]{"group1"}, PersonalRoster.Buddy.FLAG.updated);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(updated);
        assertTrue(ddr.manages(session1, buddy2));

        //added in private
        Buddy added = new Buddy(buddy2.getBareJID(), null, null, PersonalRoster.Buddy.FLAG.added);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(added);
        assertTrue(ddr.manages(session1, buddy2));

        //deleted in private
        Buddy removed = new Buddy(buddy2.getBareJID(), null, null, PersonalRoster.Buddy.FLAG.deleted);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(removed);
        assertTrue(ddr.manages(session1, buddy2));
    }

    @Test
    public void testGetBuddySubscription() throws Exception {
        DynamicDomainRoster ddr = new DynamicDomainRoster(defaultRoster, personalRoster, sessionUtil, mapper);
        BareJID owner1 = as.bare(1);
        toDefault(as.user(1));
        XMPPResourceConnection session1 = session(owner1);

        JID buddy2 = as.jid(2, 0);

        //not in default roster
        assertThat(ddr.getBuddySubscription(session1, buddy2), is(RosterAbstract.SubscriptionType.none));

        //not in the same domain
        toDefault(as.domain(2, "another domain"));
        assertThat(ddr.getBuddySubscription(session1, buddy2), is(RosterAbstract.SubscriptionType.none));

        //only added in private
        Buddy added = new Buddy(buddy2.getBareJID(), null, null, PersonalRoster.Buddy.FLAG.added);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(added);
        assertThat(ddr.getBuddySubscription(session1, buddy2), is(RosterAbstract.SubscriptionType.none));

        //in default
        toDefault(as.user(2));
        assertThat(ddr.getBuddySubscription(session1, buddy2), is(RosterAbstract.SubscriptionType.both));

        //deleted from private
        Buddy removed = new Buddy(buddy2.getBareJID(), null, null, PersonalRoster.Buddy.FLAG.deleted);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(removed);
        assertThat(ddr.getBuddySubscription(session1, buddy2), is(RosterAbstract.SubscriptionType.none));
    }

    @Test
    public void testGetBuddyItem() throws Exception {
        DynamicDomainRoster ddr = new DynamicDomainRoster(defaultRoster, personalRoster, sessionUtil, mapper);
        BareJID owner1 = as.bare(1);
        toDefault(as.user(1));
        XMPPResourceConnection session1 = session(owner1);

        JID buddy2 = as.jid(2, 0);

        //not in default
        assertNull(ddr.getBuddyItem(session1, buddy2));

        //not in the same domain
        toDefault(as.domain(2, "another domain"));
        assertNull(ddr.getBuddyItem(session1, buddy2));

        //in default
        toDefault(as.user(2));
        assertThat(ddr.getBuddyItem(session1, buddy2), is(element.element(as.bare(2), "both", as.user(2).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));

        //updated in private
        Buddy updated = new Buddy(buddy2.getBareJID(), "custom buddy 2 name", new String[]{"group1"}, PersonalRoster.Buddy.FLAG.updated);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(updated);
        assertThat(ddr.getBuddyItem(session1, buddy2), is(element.element(as.bare(2), "both", "custom buddy 2 name", Arrays.asList("group1"))));

        //deleted from private
        Buddy removed = new Buddy(buddy2.getBareJID(), null, null, PersonalRoster.Buddy.FLAG.deleted);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(removed);
        assertNull(ddr.getBuddyItem(session1, buddy2));
        
        //added to private
        Buddy added = new Buddy(buddy2.getBareJID(), null, null, PersonalRoster.Buddy.FLAG.added);
        when(personalRosterContent.find(buddy2.getBareJID())).thenReturn(added);
        assertNull(ddr.getBuddyItem(session1, buddy2));
    }

    @Test
    public void testGetBuddies() throws Exception {
        DynamicDomainRoster ddr = new DynamicDomainRoster(defaultRoster, personalRoster, sessionUtil, mapper);
        BareJID owner1 = as.bare(1);
        XMPPResourceConnection session1 = session(owner1);

        //empty default
        assertThat(ddr.getBuddies(session1).length, is(0));

        //some in default
        toDefault(as.user(1));
        toDefault(as.user(2));
        toDefault(as.user(3));
        toDefault(as.user(4));
        toDefault(as.user(5));

        assertThat(ddr.getBuddies(session1).length, is(4)); //without owner

        //some in private
        Map<BareJID, PersonalRoster.Buddy> map = new HashMap<BareJID, Buddy>();
        map.put(as.bare(1), new Buddy(as.bare(1), null, null, PersonalRoster.Buddy.FLAG.deleted)); //deleted
        map.put(as.bare(2), new Buddy(as.bare(2), "buddy 2 custom name", null, PersonalRoster.Buddy.FLAG.updated)); //edited
        map.put(as.bare(3), new Buddy(as.bare(3), null, new String[]{"group 1"}, PersonalRoster.Buddy.FLAG.updated)); //edited
        map.put(as.bare(5), new Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.added)); //added
        map.put(as.bare(6), new Buddy(as.bare(6), null, null, PersonalRoster.Buddy.FLAG.added)); //added
        map.put(as.bare(7), new Buddy(as.bare(7), null, null, PersonalRoster.Buddy.FLAG.updated)); //added
        when(personalRosterContent.asMap()).thenReturn(map);

        JID[] expected = new JID[]{as.jid(2), as.jid(3), as.jid(4)};
        Arrays.sort(expected);

        JID[] current = ddr.getBuddies(session1);
        Arrays.sort(current);

        assertThat(current, is(expected));
    }

    @Test
    public void testGetRosterItems() throws Exception {
        DynamicDomainRoster ddr = new DynamicDomainRoster(defaultRoster, personalRoster, sessionUtil, mapper);
        BareJID owner1 = as.bare(1);
        XMPPResourceConnection session1 = session(owner1);

        //empty default
        assertThat(ddr.getRosterItems(session1).size(), is(0));

        //some in default
        toDefault(as.user(1));
        toDefault(as.user(2));
        toDefault(as.user(3));
        toDefault(as.user(4));
        toDefault(as.user(5));

        assertThat(ddr.getRosterItems(session1).size(), is(4)); //without owner

        //some in private
        Map<BareJID, PersonalRoster.Buddy> map = new HashMap<BareJID, Buddy>();
        map.put(as.bare(1), new Buddy(as.bare(1), null, null, PersonalRoster.Buddy.FLAG.deleted)); //deleted
        map.put(as.bare(2), new Buddy(as.bare(2), "buddy 2 custom name", null, PersonalRoster.Buddy.FLAG.updated)); //edited
        map.put(as.bare(3), new Buddy(as.bare(3), null, new String[]{"group 1"}, PersonalRoster.Buddy.FLAG.updated)); //edited
        map.put(as.bare(5), new Buddy(as.bare(5), null, null, PersonalRoster.Buddy.FLAG.added)); //added
        map.put(as.bare(6), new Buddy(as.bare(6), null, null, PersonalRoster.Buddy.FLAG.added)); //added
        map.put(as.bare(7), new Buddy(as.bare(7), null, null, PersonalRoster.Buddy.FLAG.updated)); //added
        when(personalRosterContent.asMap()).thenReturn(map);

        List<Element> current = ddr.getRosterItems(session1);
        Map<BareJID, Element> currentMap = toMap(current);

        assertThat(currentMap.size(), is(3));
        assertThat(currentMap.get(as.bare(2)), is(element.element(as.bare(2), "both", "buddy 2 custom name", Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(currentMap.get(as.bare(3)), is(element.element(as.bare(3), "both", as.user(3).getFullName(), Arrays.asList("group 1"))));
        assertThat(currentMap.get(as.bare(4)), is(element.element(as.bare(4), "both", as.user(4).getFullName(), Arrays.asList(TestGroupService.DEFAULT_GROUP))));   
    }
    
}
