/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.tigase.user.UserUtil;
import com.kerio.im.tigase.user.XmppUsersStore;
import com.kerio.im.tigase.user.XmppUsersStorePool;
import java.util.HashMap;
import java.util.Map;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import tigase.xmpp.BareJID;
import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;
import org.mockito.Mockito;
import static org.mockito.Mockito.*;

/**
 *
 * @author tskoda
 */
public class PersonalRosterTest {

    private static final UserUtil as = new UserUtil();
    private static final BuddyUtil buddy = new BuddyUtil();

    private XmppUsersStore userStoreMock;
    
    public PersonalRosterTest() {
    }

    @Before
    public void setUp() {
        this.userStoreMock = mock(XmppUsersStore.class);
    }

    @After
    public void tearDown() {
    }

    @Test
    public void testOf() {

        XmppUsersStorePool userStorePoolMock = mock(XmppUsersStorePool.class);
        when(userStorePoolMock.get(Mockito.any(String.class))).thenReturn(this.userStoreMock);
        
        Map<BareJID, PersonalRoster.PersonalUserRoster> content = new HashMap<BareJID, PersonalRoster.PersonalUserRoster>();
        PersonalRoster roster = new PersonalRoster(content, userStorePoolMock);

        assertThat(content.size(), is(0));

        PersonalRoster.PersonalUserRoster personalRoster1 = roster.of(as.bare(1));
        assertNotNull(personalRoster1);

        PersonalRoster.PersonalUserRoster personalRoster2 = roster.of(as.bare(2));
        assertNotNull(personalRoster2);

        assertSame(personalRoster1, roster.of(as.bare(1)));
    }
    
    @Test
    public void testModifyOwner() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(0), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        
        assertFalse(personalRoster.managed(as.bare(0)));
        personalRoster.markAdded(as.bare(0));
        assertFalse(personalRoster.managed(as.bare(0)));
        personalRoster.markRemoved(as.bare(0));
        assertFalse(personalRoster.managed(as.bare(0)));
        
        personalRoster.update(as.bare(0), "name", new String[]{"group1", "group2"});
        assertFalse(personalRoster.managed(as.bare(0)));
        assertNull(personalRoster.find(as.bare(0)));
        
    }

    @Test
    public void testAsMap() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(10), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        assertThat(personalRoster.asMap().size(), is(0));

        personalRoster.update(as.bare(0), "name", new String[]{"group"});
        assertThat(personalRoster.asMap().size(), is(1));

        personalRoster.update(as.bare(1), "name", new String[]{"group"});
        assertThat(personalRoster.asMap().size(), is(2));

        personalRoster.remove(as.bare(1));
        assertThat(personalRoster.asMap().size(), is(1));

        personalRoster.remove(as.bare(0));
        assertThat(personalRoster.asMap().size(), is(0));

    }

    @Test
    public void testMarkRemovedNotExist() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        PersonalRoster.Buddy removed;

        removed = personalRoster.find(as.bare(0));
        assertNull(removed);

        personalRoster.markRemoved(as.bare(0));

        removed = personalRoster.find(as.bare(0));
        assertNotNull(removed);
        assertThat(removed, allOf(buddy.removed(), buddy.name(null), buddy.groups((String[]) null)));
    }

    @Test
    public void testMarkRemovedExistAlreadyRemoved() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.markRemoved(as.bare(0));
        PersonalRoster.Buddy removed = personalRoster.find(as.bare(0));

        personalRoster.markRemoved(as.bare(0));
        PersonalRoster.Buddy removedSecondTime = personalRoster.find(as.bare(0));

        assertThat(removed, is(removedSecondTime));
    }

    @Test
    public void testMarkRemovedExistAdded() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());

        personalRoster.markAdded(as.bare(0));
        personalRoster.markRemoved(as.bare(0));

        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.removed(), buddy.name(null), buddy.groups((String[]) null)));
    }

    @Test
    public void testMarkRemovedExistJustUpdated() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(0), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.update(as.bare(1), "name1", new String[]{"group1", "group2"});
        personalRoster.update(as.bare(2), "name2", null);
        personalRoster.update(as.bare(3), null, new String[]{"group3", "group4"});

        personalRoster.markRemoved(as.bare(1));
        assertThat(personalRoster.find(as.bare(1)), allOf(buddy.removed(), buddy.name("name1"), buddy.groups("group1", "group2")));

        personalRoster.markRemoved(as.bare(2));
        assertThat(personalRoster.find(as.bare(2)), allOf(buddy.removed(), buddy.name("name2"), buddy.groups((String[]) null)));

        personalRoster.markRemoved(as.bare(3));
        assertThat(personalRoster.find(as.bare(3)), allOf(buddy.removed(), buddy.name(null), buddy.groups("group3", "group4")));
    }

    @Test
    public void testClearEmtpy() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        assertNull(personalRoster.find(as.bare(0)));

        personalRoster.clear();

        assertNull(personalRoster.find(as.bare(0)));
    }

    @Test
    public void testClearNonEmpty() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(2), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.markAdded(as.bare(0));
        personalRoster.markAdded(as.bare(1));

        assertNotNull(personalRoster.find(as.bare(0)));
        assertNotNull(personalRoster.find(as.bare(1)));

        personalRoster.clear();

        assertNull(personalRoster.find(as.bare(0)));
        assertNull(personalRoster.find(as.bare(1)));
    }

    @Test
    public void testManaged() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(4), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());

        //user does not exist
        assertFalse(personalRoster.managed(as.bare(0)));

        //just added is marked as managed
        personalRoster.markAdded(as.bare(0));
        personalRoster.markAdded(as.bare(1));
        assertTrue(personalRoster.managed(as.bare(0)));
        assertTrue(personalRoster.managed(as.bare(1)));

        //just removed is marked as managed
        personalRoster.markRemoved(as.bare(0));
        personalRoster.markRemoved(as.bare(2));
        assertTrue(personalRoster.managed(as.bare(0)));
        assertTrue(personalRoster.managed(as.bare(2)));

        //only updated are never marked as managed
        personalRoster.update(as.bare(1), "name1", new String[]{"group"});
        personalRoster.update(as.bare(2), "name2", new String[]{"group"});
        personalRoster.update(as.bare(3), "name3", new String[]{"group"});
        assertFalse(personalRoster.managed(as.bare(1))); //added flag is preserved on update
        assertFalse(personalRoster.managed(as.bare(2))); //deleted flag is removed on update
        assertFalse(personalRoster.managed(as.bare(3)));
    }

    @Test
    public void testUpdateNonExistent() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(3), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());

        assertNull(personalRoster.find(as.bare(0)));
        personalRoster.update(as.bare(0), "name0", null);
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.name("name0"), buddy.groups((String[]) null), buddy.notManaged()));

        assertNull(personalRoster.find(as.bare(1)));
        personalRoster.update(as.bare(1), "name1", new String[]{"group"});
        assertThat(personalRoster.find(as.bare(1)), allOf(buddy.name("name1"), buddy.groups("group"), buddy.notManaged()));

        assertNull(personalRoster.find(as.bare(2)));
        personalRoster.update(as.bare(2), null, new String[]{"group"});
        assertThat(personalRoster.find(as.bare(2)), allOf(buddy.name(null), buddy.groups("group"), buddy.notManaged()));

    }

    @Test
    public void testUpdateRemoved() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.markRemoved(as.bare(0));

        personalRoster.update(as.bare(0), "name0", new String[]{"group0"});
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.name("name0"), buddy.groups("group0"), buddy.notManaged()));
    }

    @Test
    public void testUpdateAdded() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(0), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.markAdded(as.bare(1));

        personalRoster.update(as.bare(1), "name1", new String[]{"group1"});
        assertThat(personalRoster.find(as.bare(1)), allOf(buddy.name("name1"), buddy.groups("group1"), buddy.notManaged()));
    }

    @Test
    public void testUpdateUpdated() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.update(as.bare(0), "name0", new String[]{"group0"});

        personalRoster.update(as.bare(0), "name1", new String[]{"group1"});
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.name("name1"), buddy.groups("group1"), buddy.notManaged()));

        personalRoster.update(as.bare(0), "name2", (String[]) null);
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.name("name2"), buddy.groups((String[]) null), buddy.notManaged()));

        personalRoster.update(as.bare(0), null, new String[]{"group2"});
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.name(null), buddy.groups("group2"), buddy.notManaged()));

    }
    
    @Test
    public void testUpdateEmptyGroup() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.update(as.bare(0), "name0", new String[]{null});
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.name("name0"), buddy.groups(new String[]{null}), buddy.notManaged()));
        
        personalRoster.update(as.bare(0), "name0", new String[]{null, "Colleagues"});
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.name("name0"), buddy.groups(new String[]{null, "Colleagues"}), buddy.notManaged()));
    }

    @Test
    public void testRemoveNonExists() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(1), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());

        personalRoster.remove(as.bare(0));
        assertNull(personalRoster.find(as.bare(0)));
    }

    @Test
    public void testRemove() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(10), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        personalRoster.markAdded(as.bare(0));
        personalRoster.markRemoved(as.bare(1));
        personalRoster.update(as.bare(2), "name", new String[]{"group"});

        personalRoster.remove(as.bare(0));
        assertNull(personalRoster.find(as.bare(0)));

        personalRoster.remove(as.bare(1));
        assertNull(personalRoster.find(as.bare(1)));

        personalRoster.remove(as.bare(2));
        assertNull(personalRoster.find(as.bare(2)));
    }

    @Test
    public void testMarkAddedNotExist() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(10), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());

        assertNull(personalRoster.find(as.bare(0)));
        personalRoster.markAdded(as.bare(0));
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.added(), buddy.name(null), buddy.groups((String[]) null)));
    }

    @Test
    public void testMarkAddedExistAlreadyAdded() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(10), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());

        personalRoster.markAdded(as.bare(0));
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.added(), buddy.name(null), buddy.groups((String[]) null)));

        personalRoster.markAdded(as.bare(0));
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.added(), buddy.name(null), buddy.groups((String[]) null)));
    }

    @Test
    public void testMarkAddedExistRemoved() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(10), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());

        personalRoster.markRemoved(as.bare(0));
        personalRoster.markAdded(as.bare(0));
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.added(), buddy.name(null), buddy.groups((String[]) null)));
    }

    @Test
    public void testMarkAddedJustUpdated() {
        PersonalRoster.PersonalUserRoster personalRoster = new PersonalRoster.PersonalUserRoster(as.bare(10), this.userStoreMock, new HashMap<BareJID, PersonalRoster.Buddy>());
        
        personalRoster.update(as.bare(0), "name", new String[]{"group"});
        personalRoster.markAdded(as.bare(0));
        assertThat(personalRoster.find(as.bare(0)), allOf(buddy.added(), buddy.name("name"), buddy.groups("group")));
        
    }

}
