/*
 *  Copyright 2013 Kerio Technologies s.r.o. All rights reserved.
 */
package com.kerio.im.tigase.roster;

import com.kerio.im.bean.User;
import com.kerio.im.core.api.service.GroupService;
import com.kerio.im.tigase.user.TestPacketSender;
import com.kerio.im.tigase.user.TestSessionProvider;
import com.kerio.im.tigase.user.UserUtil;
import java.util.Arrays;
import java.util.Collections;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import static org.junit.Assert.*;
import static org.hamcrest.CoreMatchers.is;
import static org.mockito.Mockito.*;
import tigase.xmpp.BareJID;
import tigase.xmpp.NotAuthorizedException;
import tigase.xmpp.impl.roster.RosterAbstract;

/**
 *
 * @author tskoda
 */
public class RosterNotifyTest {

    protected static final UserUtil as = new UserUtil();
    protected static final PacketUtil packets = new PacketUtil();

    private GroupService groups;
    private PacketBuilder packetBuilder;
    private TestSessionProvider sessions;
    private TestPacketSender packetSender;

    public RosterNotifyTest() {
    }

    @Before
    public void setUp() {
        this.groups = new TestGroupService();
        this.packetBuilder = new PacketBuilder();
        this.sessions = new TestSessionProvider();
        this.packetSender = new TestPacketSender();
    }

    @After
    public void tearDown() {
    }

    @Test
    public void testAdded() {
    }

    @Test
    public void testAboutAddEmptyRecipients() {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User added = as.user(1);

        notify.aboutAdd(added, Collections.EMPTY_LIST, new TestFilter(), mock(RosterNotify.HashProvider.class));

        //no notification should be produced
        assertThat(this.packetSender.count(), is(0));
    }

    @Test
    public void testAboutAddOffline() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User added = as.user(1);
        BareJID addedJID = as.bare(added);

        User offline = as.user(2);
        BareJID offlineJID = as.bare(offline);

        //notify all recipients about new user, sessions does report all offline by default, filter allows all by default
        notify.aboutAdd(added, Arrays.asList(addedJID, offlineJID), new TestFilter<PacketBuilder.AddUserPacket>(), mock(RosterNotify.HashProvider.class));

        //no notification should be produced
        assertThat(this.packetSender.count(), is(0));
    }

    @Test
    public void testAboutAddOnlineBlockAll() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User added = as.user(1);

        User onlineBlocked = as.user(2);
        BareJID onlineBlockedJID = as.bare(onlineBlocked);

        User onlineBlocked2 = as.user(3);
        BareJID onlineBlockedJID2 = as.bare(onlineBlocked2);

        //block both users
        TestFilter<PacketBuilder.AddUserPacket> filter = new TestFilter<PacketBuilder.AddUserPacket>();
        filter.block(onlineBlockedJID, onlineBlockedJID2);

        //report these as online
        this.sessions.online(onlineBlockedJID, onlineBlockedJID2);

        //notify all recipient about new user
        notify.aboutAdd(added, Arrays.asList(onlineBlockedJID, onlineBlockedJID2), filter, mock(RosterNotify.HashProvider.class));

        assertThat(filter.getCalls(), is(2));

        //no notification should be produced
        assertThat(this.packetSender.count(), is(0));

    }

    @Test
    public void testAboutAddOnlineAllow() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User added = as.user(1);

        User offline = as.user(2);
        BareJID offlineJID = as.bare(offline);

        User onlineBlocked = as.user(3);
        BareJID onlineBlockedJID = as.bare(onlineBlocked);

        User onlineAllowed = as.user(4);
        BareJID onlineAllowedJID = as.bare(onlineAllowed);

        User onlineAllowed2 = as.user(5);
        BareJID onlineAllowed2JID = as.bare(onlineAllowed2);

        //make all *Blocked blocked
        TestFilter<PacketBuilder.AddUserPacket> filter = new TestFilter<PacketBuilder.AddUserPacket>();
        filter.block(onlineBlockedJID);

        //make all online* online
        this.sessions.online(onlineBlockedJID, onlineAllowedJID, onlineAllowed2JID);

        //notify all recipient about new user
        notify.aboutAdd(added, Arrays.asList(offlineJID, onlineBlockedJID, onlineAllowedJID, onlineAllowed2JID), filter, mock(RosterNotify.HashProvider.class));

        //three recipients
        assertThat(filter.getCalls(), is(3));

        //only two online
        assertThat(this.packetSender.count(), is(2));
        assertThat(this.packetSender.get(onlineAllowedJID, 0), is(packets.added(onlineAllowed, 0, added, Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(this.packetSender.get(onlineAllowed2JID, 0), is(packets.added(onlineAllowed2, 0, added, Arrays.asList(TestGroupService.DEFAULT_GROUP))));
    }

    @Test
    public void testAboutAddTwoResources() {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User added = as.user(1);

        User onlineWithThreeResources = as.user(2);
        BareJID onlineWithThreeResourcesJID = as.bare(onlineWithThreeResources);

        //allows all by default
        TestFilter<PacketBuilder.AddUserPacket> filter = new TestFilter<PacketBuilder.AddUserPacket>();

        //initialize three resources
        this.sessions.online(onlineWithThreeResourcesJID);
        this.sessions.resources(onlineWithThreeResourcesJID, 3);

        notify.aboutAdd(added, Arrays.asList(onlineWithThreeResourcesJID), filter, mock(RosterNotify.HashProvider.class));

        //three recipients
        assertThat(filter.getCalls(), is(1));

        //produces three packets
        assertThat(this.packetSender.count(), is(3));
        assertThat(this.packetSender.get(onlineWithThreeResourcesJID, 0), is(packets.added(onlineWithThreeResources, 0, added, Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(this.packetSender.get(onlineWithThreeResourcesJID, 1), is(packets.added(onlineWithThreeResources, 1, added, Arrays.asList(TestGroupService.DEFAULT_GROUP))));
        assertThat(this.packetSender.get(onlineWithThreeResourcesJID, 2), is(packets.added(onlineWithThreeResources, 2, added, Arrays.asList(TestGroupService.DEFAULT_GROUP))));
    }

    @Test
    public void testAboutAddModifiedPacket() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User added = as.user(1);

        User recipient = as.user(2);
        BareJID recipientJID = as.bare(recipient);

        PacketBuilder.AddUserPacket modifiedPacket = new PacketBuilder.AddUserPacket(as.bare(added), RosterAbstract.SubscriptionType.both, "changed name", new String[]{"group1", "group2"});
        TestFilter<PacketBuilder.AddUserPacket> filter = new TestFilter<PacketBuilder.AddUserPacket>();
        filter.passWith(recipientJID, modifiedPacket);

        this.sessions.online(recipientJID);

        //notify all recipient about new user
        notify.aboutAdd(added, Arrays.asList(recipientJID), filter, mock(RosterNotify.HashProvider.class));

        //three recipients
        assertThat(filter.getCalls(), is(1));

        //produces three packets
        assertThat(this.packetSender.count(), is(1));
        assertThat(this.packetSender.get(recipientJID, 0), is(packets.added(recipient, 0, as.renamed(added, "changed name"), Arrays.asList("group1", "group2"))));

    }

    @Test
    public void testAboutRemoveEmptyRecipients() {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User removed = as.user(1);

        notify.aboutRemove(removed, Collections.EMPTY_LIST, new TestFilter<PacketBuilder.RemoveUserPacket>(), mock(RosterNotify.HashProvider.class));

        //no notification should be produced
        assertThat(this.packetSender.count(), is(0));
    }

    @Test
    public void testAboutRemoveOffline() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User removed = as.user(1);
        BareJID addedJID = as.bare(removed);

        User offline = as.user(2);
        BareJID offlineJID = as.bare(offline);

        //notify all recipients about new user, sessions does report all offline by default, filter allows all by default
        notify.aboutRemove(removed, Arrays.asList(addedJID, offlineJID), new TestFilter<PacketBuilder.RemoveUserPacket>(), mock(RosterNotify.HashProvider.class));

        //no notification should be produced
        assertThat(this.packetSender.count(), is(0));
    }

    @Test
    public void testAboutRemoveOnlineBlockAll() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User removed = as.user(1);

        User onlineBlocked = as.user(2);
        BareJID onlineBlockedJID = as.bare(onlineBlocked);

        User onlineBlocked2 = as.user(3);
        BareJID onlineBlockedJID2 = as.bare(onlineBlocked2);

        //block both users
        TestFilter<PacketBuilder.RemoveUserPacket> filter = new TestFilter<PacketBuilder.RemoveUserPacket>();
        filter.block(onlineBlockedJID, onlineBlockedJID2);

        //report these as online
        this.sessions.online(onlineBlockedJID, onlineBlockedJID2);

        //notify all recipient about new user
        notify.aboutRemove(removed, Arrays.asList(onlineBlockedJID, onlineBlockedJID2), filter, mock(RosterNotify.HashProvider.class));

        assertThat(filter.getCalls(), is(2));

        //no notification should be produced
        assertThat(this.packetSender.count(), is(0));

    }

    @Test
    public void testAboutRemoveOnlineAllow() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User removed = as.user(1);

        User offline = as.user(2);
        BareJID offlineJID = as.bare(offline);

        User onlineBlocked = as.user(3);
        BareJID onlineBlockedJID = as.bare(onlineBlocked);

        User onlineAllowed = as.user(4);
        BareJID onlineAllowedJID = as.bare(onlineAllowed);

        User onlineAllowed2 = as.user(5);
        BareJID onlineAllowed2JID = as.bare(onlineAllowed2);

        //make all *Blocked blocked
        TestFilter<PacketBuilder.RemoveUserPacket> filter = new TestFilter<PacketBuilder.RemoveUserPacket>();
        filter.block(onlineBlockedJID);

        //make all online* online
        this.sessions.online(onlineBlockedJID, onlineAllowedJID, onlineAllowed2JID);

        //notify all recipient about new user
        notify.aboutRemove(removed, Arrays.asList(offlineJID, onlineBlockedJID, onlineAllowedJID, onlineAllowed2JID), filter, mock(RosterNotify.HashProvider.class));

        //three recipients
        assertThat(filter.getCalls(), is(3));

        //only two online
        assertThat(this.packetSender.count(), is(2));
        assertThat(this.packetSender.get(onlineAllowedJID, 0), is(packets.removed(onlineAllowed, 0, removed)));
        assertThat(this.packetSender.get(onlineAllowed2JID, 0), is(packets.removed(onlineAllowed2, 0, removed)));
    }

    @Test
    public void testRemoveAddTwoResources() {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User removed = as.user(1);

        User onlineWithThreeResources = as.user(2);
        BareJID onlineWithThreeResourcesJID = as.bare(onlineWithThreeResources);

        //allows all by default
        TestFilter<PacketBuilder.RemoveUserPacket> filter = new TestFilter<PacketBuilder.RemoveUserPacket>();

        //initialize three resources
        this.sessions.online(onlineWithThreeResourcesJID);
        this.sessions.resources(onlineWithThreeResourcesJID, 3);

        notify.aboutRemove(removed, Arrays.asList(onlineWithThreeResourcesJID), filter, mock(RosterNotify.HashProvider.class));

        //three recipients
        assertThat(filter.getCalls(), is(1));

        //produces three packets
        assertThat(this.packetSender.count(), is(3));
        assertThat(this.packetSender.get(onlineWithThreeResourcesJID, 0), is(packets.removed(onlineWithThreeResources, 0, removed)));
        assertThat(this.packetSender.get(onlineWithThreeResourcesJID, 1), is(packets.removed(onlineWithThreeResources, 1, removed)));
        assertThat(this.packetSender.get(onlineWithThreeResourcesJID, 2), is(packets.removed(onlineWithThreeResources, 2, removed)));
    }

    @Test
    public void testAboutRemoveModifiedPacket() throws NotAuthorizedException {
        RosterNotify notify = new RosterNotify(this.groups, this.packetBuilder, this.sessions, this.packetSender);

        User removed = as.user(1);

        User recipient = as.user(2);
        BareJID recipientJID = as.bare(recipient);

        PacketBuilder.RemoveUserPacket modifiedPacket = new PacketBuilder.RemoveUserPacket(as.bare(removed));
        TestFilter<PacketBuilder.RemoveUserPacket> filter = new TestFilter<PacketBuilder.RemoveUserPacket>();
        filter.passWith(recipientJID, modifiedPacket);

        this.sessions.online(recipientJID);

        //notify all recipient about new user
        notify.aboutRemove(removed, Arrays.asList(recipientJID), filter, mock(RosterNotify.HashProvider.class));

        //three recipients
        assertThat(filter.getCalls(), is(1));

        //produces three packets
        assertThat(this.packetSender.count(), is(1));
        assertThat(this.packetSender.get(recipientJID, 0), is(packets.removed(recipient, 0, as.renamed(removed, "changed name"))));
    }
}
